#include "ba0_exception.h"
#include "ba0_stack.h"
#include "ba0_garbage.h"
#include "ba0_analex.h"
#include "ba0_gmp.h"
#include "ba0_basic_io.h"
#include "ba0_global.h"
#include "ba0_x64_gmp.h"

#define gmp_alloc_function	ba0_global.gmp.gmp_alloc_function
#define gmp_realloc_function	ba0_global.gmp.gmp_realloc_function
#define gmp_free_function	ba0_global.gmp.gmp_free_function

BA0_DLL void ba0_set_settings_gmp (ba0_mp_set_memory_functions* set)
{
    ba0_gmp_set_memory_functions = set ? set : &mp_set_memory_functions;
}

BA0_DLL void ba0_get_settings_gmp (ba0_mp_set_memory_functions** set)
{
    if (set) *set = ba0_gmp_set_memory_functions;
}

BA0_DLL void ba0_record_gmp_memory_functions (void)
{
    mp_get_memory_functions
	(&gmp_alloc_function, &gmp_realloc_function, &gmp_free_function);
}


BA0_DLL void ba0_restore_gmp_memory_functions (void)
{
    (*ba0_gmp_set_memory_functions)
	(gmp_alloc_function, gmp_realloc_function, gmp_free_function);
}

BA0_DLL bool ba0_domain_mpz (void)
{
    return true;
}

BA0_DLL bool ba0_domain_mpq (void)
{
    return true;
}

BA0_DLL void* ba0_gmp_alloc (size_t n)
{
    ba0_global.gmp.gmp_alloc_function_called = true;
    if (n > 10000)
	ba0_process_check_interrupt ();
    return ba0_alloc ((ba0_int_p)n);
}

BA0_DLL void* ba0_gmp_realloc (void* p, size_t oldsize, size_t newsize)
{
    if (newsize <= oldsize)
        return p;
    else
    {   ba0_int_p *q;

	q = (ba0_int_p*) ba0_alloc ((ba0_int_p) newsize);
	memcpy (q, p, oldsize);
#if defined(BA0_HEAVY_DEBUG)
	{   struct ba0_stack *sp, *sq;
	    sp = ba0_which_stack (p);
	    sq = ba0_which_stack (q);
	    if (sp != sq)
	    {	
/*
		fprintf (stderr, "p = %lx, oldsize = %ld in %s\n", (unsigned long)p, (long)oldsize, sp ? sp->ident : "unknown");
		fprintf (stderr, "q = %lx, newsize = %ld in %s\n", (unsigned long)q, (long)newsize, sq ? sq->ident : "unknown");
*/
	    	BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    }
	}
#endif

        return q;
    }
}

/*
 * On Windows, GMP does not use gmp_realloc.
 * This causes memory problems in ba0_quiet_stack
 */

BA0_DLL void ba0_gmp_free (void* ptr, size_t size)
{
#if defined (_MSC_VER)
    size = ba0_allocated_size (size);
    if (ba0_stack_of_stacks [ba0_stack_sp] == &ba0_quiet_stack &&
	ptr >= ba0_quiet_stack.cells.tab 
				[ba0_quiet_stack.free.index_in_cells] &&
	(char*)ptr + size == (char*)ba0_quiet_stack.free.address)
    {
	  ba0_quiet_stack.free.address = ptr;
	  ba0_quiet_stack.free.ce_qui_reste += size;
    } 
#else
    ptr = 0;	/* To avoid annoying warnings */
    size = 0;
#endif
}

BA0_DLL __mpz_struct* ba0_new_mpz (void)
{   __mpz_struct* z;

    z = (__mpz_struct*) ba0_alloc (sizeof (__mpz_struct));
    ba0_mpz_init (z);
    return z;
}
    
/*
 * Reads a {\tt mpz_t} through the lexical analyzer and stores it
 * in~$z$ if $z$ is nonzero. Allocates otherwise a new {\tt mpz_t}.
 * Returns the read {\tt mpz_t}.
 */

BA0_DLL void* ba0_scanf_mpz (void* z)
{   __mpz_struct* c;
    bool minus;

    if (z == (void*)0)
	c = ba0_new_mpz ();
     else
        c = (__mpz_struct*)z;
    if (ba0_sign_token_analex ("-"))
    {	minus = true;
	ba0_get_token_analex ();
    } else
	minus = false;
    if (ba0_type_token_analex () != ba0_integer_token)
        BA0_RAISE_PARSER_EXCEPTION (BA0_ERRINT);
    mpz_set_str (c, ba0_value_token_analex (), (int)10);
    if (minus)
	ba0_mpz_neg (c, c);
    return c;
}
 
/*
 * Prints the {\tt mpz_t} $z$.
 */

BA0_DLL void ba0_printf_mpz (void* z)
{   __mpz_struct* c = (__mpz_struct*)z;
    char* s;
    struct ba0_mark M;

    ba0_record (&M);
    s = mpz_get_str ((char*)0, 10, c);
    ba0_put_string (s);
    ba0_restore (&M);
}

/*
 * Readonly static data
 */

static char mpz_struct_ [] = "__mpz_struct";
static char mpz_struct_mp_d_ [] = "__mpz_struct._mp_d";

BA0_DLL ba0_int_p ba0_garbage1_mpz (void* _c, enum ba0_garbage_code code)
{   __mpz_struct* c = (__mpz_struct*)_c;
    ba0_int_p n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info (c, sizeof (__mpz_struct), mpz_struct_);
    n += ba0_new_gc_info 
	   (c->_mp_d, c->_mp_alloc * sizeof (mp_limb_t), mpz_struct_mp_d_);
    return n;
}
 
BA0_DLL void* ba0_garbage2_mpz (void* _c, enum ba0_garbage_code code)
{   __mpz_struct* c;

    if (code == ba0_isolated)
	c = (__mpz_struct*)ba0_new_addr_gc_info (_c, mpz_struct_);
    else
	c = (__mpz_struct*)_c;

    c->_mp_d = ba0_new_addr_gc_info (c->_mp_d, mpz_struct_mp_d_);
    return c;
}

BA0_DLL void* ba0_copy_mpz (void* z)
{   __mpz_struct* c;

    c = ba0_new_mpz ();
    ba0_mpz_init_set (c, (__mpz_struct*)z);
    return c;
}
 
BA0_DLL __mpq_struct* ba0_new_mpq (void)
{   __mpq_struct* q;

    q = (__mpq_struct*)ba0_alloc (sizeof (__mpq_struct));
    ba0_mpq_init (q);
    return q;
}

/*
 * Reads a ba0_mpq_t through the lexical analyzer and assigns it to z,
 * if z in nonzero. Exceptions BA0_ERRRAT, BA0_ERRDPZ may be raised.
 */

BA0_DLL void* ba0_scanf_mpq (void* z)
{   __mpq_struct* q;
    struct ba0_mark M;
    ba0_int_p token_lmax;
    char* buf1;
    char* buf2;
    bool minus_sign;
    bool mantissa;
    ba0_int_p exponent;

    ba0_get_settings_analex (0, &token_lmax);
    ba0_push_another_stack ();
    ba0_record (&M);
    buf1 = (char*)ba0_alloc (token_lmax);
    buf2 = (char*)ba0_alloc (token_lmax);
    ba0_pull_stack ();

    if (z == (void*)0)
	q = ba0_new_mpq ();
    else
	q = (__mpq_struct*)z;
/*
   First the sign
*/
    if (ba0_sign_token_analex ("-"))
    {	minus_sign = true;
	ba0_get_token_analex ();
    } else
    {   if (ba0_sign_token_analex ("+"))
	    ba0_get_token_analex ();
	minus_sign = false;
    }
/*
   Possibly a non empty integer part
   Spaces are authorized between the minus sign and what follows.
   Mantissa = true iff an integer is read in buf1.
*/
    mantissa = false;
    buf1 [0] = '\0';
    if (ba0_type_token_analex () == ba0_integer_token)
    {   strcat (buf1, ba0_value_token_analex ());
        ba0_get_token_analex ();
        mantissa = true;
    }
/*
   If no integer is read then there must be a dot + a fractionnal part
*/
    if (! mantissa && ! ba0_sign_token_analex ("."))
	BA0_RAISE_PARSER_EXCEPTION (BA0_ERRRAT);
/*
   The case of the double notation
*/
    strcpy (buf2, ba0_value_token_analex ());
    if ((! mantissa || ! ba0_spaces_before_token_analex ()) &&
		(ba0_sign_token_analex (".") || 
		 ((buf2 [0] == 'e' || buf2 [0] == 'E' || buf2 [0] == '@') &&
		  (buf2 [1] == '\0' || isdigit ((int)buf2 [1])))))
    {	exponent = 0;
/*
   At the end, the integer read is multiplied by 10**exponent.
*/
	if (ba0_sign_token_analex ("."))
	{   ba0_get_token_analex ();
	    if (! ba0_spaces_before_token_analex () && 
			ba0_type_token_analex () == ba0_integer_token)
	    {   strcat (buf1, ba0_value_token_analex ());
	        exponent = - (ba0_int_p)strlen (ba0_value_token_analex ());
	        ba0_get_token_analex ();
	    }
	}
/*
   Possibly an exponent in the double notation.
*/
	strcpy (buf2, ba0_value_token_analex ());
	if (! ba0_spaces_before_token_analex () &&
		(buf2 [0] == 'e' || buf2 [0] == 'E' || buf2 [0] == '@') &&
		(buf2 [1] == '\0' || isdigit ((int)buf2 [1])))
	{   
/*
   To simplify the coming analysis, the current token may be split
   into two different tokens
*/
	    if (isdigit ((int)buf2 [1]))
	    {   ba0_int_p i;
	        i = 2;
	        while (isdigit ((int)buf2 [i]))
		    i++;
	        if (buf2 [i] == '\0')
		    ba0_unget_given_token_analex 
					(buf2+1, ba0_integer_token, false);
	        else
	        {   ba0_unget_given_token_analex 
					(buf2+i, ba0_string_token, false);
		    buf2 [i] = '\0';
		    ba0_unget_given_token_analex 
					(buf2+1, ba0_integer_token, false);
	        }
	    }
/*
   The analysis of the exponent starts here
*/
	    ba0_get_token_analex ();
	    if (! ba0_spaces_before_token_analex () && 
			ba0_type_token_analex () == ba0_integer_token)
		exponent = exponent + atoi (ba0_value_token_analex ());
	    else if (! ba0_spaces_before_token_analex () && 
					(ba0_sign_token_analex ("+") || 
					 ba0_sign_token_analex ("-")))
	    {   bool minus_sign2 = ba0_sign_token_analex ("-");
	        ba0_get_token_analex ();
	        if (ba0_spaces_before_token_analex () ||
				ba0_type_token_analex () != ba0_integer_token)
		    BA0_RAISE_PARSER_EXCEPTION (BA0_ERRRAT);
		if (minus_sign2)
		    exponent -= atoi (ba0_value_token_analex ());
		else
		    exponent += atoi (ba0_value_token_analex ());
	    } else
	        BA0_RAISE_PARSER_EXCEPTION (BA0_ERRRAT);
        } else
	    ba0_unget_token_analex ();

	mpz_set_str (mpq_numref (q), buf1, 10);
	if (exponent > 0)
	{   ba0_mpz_ui_pow_ui (mpq_denref (q), 10, exponent);
	    ba0_mpz_mul (mpq_numref (q), mpq_numref (q), mpq_denref (q));
	    ba0_mpz_set_ui (mpq_denref (q), 1);
	} else
	{   ba0_mpz_ui_pow_ui (mpq_denref (q), 10, - exponent);
	    mpq_canonicalize (q);
	}
    } else
/*
   The traditional rational notation
*/
    {	mpz_set_str (mpq_numref (q), buf1, 10);

        if (ba0_sign_token_analex ("/"))
        {   ba0_get_token_analex ();
/*
   A minus sign is allowed before the denominator
*/
	    if (ba0_sign_token_analex ("-") || ba0_sign_token_analex ("+"))
	    {	if (ba0_sign_token_analex ("-"))
		    minus_sign = ! minus_sign;
		ba0_get_token_analex ();
	    }
	    if (ba0_type_token_analex () != ba0_integer_token)
	        BA0_RAISE_PARSER_EXCEPTION (BA0_ERRRAT);
	    mpz_set_str (mpq_denref (q), ba0_value_token_analex (), 10);
	    if (ba0_mpz_sgn (mpq_denref (q)) == 0)
	        BA0_RAISE_PARSER_EXCEPTION (BA0_ERRIVZ);
	    mpq_canonicalize (q);
        } else
	    ba0_unget_token_analex ();
    }
    if (minus_sign)
	ba0_mpq_neg (q, q);
    ba0_restore (&M);
    return q;
}

BA0_DLL void ba0_printf_mpq (void* z)
{   __mpq_struct* q = (__mpq_struct*) z;
    char* s;
    struct ba0_mark M;

    ba0_record (&M);
    s = mpz_get_str ((char*)0, 10, mpq_numref (q));
    ba0_put_string (s);
    if (ba0_mpz_cmp_ui (mpq_denref (q), 1) != 0)
    {	ba0_put_char ('/');
	s = mpz_get_str ((char*)0, 10, mpq_denref (q));
	ba0_put_string (s);
    }
    ba0_restore (&M);
}

/*
 * Readonly static data
 */

static char mpq_struct_ [] = "__mpq_struct";
static char mpq_struct_mp_num_ [] = "__mpq_struct._mp_num._mp_d";
static char mpq_struct_mp_den_ [] = "__mpq_struct._mp_den._mp_d";

BA0_DLL ba0_int_p ba0_garbage1_mpq (void* _q, enum ba0_garbage_code code)
{   __mpq_struct* q = (__mpq_struct*)_q;
    ba0_int_p n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info (q, sizeof (__mpq_struct), mpq_struct_);

    n += ba0_new_gc_info 
		(q->_mp_num._mp_d,
		 q->_mp_num._mp_alloc * sizeof (mp_limb_t), 
		 mpq_struct_mp_num_);
    n += ba0_new_gc_info
		(q->_mp_den._mp_d,
		 q->_mp_den._mp_alloc * sizeof (mp_limb_t),
		 mpq_struct_mp_den_);
    return n;
}
 
BA0_DLL void* ba0_garbage2_mpq (void* _q, enum ba0_garbage_code code)
{   __mpq_struct* q;

    if (code == ba0_isolated)
	q = (__mpq_struct*)ba0_new_addr_gc_info (_q, mpq_struct_);
    else
	q = (__mpq_struct*)_q;

    q->_mp_num._mp_d = ba0_new_addr_gc_info 
			(q->_mp_num._mp_d, mpq_struct_mp_num_);
    q->_mp_den._mp_d = ba0_new_addr_gc_info
			(q->_mp_den._mp_d, mpq_struct_mp_den_);
    return q;
}

BA0_DLL void* ba0_copy_mpq (void* z)
{   __mpq_struct* q;

    q = ba0_new_mpq ();
    ba0_mpq_init (q);
    ba0_mpq_set (q, (__mpq_struct*)z);
    return q;
}

