#include "ba0_exception.h"
#include "ba0_stack.h"
#include "ba0_list.h"
#include "ba0_table.h"
#include "ba0_array.h"
#include "ba0_matrix.h"
#include "ba0_format.h"
#include "ba0_scanf.h"
#include "ba0_basic_io.h"
#include "ba0_analex.h"
#include "ba0_int_p.h"
#include "ba0_global.h"

/*
   The format f describes the type of the elements of the list
   po = opening bracket
   pf = closing bracket
   struct_list_in_another_stack : if true then the struct ba0_list are 
   allocated in another stack than the list values. This is useful for parsing 
   tables. If false then the list gets allocated in the current stack.
*/

static ba0_list ba0_scanf_list 
	(ba0_format f, char po, char pf, bool struct_list_in_another_stack)
{   char delimiteur [] = " ";
    ba0_list L = (ba0_list)0;

    delimiteur [0] = po;
    if (! ba0_sign_token_analex (delimiteur))
	BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);

    ba0_get_token_analex ();
    delimiteur [0] = pf;
    if (ba0_sign_token_analex (delimiteur))
	return L;

    for (;;)
    {	void* value;
/*
   Lists of ba0_int_p are a special case :
   We want the integers to be stored in the pointers.
*/
	if (f->linknmb > 0 && 
		(f->link [0]->u.leaf.scanf == &ba0_scanf_int_p ||
		 f->link [0]->u.leaf.scanf == &ba0_scanf_hexint_p))
	    ba0_scanf_int_p (&value);
	else
	    ba0__scanf__ (f, &value, false);
	if (struct_list_in_another_stack)
	{   ba0_push_another_stack ();
	    L = ba0_cons_list (value, L);
	    ba0_pull_stack ();
	} else
	    L = ba0_cons_list (value, L);
	ba0_get_token_analex ();
	if (! ba0_sign_token_analex (","))
	    break;
	ba0_get_token_analex ();
    }

    if (! ba0_sign_token_analex (delimiteur))
	BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);

    return ba0_reverse_list (L);
}

static ba0_table ba0_scanf_table (ba0_format f, char po, char pf, ba0_table T)
{   ba0_list L;
    struct ba0_mark M;
/*
   A table is parsed as a list.
   A mark is set in another heap before parsing the list
*/
    ba0_push_another_stack ();
    ba0_record (&M);
    ba0_pull_stack ();

    L = ba0_scanf_list (f, po, pf, true);

    if (T == (ba0_table)0)
	T = ba0_new_table ();
    ba0_set_table_list (T, L);
/*
   The struct ba0_list are freed
*/
    ba0_restore (&M);

    return T;
}

/*
   There is some waste of memory.
   It could be saved by applying the garbage collector but this would
   force the user to write a garbage collector in order to parse arrays.
*/

static ba0_array ba0_scanf_array (ba0_format f, char po, char pf, ba0_array A)
{   ba0_list L;
    ba0_int_p sizelt, length, i;
    struct ba0_mark M;
/*
   An array is parsed as a list.
   A mark is set in another heap before parsing the list
*/
    ba0_push_another_stack ();
    ba0_record (&M);
    ba0_pull_stack ();

    if (f->linknmb == 0 || f->link [0]->u.leaf.sizelt < 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    L = ba0_scanf_list (f, po, pf, true);
    length = ba0_length_list (L);

    if (A == (ba0_array)0)
	A = ba0_new_array ();
    sizelt = f->link [0]->u.leaf.sizelt;
    ba0_realloc_array (A, length, sizelt);
/*
   Machine integers are a special case. They are stored in list pointers.
*/
    if (f->link [0]->u.leaf.scanf == &ba0_scanf_int_p ||
                 f->link [0]->u.leaf.scanf == &ba0_scanf_hexint_p)
    {	for (i = 0; i < length; i++)
	{   memcpy (A->tab + i * sizelt, &L->value, sizelt);
	    L = L->next;
	}
    } else
    {	for (i = 0; i < length; i++)
	{   memcpy (A->tab + i * sizelt, L->value, sizelt);
	    L = L->next;
	}
    }
    A->size = length; 
/*
   The struct ba0_list are freed
*/
    ba0_restore (&M);

    return A;
}

/*
   If struct_list_in_another_stack then the struct ba0_list of the outer
   list and the inner list are allocated in another stack than the current one.
*/

static ba0_list ba0_scanf_listof_list 
	(ba0_format f, char po, char pf, bool struct_list_in_another_stack)
{   char delimiteur [] = " ";
    ba0_list L = (ba0_list)0;

    delimiteur [0] = po;
    if (! ba0_sign_token_analex (delimiteur))
	BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);

    ba0_get_token_analex ();
    delimiteur [0] = pf;
    if (ba0_sign_token_analex (delimiteur))
	return L;

    for (;;)
    {	ba0_list value;
	value = ba0_scanf_list (f, po, pf, struct_list_in_another_stack);

	if (struct_list_in_another_stack)
	{   ba0_push_another_stack ();
	    L = ba0_cons_list (value, L);
	    ba0_pull_stack ();
	} else
	    L = ba0_cons_list (value, L);

	ba0_get_token_analex ();
	if (! ba0_sign_token_analex (","))
	    break;
	ba0_get_token_analex ();
    }

    if (! ba0_sign_token_analex (delimiteur))
	BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);

    return ba0_reverse_list (L);
}

static ba0_matrix ba0_scanf_matrix 
				(ba0_format f, char po, char pf, ba0_matrix M)
{   ba0_list L, LL;
    ba0_int_p m, n, o, i, j;
    struct ba0_mark mark;
/*
   A matrix is parsed as a list of lists.
   A mark is set in another heap before parsing the list
*/
    ba0_push_another_stack ();
    ba0_record (&mark);
    ba0_pull_stack ();

    L = ba0_scanf_listof_list (f, po, pf, true);

    n = ba0_length_list (L);
    m = 0;
    for (LL = L; LL != (ba0_list)0; LL = LL->next)
    {	o = ba0_length_list ((ba0_list)LL->value);
	m = BA0_MAX (m, o);
    }
    if (M == (ba0_matrix)0)
	M = ba0_new_matrix ();
    ba0_realloc_matrix (M, n, m);
    M->nrow = n;
    M->ncol = m;
    for (i = 0; i < n; i++)
    {	LL = (ba0_list)L->value;
	for (j = 0; j < m; j++)
	{   BA0_MAT (M, i, j) = LL->value;
	    LL = LL->next;
	}
	L = L->next;
    }
/*
   The struct ba0_list are freed
*/
    ba0_restore (&mark);

    return M;
}

/*
   If toplevel is true
	Then tabres contains the addresses of the variables which must
	receive the read data. These variables are assumed to be initialized.
   If toplevel is false
	Then the read data are stored directly in the tabres entries. 
	The read data are stored in newly allocated areas.
*/

BA0_DLL void ba0__scanf__ (ba0_format f, void** tabres, bool toplevel)
{   static char buffer [256];
    ba0_int_p i = 0, j = 0;

    ba0_unget_token_analex ();
    while (isspace ((int)f->text [i]))
	i++;
    while (f->text [i] != '\0')
    {	ba0_get_token_analex ();
	if (isalpha ((int)f->text [i]))
	{   char* q = buffer;
	    while (isalnum ((int)f->text [i]) || f->text [i] == '_')
		*q++ = f->text [i++];
	    *q = '\0';
	    if (ba0_type_token_analex () != ba0_string_token ||
			strcmp (ba0_value_token_analex (), buffer) != 0)
		BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
	} else if (f->text [i] == '%')
	{   i++;
	    switch (f->link [j]->code)
	    {	case ba0_leaf_format:
		    if (toplevel)
			(*f->link [j]->u.leaf.scanf) (tabres [j]);
		    else
			tabres [j] = (*f->link [j]->u.leaf.scanf) ((void*)0);
		    break;
		case ba0_list_format:
		    if (toplevel)
			*(ba0_list*)tabres [j] = 
				ba0_scanf_list (f->link [j]->u.node.op,
						f->link [j]->u.node.po, 
						f->link [j]->u.node.pf, false);
		    else
			tabres [j] = ba0_scanf_list (f->link [j]->u.node.op, 
						f->link [j]->u.node.po, 
						f->link [j]->u.node.pf, false);
		    break;
		case ba0_table_format:
		    if (toplevel)
			ba0_scanf_table (f->link [j]->u.node.op,
				f->link [j]->u.node.po, 
				f->link [j]->u.node.pf, 
				(ba0_table)tabres [j]);
		    else
			tabres [j] = ba0_scanf_table (f->link [j]->u.node.op,
						f->link [j]->u.node.po, 
						f->link [j]->u.node.pf,
						(ba0_table)0);
		    break;
		case ba0_array_format:
		    if (toplevel)
			ba0_scanf_array (f->link [j]->u.node.op,
				f->link [j]->u.node.po, 
				f->link [j]->u.node.pf, 
				(ba0_array)tabres [j]);
		    else
			tabres [j] = ba0_scanf_array (f->link [j]->u.node.op,
						f->link [j]->u.node.po, 
						f->link [j]->u.node.pf,
						(ba0_array)0);
		    break;
		case ba0_matrix_format:
		    if (toplevel)
			ba0_scanf_matrix (f->link [j]->u.node.op,
				f->link [j]->u.node.po,
				f->link [j]->u.node.pf,
				(ba0_matrix)tabres [j]);
		    else
			tabres [j] = ba0_scanf_matrix (f->link [j]->u.node.op,
                                                f->link [j]->u.node.po,
                                                f->link [j]->u.node.pf,
						(ba0_matrix)0);
                    break;
	    }
	    j++;
	} else
	{   buffer [0] = f->text [i++];
	    if (buffer [0] == ':')
	    {	buffer [1] = f->text [i++];
		buffer [2] = '\0';
	    } else
		buffer [1] = '\0';
	    if (! ba0_sign_token_analex (buffer))
		BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
	}
	while (isspace ((int)f->text [i]))
	    i++;
    }
}

/*
   ba0_scanf (format-string, address1, ..., addressk)
   The current token should be the first token to be read.
*/

BA0_DLL void ba0_scanf (char* s, ...)
{   ba0_format f;
    va_list arg;
    void** tabres;
    ba0_int_p i;
    struct ba0_mark M;

    f = ba0_get_format (s);

    ba0_push_another_stack ();
    ba0_record (&M);

    tabres = (void**) ba0_alloc (sizeof (void*) * f->linknmb);

    ba0_pull_stack ();

    va_start (arg, s);
    for (i = 0; i < f->linknmb; i++)
 	tabres [i] = va_arg (arg, void*);
    va_end (arg);

    ba0__scanf__ (f, tabres, true);

    ba0_restore (&M);
}

/*
   Varant of the above one, which reads the first token to be read.
*/

BA0_DLL void ba0_scanf2 (char* s, ...)
{   ba0_format f;
    va_list arg;
    void** tabres;
    ba0_int_p i;
    struct ba0_mark M;

    f = ba0_get_format (s);

    ba0_push_another_stack ();
    ba0_record (&M);

    tabres = (void**) ba0_alloc (sizeof (void*) * f->linknmb);

    ba0_pull_stack ();

    va_start (arg, s);
    for (i = 0; i < f->linknmb; i++)
        tabres [i] = va_arg (arg, void*);
    va_end (arg);

    ba0_get_token_analex ();

    ba0__scanf__ (f, tabres, true);

    ba0_restore (&M);
}

/*
   Variant of ba0_scanf
*/

BA0_DLL void ba0_sscanf (char* buffer, char* s, ...)
{   ba0_format f;
    va_list arg;
    void** tabres;
    struct ba0_exception_code code;
    ba0_int_p i;
    struct ba0_mark M;

    f = ba0_get_format (s);

    ba0_push_another_stack ();
    ba0_record (&M);

    tabres = (void**) ba0_alloc (sizeof (void*) * f->linknmb);

    ba0_pull_stack ();

    va_start (arg, s);
    for (i = 0; i < f->linknmb; i++)
 	tabres [i] = va_arg (arg, void*);
    va_end (arg);

    ba0_record_analex ();
    ba0_set_analex_string (buffer);

    BA0_PUSH_EXCEPTION (code);
    if (ba0_exception_is_set (code))
    {	ba0__scanf__ (f, tabres, true);
	ba0_pull_exception (code);
	ba0_restore_analex ();
	ba0_restore (&M);
    } else
    {	ba0_restore_analex ();
	BA0_RAISE_EXCEPTION (ba0_mesgerr);
    }
}

/*
   Variant of ba0_scanf2
*/

BA0_DLL void ba0_sscanf2 (char* buffer, char* s, ...)
{   ba0_format f;
    va_list arg;
    void** tabres;
    struct ba0_exception_code code;
    ba0_int_p i;
    struct ba0_mark M;

    f = ba0_get_format (s);

    ba0_push_another_stack ();
    ba0_record (&M);

    tabres = (void**) ba0_alloc (sizeof (void*) * f->linknmb);

    ba0_pull_stack ();

    va_start (arg, s);
    for (i = 0; i < f->linknmb; i++)
 	tabres [i] = va_arg (arg, void*);
    va_end (arg);

    ba0_record_analex ();
    ba0_set_analex_string (buffer);

    BA0_PUSH_EXCEPTION (code);
    if (ba0_exception_is_set (code))
    {	ba0_get_token_analex ();
	ba0__scanf__ (f, tabres, true);
	ba0_pull_exception (code);
	ba0_restore_analex ();
	ba0_restore (&M);
    } else
    {	ba0_restore_analex ();
	BA0_RAISE_EXCEPTION (ba0_mesgerr);
    }
}

BA0_DLL void ba0_fscanf (FILE* file, char* s, ...)
{   ba0_format f;
    va_list arg;
    void** tabres;
    struct ba0_exception_code code;
    ba0_int_p i;
    struct ba0_mark M;

    f = ba0_get_format (s);

    ba0_push_another_stack ();
    ba0_record (&M);

    tabres = (void**) ba0_alloc (sizeof (void*) * f->linknmb);

    ba0_pull_stack ();

    va_start (arg, s);
    for (i = 0; i < f->linknmb; i++)
 	tabres [i] = va_arg (arg, void*);
    va_end (arg);

    ba0_record_analex ();
    ba0_set_analex_FILE (file);

    BA0_PUSH_EXCEPTION (code);
    if (ba0_exception_is_set (code))
    {	ba0__scanf__ (f, tabres, true);
	ba0_pull_exception (code);
	ba0_restore_analex ();
	ba0_restore (&M);
    } else
    {	ba0_restore_analex ();
	BA0_RAISE_EXCEPTION (ba0_mesgerr);
    }
}

BA0_DLL void ba0_fscanf2 (FILE* file, char* s, ...)
{   ba0_format f;
    va_list arg;
    void** tabres;
    struct ba0_exception_code code;
    ba0_int_p i;
    struct ba0_mark M;

    f = ba0_get_format (s);

    ba0_push_another_stack ();
    ba0_record (&M);

    tabres = (void**) ba0_alloc (sizeof (void*) * f->linknmb);

    ba0_pull_stack ();

    va_start (arg, s);
    for (i = 0; i < f->linknmb; i++)
 	tabres [i] = va_arg (arg, void*);
    va_end (arg);

    ba0_record_analex ();
    ba0_set_analex_FILE (file);

    BA0_PUSH_EXCEPTION (code);
    if (ba0_exception_is_set (code))
    {	ba0_get_token_analex ();
	ba0__scanf__ (f, tabres, true);
	ba0_pull_exception (code);
	ba0_restore_analex ();
	ba0_restore (&M);
    } else
    {	ba0_restore_analex ();
	BA0_RAISE_EXCEPTION (ba0_mesgerr);
    }
}
