#include "ba0_stack.h"
#include "ba0_exception.h"
#include "ba0_analex.h"
#include "ba0_global.h"

/*
 * A magic number used to check memory overflow
 */

#ifdef BA0_MEMCHECK
#ifdef BA0_64BITS
static ba0_int_p magic = 0x3352074971119021;
#else
static ba0_int_p magic = 0x33520749;
#endif
#endif

#define system_malloc	ba0_initialized_global.stack.system_malloc
#define system_free	ba0_initialized_global.stack.system_free
#define no_oom		ba0_initialized_global.malloc.no_oom

/*
   Every call to malloc should go through the following function.
*/

BA0_DLL void* ba0_malloc (ba0_int_p n)
{   void* m;

    if (no_oom == false && n + ba0_malloc_counter > ba0_memory_limit)
	BA0_RAISE_EXCEPTION (BA0_ERROOM);
    m = (*system_malloc) (n);
    if (m == (void*)0)
	BA0_RAISE_EXCEPTION (BA0_ERROOM);
    ba0_malloc_counter += n;
    ba0_malloc_nbcalls++;
/*
    fprintf (stderr, "%s\n", mesg);
*/
    return m;
}

/*
 * Same function as above but the counters are not incremented.
 * The allocated memory is not assumed to be freed by ba0_terminate.
 */

BA0_DLL void* ba0_persistent_malloc (ba0_int_p n)
{   void* m;

    if (no_oom == false && n + ba0_malloc_counter > ba0_memory_limit)
	BA0_RAISE_EXCEPTION (BA0_ERROOM);
    m = (*system_malloc) (n);
    if (m == (void*)0)
	BA0_RAISE_EXCEPTION (BA0_ERROOM);
    return m;
}

BA0_DLL void ba0_free (void* p)
{
    ba0_malloc_nbcalls--;
/*
    fprintf (stderr, "nbfree = %d (%s)\n", ba0_malloc_nbcalls, mesg);
*/
    (*system_free) (p);
}

BA0_DLL void ba0_set_settings_no_oom (bool b)
{
    no_oom = b ? b : false;
}

BA0_DLL void ba0_get_settings_no_oom (bool* b)
{
    if (b) *b = no_oom;
}

/*
 * To be called before ba0_restart !
 */

BA0_DLL void ba0_set_settings_memory_functions 
				(void* (*m) (size_t), void (*f) (void*))
{
    system_malloc = m ? m : &malloc;
    system_free = f ? f : &free;
}

BA0_DLL void ba0_get_settings_memory_functions 
				(void* (**m) (size_t), void (**f) (void*))
{
    if (m) *m = system_malloc;
    if (f) *f = system_free;
}

/*
 * Returns the smallest integer m greater than or equal to n such that
 * m is a multiple of BA0_ALIGN (memory alignment).
 */

BA0_DLL unsigned ba0_int_p ba0_ceil_align (unsigned ba0_int_p n)
{
    if ((n & (BA0_ALIGN - 1)) != 0)
	n = (n & -BA0_ALIGN) + BA0_ALIGN;
    return n;
}

/*
 * Returns the size actually allocated by ba0_alloc.
 * Beware to the fact that, on some architectures, 
 * 		BA0_ALIGN > sizeof (ba0_int_p)
 */

BA0_DLL unsigned ba0_int_p ba0_allocated_size (unsigned ba0_int_p n)
{
    n = ba0_ceil_align (n);
#if defined (BA0_MEMCHECK)
    n += BA0_ALIGN;
#endif
    return n;
}

/*************************************************************
 * STACKS SETTINGS
 *************************************************************/

#define ba0_sizeof_main_cell	ba0_initialized_global.stack.sizeof_main_cell
#define ba0_sizeof_quiet_cell	ba0_initialized_global.stack.sizeof_quiet_cell
#define ba0_sizeof_analex_cell	ba0_initialized_global.stack.sizeof_analex_cell
#define ba0_nb_cells_per_stack	ba0_initialized_global.stack.nb_cells_per_stack
#define ba0_sizeof_stack_of_stack	ba0_initialized_global.stack.sizeof_stack_of_stack

BA0_DLL void ba0_set_settings_stack (ba0_int_p main, ba0_int_p quiet, 
			     ba0_int_p analex, ba0_int_p nbcells, 
			     ba0_int_p sizess)
{
    main = ba0_ceil_align (main);
    quiet = ba0_ceil_align (quiet);
    analex = ba0_ceil_align (analex);

    ba0_sizeof_main_cell = main ? main : BA0_SIZE_CELL_MAIN_STACK;
    ba0_sizeof_quiet_cell = quiet ? quiet : BA0_SIZE_CELL_QUIET_STACK;
    ba0_sizeof_analex_cell = analex > 2 * BA0_TOKEN_LMAX ? analex : 
					BA0_SIZE_CELL_ANALEX_STACK;
    ba0_nb_cells_per_stack = nbcells ? nbcells : BA0_NB_CELLS_PER_STACK;
    ba0_sizeof_stack_of_stack = sizess ? sizess : BA0_SIZE_STACK_OF_STACK;
}

BA0_DLL void ba0_get_settings_stack 
			    (ba0_int_p* main, ba0_int_p* quiet, 
			     ba0_int_p* analex, ba0_int_p* nbcells, 
			     ba0_int_p* sizess)
{
    if (main) *main = ba0_sizeof_main_cell;
    if (quiet) *quiet = ba0_sizeof_quiet_cell;
    if (analex) *analex = ba0_sizeof_analex_cell;
    if (nbcells) *nbcells = ba0_nb_cells_per_stack;
    if (sizess) *sizess = ba0_sizeof_stack_of_stack;
}

/**********************************************************************
 STACKS
 **********************************************************************/

static void ba0_set_mark (struct ba0_mark*, struct ba0_stack*,
                        ba0_int_p, void* address, unsigned ba0_int_p);

/*
   For debugging purpose.
   Returns true if p points to an area located in the used part of H.
*/

BA0_DLL bool ba0_in_stack (void* p, struct ba0_stack* H)
{
    return ba0_cell_index_mark (p, &H->free) != -1;
}

/*
 * Used by ba0_raise_exception2.
 * Returns the stack in which p points to.
 * Does not investigate all the stacks.
 */

BA0_DLL struct ba0_stack* ba0_which_stack (void* p)
{
    if (ba0_in_stack (p, &ba0_main_stack))
	return &ba0_main_stack;
    else if (ba0_in_stack (p, &ba0_second_stack))
	return &ba0_second_stack;
    else if (ba0_in_stack (p, &ba0_quiet_stack))
	return &ba0_quiet_stack;
    else
	return (struct ba0_stack*)0;
}

/*
 * Returns the current stack
 */

struct ba0_stack* ba0_current_stack (void)
{
    return ba0_stack_of_stacks [ba0_stack_sp];
}

/*
   For information.
   Returns the maximum allocated memory in the stack H
*/

BA0_DLL unsigned ba0_int_p ba0_max_alloc_stack (struct ba0_stack* H)
{
    return (H->max_alloc.index_in_cells + 1) * H->cell_size - 
		H->max_alloc.ce_qui_reste;
}

/*
   Initializes H
   Called by ba0_restart. 
*/

BA0_DLL void ba0_init_stack (struct ba0_stack* H)
{
    if (H == &ba0_main_stack)
    {	H->cell_size = ba0_sizeof_main_cell;
	H->ident = "ba0_main_stack";
    } else if (H == &ba0_second_stack)
    {	H->cell_size = ba0_sizeof_main_cell;
	H->ident = "ba0_second_stack";
    } else if (H == &ba0_analex_stack)
    {   H->cell_size = ba0_sizeof_analex_cell;
        H->ident = "ba0_analex_stack";
    } else if (H == &ba0_quiet_stack)
    {	H->cell_size = ba0_sizeof_quiet_cell;
	H->ident = "ba0_quiet_stack";
    } else if (H == &ba0_format_stack)
    {	H->cell_size = ba0_sizeof_quiet_cell;
	H->ident = "ba0_format_stack";
    } else 
    {	H->cell_size = ba0_sizeof_quiet_cell;
	H->ident = "a non predefined stack";
    } 

    H->cells.tab = (void**)ba0_malloc 
	(sizeof(void*)*ba0_nb_cells_per_stack);
    H->cells.size = 0;
    H->cells.alloc = ba0_nb_cells_per_stack;
    memset (H->cells.tab, 0, H->cells.alloc * sizeof(void*));

    ba0_set_mark (&H->free, H, -1, (void*)0, 0);
    H->max_alloc = H->free;
    H->nb_calls_to_alloc = 0;
#if defined (BA0_MEMCHECK)
    H->bound = &magic;
#endif
}

BA0_DLL void ba0_init_one_cell_stack 
	(struct ba0_stack* H, char* ident, void* cell, ba0_int_p cell_size)
{
    H->ident = ident;
    H->cells.tab = (void**)ba0_malloc (sizeof(void*));
    H->cells.tab [0] = cell;
    H->cells.size = 1;
    H->cells.alloc = 1;

    H->cell_size = cell_size;

    ba0_set_mark (&H->free, H, -1, (void*)0, 0);
    H->max_alloc = H->free;

    H->nb_calls_to_alloc = 0;
#if defined (BA0_MEMCHECK)
    H->bound = &magic;
#endif
}

BA0_DLL void ba0_clear_one_cell_stack (struct ba0_stack* H)
{
    ba0_free (H->cells.tab);
}

/*
   Logically frees everything in H.
*/

BA0_DLL void ba0_reset_stack (struct ba0_stack* H)
{
    ba0_set_mark (&H->free, H, -1, (void*)0, 0);
#if defined (BA0_MEMCHECK)
    H->bound = &magic;
#endif
}

/*
   Physically frees the cells
   Called by ba0_terminate. 
*/

BA0_DLL void ba0_clear_cells_stack (struct ba0_stack* H)
{
    ba0_reset_stack (H);
    while (H->cells.size > 0)
    {	ba0_free (H->cells.tab [H->cells.size - 1]);
	H->cells.tab [H->cells.size - 1] = (void*)0;
	ba0_malloc_counter -= H->cell_size;
	H->cells.size--;
    }
}

BA0_DLL void ba0_clear_stack (struct ba0_stack* H)
{
    ba0_clear_cells_stack (H);
    ba0_free (H->cells.tab);
    ba0_malloc_counter -= sizeof(void*)*H->cells.alloc;
}

/*
   Subfunction of ba0_alloc.

   Allocates at least n bytes in the current stack.

   An extra area is appended to the allocated n bytes to store a magic number.
   The magic number is *not* stored.
*/

static void* ba0_alloc_but_do_not_set_magic_mark 
				(struct ba0_mark*, unsigned ba0_int_p);

BA0_DLL void* ba0_alloc_but_do_not_set_magic (unsigned ba0_int_p n)
{   struct ba0_stack* H = ba0_stack_of_stacks [ba0_stack_sp];
    void* m;

#if defined (BA0_MEMCHECK)
    if (*H->bound != magic)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    n += BA0_ALIGN;
#endif
    n = ba0_ceil_align (n);

    m = ba0_alloc_but_do_not_set_magic_mark (&H->free, n);

    H->nb_calls_to_alloc++;
    ba0_alloc_counter += n;

    if (H->free.index_in_cells > H->max_alloc.index_in_cells ||
	(H->free.index_in_cells == H->max_alloc.index_in_cells &&
	 (unsigned ba0_int_p)H->free.address > 
				(unsigned ba0_int_p)H->max_alloc.address))
	H->max_alloc = H->free;
#if defined (BA0_MEMCHECK)
    H->bound = (ba0_int_p*)((unsigned ba0_int_p)m + n - BA0_ALIGN);
#endif
    return m;
}

/*
   Subfunction of ba0_alloc
*/

BA0_DLL void ba0_alloc_set_magic (void)
{
#if defined (BA0_MEMCHECK)
    struct ba0_stack* H = ba0_stack_of_stacks [ba0_stack_sp];
    *H->bound = magic;
#endif
}

/*
   Allocates n bytes (+ magic number) in the current stack.
*/

BA0_DLL void* ba0_alloc (unsigned ba0_int_p n)
{   void* p;

    p = ba0_alloc_but_do_not_set_magic (n);
    ba0_alloc_set_magic ();
    return p;
}

/*
   Specific array allocation for polynomials.
   If the array does not fit in the current cell, then its size is decreased.
*/

BA0_DLL void ba0_t2_alloc 
	(unsigned ba0_int_p sc, unsigned ba0_int_p st, unsigned ba0_int_p n, 
	 void** tc, void** tt, unsigned ba0_int_p* m)
{   struct ba0_stack* H = ba0_stack_of_stacks [ba0_stack_sp];

    if (n == 0 || sc == 0 || st == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (sc < st)
    {	BA0_SWAP (ba0_int_p, sc, st);
	BA0_SWAP (void**, tc, tt);
    }
    if (2 * n * sc <= H->free.ce_qui_reste)
	*m = n;
    else if (H->free.ce_qui_reste > H->cell_size / 128)
	*m = H->free.ce_qui_reste / (2 * sc);
    else
    {	*m = H->cell_size / (2 * sc);
	*m = BA0_MIN (n, *m);
    }

    *tc = ba0_alloc (*m * sc);
    *tt = ba0_alloc (*m * st);
}

/*
   Variant of the above one.
*/

BA0_DLL void ba0_t1_alloc (unsigned ba0_int_p se, unsigned ba0_int_p n, 
				 void** te, unsigned ba0_int_p* m)
{   struct ba0_stack* H = ba0_stack_of_stacks [ba0_stack_sp];

    if (n == 0 || se == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (n * se <= H->free.ce_qui_reste)
	*m = n;
    else if (H->free.ce_qui_reste > H->cell_size / 128)
	*m = H->free.ce_qui_reste / se;
    else
    {	*m = H->cell_size / se;
	*m = BA0_MIN (n, *m);
    }

    *te = ba0_alloc (*m * se);
}

/**********************************************************************
 THE STACK OF STACKS
 **********************************************************************/

BA0_DLL void ba0_init_stack_of_stacks (void)
{
    ba0_stack_of_stacks = (struct ba0_stack**)ba0_malloc 
		(sizeof (struct ba0_stack*) * ba0_sizeof_stack_of_stack);
    ba0_reset_stack_of_stacks ();
}

BA0_DLL void ba0_reset_stack_of_stacks (void)
{
    ba0_stack_sp = 0;
    ba0_stack_of_stacks [ba0_stack_sp] = &ba0_main_stack;
}

BA0_DLL void ba0_clear_stack_of_stacks (void)
{
    ba0_free (ba0_stack_of_stacks);
    ba0_malloc_counter -= 
		sizeof (struct ba0_stack*) * ba0_sizeof_stack_of_stack;
}

/*
   After this call, H becomes the current stack.
*/

BA0_DLL void ba0_push_stack (struct ba0_stack* H)
{
    ba0_stack_sp++;
    if (ba0_stack_sp >= BA0_SIZE_STACK_OF_STACK)
        BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_stack_of_stacks [ba0_stack_sp] = H;
}

BA0_DLL void ba0_pull_stack (void)
{
    ba0_stack_sp--;
    if (ba0_stack_sp < 0)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
}

/*
   Changes the current stack.
*/

BA0_DLL void ba0_push_another_stack (void)
{
    if (ba0_stack_of_stacks [ba0_stack_sp] != &ba0_main_stack)
        ba0_push_stack (&ba0_main_stack);
    else
	ba0_push_stack (&ba0_second_stack);
}

/**********************************************************************
   MARKS
 **********************************************************************/

/*
   A mark is essentially a pointer in a stack.
*/

static void ba0_set_mark (struct ba0_mark* M, struct ba0_stack* H, 
			ba0_int_p index_in_cells, void* address, 
			unsigned ba0_int_p ce_qui_reste)
{
    M->stack = H;
    M->index_in_cells = index_in_cells;
    M->address = address;
    M->ce_qui_reste = ce_qui_reste;
}

/*
   Returns the index in M->stack->cells of the cell which involves p.
   Searches between the beginning and the address M points to.
   Returns -1 if not found
*/

BA0_DLL ba0_int_p ba0_cell_index_mark (void* p, struct ba0_mark* M)
{   struct ba0_stack* H = M->stack;
    unsigned ba0_int_p deb, fin;
    ba0_int_p i;

    for (i = 0; i < M->index_in_cells; i++)
    {	deb = (unsigned ba0_int_p)H->cells.tab [i];
	fin = deb + H->cell_size;
	if ((unsigned ba0_int_p)p >= deb && (unsigned ba0_int_p)p < fin)
	    return i;
    }
    deb = (unsigned ba0_int_p)H->cells.tab [M->index_in_cells];
    fin = (unsigned ba0_int_p)M->address;
    if ((unsigned ba0_int_p)p >= deb && (unsigned ba0_int_p)p < fin)
	return i;
    return -1;
}

/*
   The current cell M is pointing to is full.
   Moves M to the beginning of the next one.
   Quite often, M is actually the free pointer of the stack it is pointing to.
*/

static void ba0_move_to_next_cell_mark (struct ba0_mark* M)
{   struct ba0_stack* H = M->stack;

    if (M->index_in_cells + 1 == H->cells.size)
    {	if (H->cells.size == H->cells.alloc)
	    BA0_RAISE_EXCEPTION (BA0_ERROOM);
	H->cells.tab [H->cells.size] = ba0_malloc (H->cell_size);
	H->cells.size++;
    }
    M->index_in_cells++;
    M->address =  H->cells.tab [M->index_in_cells];
    M->ce_qui_reste = H->cell_size;
}

/*
   Simulates an allocation of n bytes, assuming M is the free pointer.
   The number n is assumed to be already aligned.
*/

static void* ba0_alloc_but_do_not_set_magic_mark
				(struct ba0_mark* M, unsigned ba0_int_p n)
{   void* m;

    m = M->address;
    if (n <= M->ce_qui_reste)
    {	M->address = (void*)((unsigned ba0_int_p)m + n);
	M->ce_qui_reste -= n;
    } else if (n >= M->stack->cell_size)
	BA0_RAISE_EXCEPTION (BA0_ERROOM);
    else 
    {
	ba0_move_to_next_cell_mark (M);
	m = M->address;
	M->address = (void*)((unsigned ba0_int_p)m + n);
	M->ce_qui_reste -= n;
    }
    return m;
}

/*
   Simulates a memory allocation, assuming that M is the free pointer
   Manages the magic number allocation and the alignment.
*/

BA0_DLL void* ba0_alloc_mark (struct ba0_mark* M, unsigned ba0_int_p n)
{   void* m;
    n = ba0_allocated_size (n);
    m = ba0_alloc_but_do_not_set_magic_mark (M, n);
    return m;
}
 
/*
   Returns true if ba0_alloc (n) would need to allocate a new cell
   through ba0_malloc.
*/

BA0_DLL bool ba0_alloc_would_malloc_a_cell (unsigned ba0_int_p n)
{   struct ba0_stack* H = ba0_stack_of_stacks [ba0_stack_sp];
    n = ba0_allocated_size (n);
    return n > H->free.ce_qui_reste &&
                        H->free.index_in_cells + 1 == H->cells.size;
}

/*
   Performs a rotation between some cells of the current stack :

   Let n = cells.size - 1

            -----------------------------------------
   Before : | 0 | 1 | ... | i-1 | i | i+1 | ... | n |
            -----------------------------------------

            -----------------------------------------
   After  : | 0 | 1 | ... | i-1 | i+1 | ... | n | i |
            -----------------------------------------

   Use with care. Marks pointing to the current stack may become corrupted.
   Actually used by the lexical analyzer which makes use of fifo.
*/

BA0_DLL void ba0_rotate_cells (ba0_int_p i)
{   struct ba0_stack* H = ba0_stack_of_stacks [ba0_stack_sp];
    void* cell;

    if (H->cells.size <= i)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);

    cell = H->cells.tab [i];
    memmove (H->cells.tab + i, H->cells.tab + i + 1,
                        (H->cells.size - i - 1) * sizeof (void*));
    H->cells.tab [H->cells.size - 1] = cell;
    if (H->free.index_in_cells == i)
        H->free.index_in_cells = H->cells.size - 1;
    else if (H->free.index_in_cells > i)
        H->free.index_in_cells -= 1;
    if (H->max_alloc.index_in_cells == i)
        H->max_alloc.index_in_cells = H->cells.size - 1;
    else
        H->max_alloc.index_in_cells -= 1;
}

/*
   Records the free pointer of the current stack in M.
*/

BA0_DLL void ba0_record (struct ba0_mark* M)
{   struct ba0_stack* H = ba0_stack_of_stacks [ba0_stack_sp];

    ba0_process_check_interrupt ();
    *M = H->free;
}

/*
   Restores the free pointer of the stack referred to by M to the
   address stored in M. This function should follow a call to ba0_record.
*/

BA0_DLL void ba0_restore (struct ba0_mark* M)
{   struct ba0_stack* H = M->stack;

    if (M->index_in_cells > H->free.index_in_cells ||
	(M->index_in_cells == H->free.index_in_cells &&
	 (unsigned ba0_int_p)M->address > (unsigned ba0_int_p)H->free.address))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    H->free = *M;
}

/*
 * Returns the amount of memory between the two marks A and B.
 */

BA0_DLL unsigned ba0_int_p 
		ba0_range_mark (struct ba0_mark* A, struct ba0_mark* B)
{   struct ba0_mark M;
    unsigned ba0_int_p range;

    if (A->stack != B->stack || A->index_in_cells > B->index_in_cells)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    M = *A;
    range = 0;
    while (M.index_in_cells < B->index_in_cells)
    {	range += M.ce_qui_reste;
	ba0_move_to_next_cell_mark (&M);
    }
    range += (M.ce_qui_reste - B->ce_qui_reste);
    return range;
}

