#include "bad_reduction.h"
#include "bad_invert.h"
#include "bad_global.h"

/*
 * A0 and B0 are polynomials with leading derivative x.
 * The initial of B0 is nonzero mod C
 *
 * Computes U1, U2 and U3 such that U1 * A0 + U2 * B0 = U3 mod C
 *                                  U3 is a gcd (mod C) of A0 and B0
 *
 * (except that U2 is not computed and not returned).
 *
 * The ordering is assumed to be rearranged so that L >> N
 * where L = Leaders (C) and N = Non-Leaders (C, A0, B0)
 */

static void bad_half_Euclid 
	(bap_polynom_mpz U1, bap_polynom_mpz U3,
	 bap_polynom_mpz A0, bap_polynom_mpz B0,
	 bav_variable x, bad_regchain C)
{   struct bap_tableof_polynom_mpz polys;
    bap_product_mpz H;
    bap_polynom_mpz U_1, U_3, V_1, V_3, Q, R, S;
    ba0_int_p counter;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    U_1 = bap_new_polynom_mpz ();
    V_1 = bap_new_polynom_mpz ();

    if (bap_degree_polynom_mpz (A0, x) >= bap_degree_polynom_mpz (B0, x))
    {	bap_set_polynom_one_mpz (U_1);
	U_3 = A0;
	V_3 = B0;
    } else
    {	bap_set_polynom_one_mpz (V_1);
	V_3 = A0;
	U_3 = B0;
    }

    Q = bap_new_polynom_mpz ();
    R = bap_new_polynom_mpz ();
    S = bap_new_polynom_mpz ();
    H = bap_new_product_mpz ();
/*
 * No need to realloc2
 */
    ba0_init_table ((ba0_table)&polys);
    ba0_realloc_table ((ba0_table)&polys, 2);

    counter = 0;
/*
 * Loop invariants:
 * U_1 * A0 + U_2 * B0 = U_3 mod C
 * V_1 * A0 + V_2 * B0 = V_3 mod C
 */
    while (! bap_is_zero_polynom_mpz (V_3))
    {   
	bap_gcd_pseudo_division_polynom_mpz (Q, R, H, U_3, V_3, x);
	bad_reduce_easy_polynom_by_regchain (R, R, C, bad_algebraic_reduction);
	bad_reduce_easy_polynom_by_regchain (Q, Q, C, bad_algebraic_reduction);
	bad_ensure_initial_mod_regchain (R, R, C, bad_algebraic_reduction);
/*
 * H * U_3 = Q * V_3 + R mod C
 */
	bap_mul_polynom_mpz (Q, Q, V_1);
	bap_mul_product_polynom_mpz (H, H, U_1, 1);
	bap_expand_product_mpz (S, H);
	bap_sub_polynom_mpz (S, S, Q);
/*
 * S = H * U_1 - Q * V_1 
 */
	polys.tab [0] = S;
	polys.tab [1] = R;
	polys.size = 2;
	bap_content_tableof_polynom_mpz (H, &polys, x, true);
	bap_expand_product_mpz (Q, H);
	if (! bap_is_one_polynom_mpz (Q))
	{   bap_exquo_polynom_mpz (R, R, Q);
	    bap_exquo_polynom_mpz (S, S, Q);
	}
/*
 * Q = the gcd of the coefficients (wrt x) of S and R together
 * S = S / Q
 * R = R / Q
 */
	BA0_SWAP (bap_polynom_mpz, U_1, V_1);
	if (counter < 2)
	{   U_3 = V_3;
	    V_3 = bap_new_polynom_mpz ();
	} else 
	    BA0_SWAP (bap_polynom_mpz, U_3, V_3);
/*
 * A bit complicated stuff in order not to override A0 and B0
 */
	BA0_SWAP (bap_polynom_mpz, V_1, S);
	BA0_SWAP (bap_polynom_mpz, V_3, R);
/*
 * U_1 = V_1
 * U_3 = V_3
 * V_1 = S
 * V_3 = R
 */
	counter += 1;
    }
    ba0_pull_stack ();
    if (U1 != U_1)
	bap_set_polynom_mpz (U1, U_1);
    if (U3 != U_3)
	bap_set_polynom_mpz (U3, U_3);
    ba0_restore (&M);
}

/*
 * Computes an inverse of A modulo C, i.e products U and G such that
 *
 *             U * A = G mod C (algebraically)
 *
 *             U in K [L, N] 
 *             G in K [N]
 *
 * By a well-known theorem, G is then not a zero divisor and 1/A = U/G mod C
 *
 * Any exhibited zerodivisor is returned through ddz.
 * The case A = 0 is handled.
 *
 * The ordering is assumed to be rearranged so that L >> N
 * where L = Leaders (C) and N = Non-Leaders (C, A)
 */

static void bad_algebraic_invert_polynom_mod_regchain_zero_dimensional
                (bap_product_mpz U, bap_polynom_mpz G, bap_polynom_mpz A,
                 bad_regchain C, bap_polynom_mpz* ddz)
{   struct bad_regchain Cbar;
    bap_product_mpz U1;
    bap_polynom_mpz U1bar, U3;
    bav_variable x;
    ba0_int_p k;
    struct ba0_mark M;
/*
 * Cbar is a working copy of C.
 */
    Cbar = *C;

    ba0_push_another_stack ();
    ba0_record (&M);

    U1 = bap_new_product_mpz ();
    U3 = bap_new_polynom_mpz ();
    U1bar = bap_new_polynom_mpz ();

    bad_ensure_initial_mod_regchain (U3, A, C, bad_algebraic_reduction);
    if (bap_is_zero_polynom_mpz (U3))
	BA0_RAISE_EXCEPTION (BAD_EXRNUL);
/*
 * Loop invariant: U1 * A = U3 mod C
 */
    for (;;)
    {	
/*
 * Using bad_nonzero would change the behaviour of the function, in the
 * case of computations over base fields presented by non-trivial relations
 *
 * See tests/base_field2.c
 */
	if (bap_is_independent_polynom_mpz (U3))
	    break;
	x = bap_leader_polynom_mpz (U3);
	if (! bad_is_leader_of_regchain (x, C, &k))
	    break;
/*
 * Cbar = { C[0], ..., C[k-1] }. 
 * Reductions by Cbar will only apply to the coefficients of the polynomials.
 * Otherwise, the reduce_easy operation creates problems.
 */
	Cbar.decision_system.size = k;
	bad_half_Euclid (U1bar, U3, U3, C->decision_system.tab [k], x, &Cbar);
	if (bap_depend_polynom_mpz (U3, x))
        {   if (ddz != (bap_polynom_mpz*)0)
            {   *ddz = U3;
                BA0_RAISE_EXCEPTION2 (BAD_EXRDDZ, "%Az", (void**)ddz);
            } else
                BA0_RAISE_EXCEPTION (BAD_EXRDDZ);
        }
	bap_mul_product_polynom_mpz (U1, U1, U1bar, 1);
    }
    if (bap_is_zero_polynom_mpz (U3))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    ba0_pull_stack ();
    bap_set_product_mpz (U, U1);
    bap_set_polynom_mpz (G, U3);
    ba0_restore (&M);
}

/*
 * Computes an inverse of A modulo C, i.e products U and G such that
 *
 *             U * A = G mod C (algebraically)
 *
 *             U in K [L, N] 
 *             G in K [N]
 *
 * By a well-known theorem, G is then not a zero divisor and 1/A = U/G mod C
 *
 * Any exhibited zerodivisor is returned through ddz.
 * The case A = 0 is handled.
 *
 * This function performs a temporary change of ordering.
 */

static void bad_algebraic_invert_product_mod_regchain
                (bap_product_mpz U, bap_product_mpz G, bap_product_mpz A,
                 bad_regchain C, bap_polynom_mpz* ddz)
{
    struct bad_regchain Cbar;
    struct bap_product_mpz Abar, Ubar, Gbar, U1;
    struct bap_polynom_mpz U3;
    bap_polynom_mpz ddzbar, *_ddzbar;
    bav_Iordering r;
    volatile ba0_int_p i;
    struct ba0_exception_code code;
    struct ba0_mark M;
/*
 * Get rid of the case A is zero
 */
    if (bap_is_zero_product_mpz (A))
        BA0_RAISE_EXCEPTION (BAD_EXRNUL);

    ba0_push_another_stack ();
    ba0_record (&M);

    r = bad_ordering_eliminating_leaders_of_regchain (C);
    bav_R_push_ordering (r);
/*
 * It is better to remove the differential property since r is not a ranking
 * Moreover, one may want to compute algebraic inverses of non partially
 * reduced polynomials, modulo regular differential chains.
 */
    bad_init_regchain (&Cbar);
    bad_sort_regchain (&Cbar, C);
    bad_clear_structural_property_attchain 
			(&Cbar.attrib, bad_differential_structural_property);

    _ddzbar = ddz == (bap_polynom_mpz*)0 ? (bap_polynom_mpz*)0 : &ddzbar;

    bap_init_product_mpz (&Abar);
    bap_sort_product_mpz (&Abar, A);

    bap_init_product_mpz (&U1);
    bap_init_polynom_mpz (&U3);

    bap_init_product_mpz (&Ubar);
    bap_init_product_mpz (&Gbar);
    bap_set_product_numeric_mpz (&Gbar, Abar.num_factor);
/*
 * Denote Abar = c * mul (Abar[k] ** d[k], k = 1 .. t)
 * Loop invariant: 
 *
 *      Ubar * Abar = Gbar * mul (Abar[k] ** d[k], k = i .. t) mod C
 *
 * Thus, initially Ubar = 1 and Gbar = c
 */
    for (i = 0; i < Abar.size; i++)
    {
	if (Abar.tab [i].exponent > 0)
	{
/*
 * The exception handling is needed for fixing the ordering of ddz
 */
	    BA0_PUSH_EXCEPTION (code);
	    if (ba0_exception_is_set (code))
	    {   bad_algebraic_invert_polynom_mod_regchain_zero_dimensional 
			(&U1, &U3, &Abar.tab [i].factor, &Cbar, _ddzbar);
	        ba0_pull_exception (code);
	    } else
	    {   
/*
 * The ordering r will be freed by the next BA0_RAISE_EXCEPTION
 */
	        if (ba0_mesgerr != BAD_EXRDDZ || ddz == (bap_polynom_mpz*)0)
		    BA0_RAISE_EXCEPTION (ba0_mesgerr);
/*
 * Change of ranking over the exhibited zero divisor
 */
	        bav_R_pull_ordering ();
	        bap_sort_polynom_mpz (ddzbar, ddzbar);
	        *ddz = bap_new_polynom_mpz ();
	        bap_set_polynom_mpz (*ddz, ddzbar);
	        BA0_RAISE_EXCEPTION2 (BAD_EXRDDZ, "%Az", (void**)ddz);
	    }
/*
 * We have computed U1 and U3 such that
 *
 *      U1 * Abar[i] = U3 mod C
 *
 * Thus
 *      Ubar = Ubar * U1 ** d[i]
 *      Gbar = Gbar * U3 ** d[i]
 */
	    bap_pow_product_mpz (&U1, &U1, Abar.tab [i].exponent);
	    bap_mul_product_mpz (&Ubar, &Ubar, &U1);
	    bap_mul_product_polynom_mpz 
				(&Gbar, &Gbar, &U3, Abar.tab [i].exponent);
	}
    }

    bav_R_pull_ordering ();
    bap_sort_product_mpz (&Ubar, &Ubar);
    bap_sort_product_mpz (&Gbar, &Gbar);
    bav_R_free_ordering (r);

    ba0_pull_stack ();
    bap_set_product_mpz (U, &Ubar);
    bap_set_product_mpz (G, &Gbar);
    ba0_restore (&M);
}

/*
 * Computes an inverse of A modulo C, i.e products U and G such that
 *
 *             U * A = G mod C (differentially)
 *
 *             U in K [L, N] 
 *             G in K [N]
 *
 * By a well-known theorem, G is then not a zero divisor and 1/A = U/G mod C
 *
 * Any exhibited zerodivisor is returned through ddz.
 * The case A = 0 is handled.
 *
 * Here, L = leaders (C) and N = non-leaders (C, A)
 */

static void bad_differential_invert_polynom_mod_regchain
                (bap_product_mpz U, bap_product_mpz G, bap_polynom_mpz A,
                 bad_regchain C, bap_polynom_mpz* ddz)
{
    struct bap_product_mpz P, H;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_product_mpz (&H);
    bap_init_product_mpz (&P);

    bad_reduce_polynom_by_regchain 
	(&P, &H, A, C, bad_full_reduction, bad_all_derivatives_to_reduce);

    ba0_pull_stack ();

    bad_algebraic_invert_product_mod_regchain (U, G, &P, C, ddz);
    bap_mul_product_mpz (U, U, &H);

    ba0_restore (&M);
}

/*
 * Computes an inverse of A modulo C, i.e products U and G such that
 *
 *             U * A = G mod C (differentially or algebraically)
 *
 *             U in K [L, N] 
 *             G in K [N]
 *
 * The choice differential vs algebraic depends on the attributes of C
 *
 * By a well-known theorem, G is then not a zero divisor and 1/A = U/G mod C
 *
 * Any exhibited zerodivisor is returned through ddz.
 * The case A = 0 is handled.
 * 
 * Here, L = leaders (C) and N = non-leaders (C, A)
 */

BAD_DLL void bad_invert_polynom_mod_regchain
                (bap_product_mpz U, bap_product_mpz G, bap_polynom_mpz A,
                 bad_regchain C, bap_polynom_mpz* ddz)
{   struct bap_product_mpz P;
    struct ba0_mark M;

    if (bad_defines_a_differential_ideal_regchain (C))
	bad_differential_invert_polynom_mod_regchain (U, G, A, C, ddz);
    else
    {	if (bad_is_a_reduced_to_zero_polynom_by_regchain 
					(A, C, bad_algebraic_reduction))
	    BA0_RAISE_EXCEPTION (BAD_EXRNUL);

	ba0_push_another_stack ();
	ba0_record (&M);
	bap_init_product_mpz (&P);
	bap_factor_easy_polynom_mpz (&P, A, (bap_listof_polynom_mpz)0);
	ba0_pull_stack ();
	bad_algebraic_invert_product_mod_regchain (U, G, &P, C, ddz);
	ba0_restore (&M);
    }
    bap_gcd_product_mpz ((bap_product_mpz)0, U, G, U, G);
}

BAD_DLL void bad_invert_product_mod_regchain
                (bap_product_mpz U, bap_product_mpz G,
                 bap_product_mpz A, bad_regchain C, bap_polynom_mpz* ddz)
{   struct bap_product_mpz Uhat, Ghat, Ubar, Gbar;
    ba0_int_p i;
    struct ba0_mark M;

    if (bap_is_zero_product_mpz (A))
        BA0_RAISE_EXCEPTION (BAD_EXRNUL);

    bap_set_product_one_mpz (U);
    bap_set_product_numeric_mpz (G, A->num_factor);
    bap_realloc_product_mpz (U, A->size);
    bap_realloc_product_mpz (G, A->size);

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_product_mpz (&Ubar);
    bap_init_product_mpz (&Gbar);
    bap_set_product_numeric_mpz (&Gbar, A->num_factor);
    
    bap_init_product_mpz (&Uhat);
    bap_init_product_mpz (&Ghat);
/*
 * Denote A = c * mul (A[k] ** d[k], k = 1 .. t)
 * Loop invariant: 
 *
 *      Ubar * A = Gbar * mul (A[k] ** d[k], k = i .. t) mod C
 *
 * Thus, initially Ubar = 1 and Gbar = c
 */
    for (i = 0; i < A->size; i++)
    {   if (A->tab [i].exponent > 0)
        {   bad_invert_polynom_mod_regchain
                                (&Uhat, &Ghat, &A->tab [i].factor, C, ddz);
            bap_pow_product_mpz (&Uhat, &Uhat, A->tab [i].exponent);
            bap_pow_product_mpz (&Ghat, &Ghat, A->tab [i].exponent);

            bap_mul_product_mpz (&Ubar, &Ubar, &Uhat);
            bap_mul_product_mpz (&Gbar, &Gbar, &Ghat);
        }
    }

    ba0_pull_stack ();
    bap_set_product_mpz (U, &Ubar);
    bap_set_product_mpz (G, &Gbar);
    ba0_restore (&M);
}

