#ifndef BAI_DOP853_H
#define BAI_DOP853_H 1

/**********************************************************************
   Adaptation of the FORTRAN code provided by Ernst Hairer for DOP853.
   Page numbers refer to 

   Solving Ordinary Differential Equations I
   Ernst Hairer, Syvert Norsett, Gerhard Wanner
   Springer Verlag
   Second Revised Edition

   The Runge-Kutta method has order 8.
   There are two different embedded methods of respective orders 5 and 3
   in order to increase stability w.r.t. the "basic" 8(6) method. See page 254.

   The dense output method has order 7. See page 194.

   ODE y'(t) = f (t, y(t))

   The signature of the functions to be integrated.
   t = the independent variable
   y = the current value of the dependent variables y(t)
   f = the values to be computed f (t, y(t))
   params = whatever you want as extra parameters to provide to the function

   The function returns 0 on success and any nonzero value on failure.
 **********************************************************************/

#include "bai_common.h"
#include "bai_odex.h"
#include "bai_odex_integrated_function.h"

BEGIN_C_DECLS

/**********************************************************************
 INITIAL VALUES

 There should be one y0 [i] per dependent variable.
 The ODE is integrated from t0 to t1. t0 > t1 is allowed.
 **********************************************************************/

typedef struct bai_dop853_initial_values {
    double t0;
    double t1;
    struct ba0_arrayof_double y0;
} *bai_dop853_initial_values;

extern BAI_DLL void bai_dop853_init_initial_values (bai_dop853_initial_values);
extern BAI_DLL void bai_dop853_reset_initial_values 
			(bai_dop853_initial_values);
extern BAI_DLL bai_dop853_initial_values bai_dop853_new_initial_values (void);
extern BAI_DLL void bai_dop853_realloc_initial_values 
			(bai_dop853_initial_values, ba0_int_p);
extern BAI_DLL void bai_dop853_set_initial_values_time 
			(bai_dop853_initial_values, double, double);
extern BAI_DLL void bai_dop853_set_initial_values_variable
			(bai_dop853_initial_values, bav_variable, 
			 bai_odex_system, double);
extern BAI_DLL void bai_dop853_set_initial_values
			(bai_dop853_initial_values, bai_dop853_initial_values);

/**********************************************************************
   SETTING

   Parameters to control the behaviour of the integrator

   uround = the rounding unit.
   safe_fac, fac1 and fac2 are safety factors to forbid the step
	to increase or decrease too fast.
   beta = a number related to safety factors (stabilization)
   hmax	= the max value of the step h (absolute value).
   h0   = the initial value of the step h (signed !). 
	In the case of a zero value, the value is determined by computation. 
   nb_max_steps = the maximal number of steps
   stiffness_test_step = stiffness tests are performed at each step
	of the form k * stiffness_test_step, for all positive integer k.

   Some default values can be provided using the initial values 

   uround = 2.3e-16
   safe_fac = 0.9
   fac1 = 1./3.
   fac2 = 6.
   beta = 0.
   hmax = |t1 - t0|
   h0 = determined by computation
   nb_max_steps = 100000
   stiffness_test_step = 1000

 **********************************************************************/

typedef struct bai_dop853_setting {
    double uround;
    double safe_fac;
    double fac1;
    double fac2;
    double beta;
    double hmax;
    double h0;
    ba0_int_p nb_max_steps;
    ba0_int_p stiffness_test_step;
} *bai_dop853_setting;

extern BAI_DLL void bai_dop853_init_setting (bai_dop853_setting);
extern BAI_DLL void bai_dop853_reset_setting (bai_dop853_setting);
extern BAI_DLL void bai_dop853_set_setting 
				(bai_dop853_setting, bai_dop853_setting);
extern BAI_DLL void bai_dop853_set_default_setting (bai_dop853_setting);

/**********************************************************************
 STATISTICS

 They are inlined now in the bai_dop853_workspace structure
 **********************************************************************/

typedef struct bai_dop853_stats {
    ba0_int_p nb_evals;	/* number of calls to the function to be integrated */
    ba0_int_p nb_steps;	/* number of steps */
    ba0_int_p nb_accepts;	/* number of accepted steps */
    ba0_int_p nb_rejects;	/* number of rejected steps */
} *bai_dop853_stats;

extern BAI_DLL void bai_dop853_init_stats (bai_dop853_stats);
extern BAI_DLL void bai_dop853_reset_stats (bai_dop853_stats);
extern BAI_DLL void bai_dop853_set_stats (bai_dop853_stats, bai_dop853_stats);

/**********************************************************************
 ERROR TOLERANCE

 There are two types of error tolerance.
 bai_dop853_global_errtol: the entries relative [0] and absolute [0] are used
 bai_dop853_componentwise_errtol: the entries relative [i] and absolute [i]
	are used for each i (there should be an entry per dependent variable).
 **********************************************************************/

enum bai_dop853_typeof_errtol {
		 bai_dop853_global_errtol, bai_dop853_componentwise_errtol };

typedef struct bai_dop853_errtol {
    enum bai_dop853_typeof_errtol type;
    struct ba0_arrayof_double relative;
    struct ba0_arrayof_double absolute;
} *bai_dop853_errtol;

/*
   Initializes to bai_dop853_global_errtol with a absolute [0] = 1e-6
*/

extern BAI_DLL void bai_dop853_init_errtol (bai_dop853_errtol);
extern BAI_DLL void bai_dop853_reset_errtol (bai_dop853_errtol);
extern BAI_DLL void bai_dop853_set_default_errtol (bai_dop853_errtol);
extern BAI_DLL void bai_dop853_set_errtol 
				(bai_dop853_errtol, bai_dop853_errtol);
extern BAI_DLL void bai_dop853_realloc_errtol (bai_dop853_errtol, ba0_int_p);

/**********************************************************************
   VIEW

   What should be displayed

   view_t indicates if the independent variable must be displayed.
   Each entry of tab provides a variable or, more generally, a value
	to be displayed.
   If isvar is true then the value of the variable number varnum must be
	displayed. If isvar is false, then the *eval function is called
	and its returned value is displayed.
 **********************************************************************/

typedef double bai_dop853_eval_function (double t, void* workspace);

typedef struct bai_dop853_view_elt {
    bool isvar;
    ba0_int_p varnum;
    bai_dop853_eval_function *eval;
} *bai_dop853_view_elt;

typedef struct bai_arrayof_view_elt {
    ba0_int_p alloc;
    ba0_int_p size;
    struct bai_dop853_view_elt* tab;
    ba0_int_p sizelt;
} *bai_arrayof_view_elt;

typedef struct bai_dop853_view {
    bool view_t;
    struct bai_arrayof_view_elt elts;
} *bai_dop853_view;

extern BAI_DLL void bai_dop853_init_view (bai_dop853_view);
extern BAI_DLL void bai_dop853_reset_view (bai_dop853_view);
extern BAI_DLL void bai_dop853_realloc_view (bai_dop853_view, ba0_int_p);
extern BAI_DLL void bai_dop853_set_view_variable 
			(bai_dop853_view, bav_variable, bai_odex_system);
extern BAI_DLL void bai_dop853_set_view_function 
			(bai_dop853_view, bai_dop853_eval_function*);

/**********************************************************************
 SOLOUT

 Prints in file f the values in w of the variables given in view.
 The boolean indicates if some interpolation must be performed, i.e. if
 a curve must be printed instead of a single point.
 If the boolean is true then
 - the double provides the interpolation step.
 - the variables in view must be prepared for dense output.
 
 **********************************************************************/

struct bai_dop853_workspace;

extern BAI_DLL void bai_dop853_solout 
	(FILE*, struct bai_dop853_workspace*, bai_dop853_view, bool, double);

/**********************************************************************
 DENSE OUTPUT 

 Dense output permits to interpolate the values of some dependent variables
 between two integration steps. 

 The indices of the variables over which dense output is going to be
 processed should be stored in index.
 **********************************************************************/

typedef struct bai_dop853_dense_output {
    struct ba0_tableof_int_p index;
    struct ba0_arrayof_double cont0;
    struct ba0_arrayof_double cont1;
    struct ba0_arrayof_double cont2;
    struct ba0_arrayof_double cont3;
    struct ba0_arrayof_double cont4;
    struct ba0_arrayof_double cont5;
    struct ba0_arrayof_double cont6;
    struct ba0_arrayof_double cont7;
} *bai_dop853_dense_output;

extern BAI_DLL void bai_dop853_init_dense_output (bai_dop853_dense_output);
extern BAI_DLL void bai_dop853_reset_dense_output (bai_dop853_dense_output);
extern BAI_DLL void bai_dop853_realloc_dense_output
		(bai_dop853_dense_output, ba0_int_p);
extern BAI_DLL void bai_dop853_set_dense_output_all_variables
		(bai_dop853_dense_output, bai_odex_system);
extern BAI_DLL void bai_dop853_set_dense_output_variable
		(bai_dop853_dense_output, bav_variable, bai_odex_system);
extern BAI_DLL void bai_dop853_set_dense_output
		(bai_dop853_dense_output v, bai_dop853_dense_output w);
/*
   Assigns yi (t) to *res.
   The value of t should be in the range w->told .. w->t 
*/

extern BAI_DLL double bai_dop853_dense_output_evaluate
		(ba0_int_p i, double t, struct bai_dop853_workspace* w);
 
/**********************************************************************
 DOP 853
 **********************************************************************/

typedef struct bai_dop853_workspace {
    double told;
    double t;
    double h;
    struct ba0_arrayof_double y;
			/* y + h * a linear combination of ki (tmp. var.) */
    struct ba0_arrayof_double y1; 
			/* the famous Runge-Kutta numbers */
    struct ba0_arrayof_double k1; 
    struct ba0_arrayof_double k2;
    struct ba0_arrayof_double k3;
    struct ba0_arrayof_double k4;
    struct ba0_arrayof_double k5;
    struct ba0_arrayof_double k6;
    struct ba0_arrayof_double k7;
    struct ba0_arrayof_double k8;
    struct ba0_arrayof_double k9;
    struct ba0_arrayof_double k10;
/*
   Some internal variables of the core integrator
*/
    double hlamb;
    ba0_int_p nb_stiffs;
    ba0_int_p nb_non_stiffs;
    double facold;
    struct bai_dop853_stats stat;
/*
   Pointers to the parameters provided to the reset function
*/
    bai_odex_integrated_function* fcn;
    struct bai_dop853_initial_values iv;
    struct bai_dop853_errtol tol;
    struct bai_dop853_setting setting;
    struct bai_dop853_dense_output dow;
    void* params;
} *bai_dop853_workspace;

extern BAI_DLL void bai_dop853_init_workspace (bai_dop853_workspace);
extern BAI_DLL void bai_dop853_reset_workspace (bai_dop853_workspace);
extern BAI_DLL void bai_dop853_set_workspace 
			(bai_dop853_workspace, bai_dop853_workspace);

extern BAI_DLL void bai_dop853_start_workspace 
	(bai_dop853_workspace, 
	 bai_odex_integrated_function* fcn,
	 bai_dop853_initial_values iv,
	 bai_dop853_errtol tol,
	 bai_dop853_setting setting,
	 bai_dop853_dense_output dow,
	 void* params);

/*
   Performs one Runge-Kutta step. The main function.
*/

extern BAI_DLL void bai_dop853_step_workspace (struct bai_dop853_workspace*);

/**********************************************************************
 EVALUATION FUNCTION

 The function index i must be prepared for dense output (it must
 appear in the index array of w->dow). res is a pointer to one double.

 Assigns to yi (t) to *res. 
 **********************************************************************/

extern BAI_DLL double bai_dop853_evaluate_variable
	(bav_variable v, double t, bai_dop853_workspace w, bai_odex_system);

END_C_DECLS

#endif /* !BAI_DOP853_H */

