#include <blad.h>

#ifdef _MSC_VER
#    include <float.h>
#    define finite _finite
#endif

#ifdef HAVE_GSL_H

#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_odeiv.h>

static enum bai_odex_exit_code van_der_pol
	(double t, double* x, double* f, void* params)
{   double* p;
    bai_command_function** c;

    p = ((struct bai_params*)params)->pars.values;
    c = ((struct bai_params*)params)->cmds.fptrs;

    f[0] = x[1];
    if (! finite (f[0]))
        return bai_odex_non_finite;
    f[1] = -x[0]*x[0]*x[1]*p[0] - x[0] + x[1]*p[0];
    if (! finite (f[1]))
        return bai_odex_non_finite;
    return bai_odex_success;
}

static enum bai_odex_exit_code jac_vdp
	(double t, double* x, double* J, double* dfdt, void* params)
{   double* p;
    bai_command_function** c;

    p = ((struct bai_params*)params)->pars.values;
    c = ((struct bai_params*)params)->cmds.fptrs;

    J[0] = (double)0;
    if (! finite (J[0]))
        return bai_odex_non_finite;
    J[1] = (double)1;
    if (! finite (J[1]))
        return bai_odex_non_finite;
    J[2] = -2*x[0]*x[1]*p[0] - 1;
    if (! finite (J[2]))
        return bai_odex_non_finite;
    J[3] = -x[0]*x[0]*p[0] + p[0];
    if (! finite (J[3]))
        return bai_odex_non_finite;

    dfdt[0] = (double)0;
    if (! finite (dfdt[0]))
        return bai_odex_non_finite;
    dfdt[1] = (double)0;
    if (! finite (dfdt[1]))
        return bai_odex_non_finite;

    return bai_odex_success;
}

typedef int fcn_type (double, const double[], double[], void*);
typedef int jac_type (double, const double[], double*, double[], void*);

int main ()
{   struct bai_odex_system S;
    struct bai_params params;
/*
    void* handle;
    FILE* f;
*/
    fcn_type* fcn;
    jac_type* jac;    
    bav_Iordering r;
    bool b;

    bai_restart (0, 0);
    ba0_sscanf2 ("ordering (derivations = [t], blocks = [[y0, y1], epsilon])",
		 "%ordering", &r);
    bav_R_push_ordering (r);
/*
 * Definition of the system (the Van Der Pol oscillator)
 */ 
    bai_init_odex_system (&S);
    ba0_sscanf2 
	("odex (params = [epsilon], \
		eqns = [y0[t] = y1, y1[t] = epsilon*(1 - y0^2)*y1 - y0])", 
	 "%odex", &S);
/*
 * C code generation - with the Jacobian
 * 
 *   f = fopen ("mm.c", "w");
 *   if (f == NULL) return 1;
 *   bai_odex_generate_rhs_C_code (f, "van_der_pol", &S);
 *   bai_odex_generate_jacobianof_rhs_C_code (f, "jac_vdp", &S);
 *   fclose (f);
 * 
 * Dynamic compiling and linking on a LINUX machine
 * fcn and jac receive the addresses of the two functions
 * 
 *   system ("gcc -fPIC -shared mm.c -o mm.so");
 *   handle = dlopen ("./mm.so", RTLD_NOW);
 *   if (! handle)
 *   {   fprintf (stderr, "%s\n", dlerror());
 *       exit (1);
 *   }
 *   fcn = (fcn_type*)dlsym (handle, "van_der_pol");
 *   if (! fcn)
 *   {   fprintf (stderr, "%s\n", dlerror());
 *       exit (1);
 *   }
 *   jac = (jac_type*)dlsym (handle, "jac_vdp");
 *   if (! jac)
 *   {   fprintf (stderr, "%s\n", dlerror());
 *       exit (1);
 *   }
 */
    fcn = (fcn_type*)van_der_pol;
    jac = (jac_type*)jac_vdp;
/*
 * Set epsilon = 10.0
 * Observe that this setting is different from that given in the GSL manual
 * for the functions *fcn and *jac, which were generated by BLAD, expect
 * a special form for their last void* parameter.
 */
    bai_init_params (&params);
    bai_set_params_odex_system (&params, &S);
    bai_set_params_parameter (&params, S.params.tab [0], 10.0);
/*
 * This part is borrowed from the GSL manual.
 * The numerical integration method is that of Stoer and Bulirsch, which
 * requires the Jacobian.
 */
    {	const gsl_odeiv_step_type* T = gsl_odeiv_step_bsimp;
	gsl_odeiv_step* s = gsl_odeiv_step_alloc (T, 2);
	gsl_odeiv_control* c = gsl_odeiv_control_y_new (1e-6, 0.0);
	gsl_odeiv_evolve* e = gsl_odeiv_evolve_alloc (2);
/*
 * 	gsl_odeiv_system sys = { fcn, jac, 2, &params };
 */
	gsl_odeiv_system sys;

	double t = 0.0;
	double t1 = 100.0;
	double h = 1e-6;
	double y[2] = { 1.0, 0.0 };

	sys.function = fcn;
	sys.jacobian = jac;
	sys.dimension = 2;
	sys.params = &params;

	while (t < t1)
	{   int status = gsl_odeiv_evolve_apply (e, c, s, &sys, &t, t1, &h, y);
	    if (status != GSL_SUCCESS)
		break;
/*
	    printf ("%.5e %.5e\n", t, y[0]);
*/
	}

	b = fabs (y[0] + 1.75889) < 1e-3;

	gsl_odeiv_evolve_free (e);
	gsl_odeiv_control_free (c);
	gsl_odeiv_step_free (s);
    }

    bai_terminate (ba0_init_level);
    return b ? 0 : 1;
}

#else

int main ()
{
    return 0;
}

#endif

