/*
    Copyright (C) 2002 Francois Boulier
    University Lille I. LIFL. 59655 Villeneuve d'Ascq CEDEX. France.
    email: boulier@lifl.fr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/


#include "bap_clot_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/*
 * Clots are ordered linear combinations of terms.
 * Polynomials are built over them.
 * 
 * The field {\tt alloc} gives the number of monomials allocated to the clot.
 * The field {\tt size} gives the number of monomials used by the clot.
 * The zero clot is coded by $\mbox{\tt size} = 0$.
 * The field {\tt ordering} gives the ordering of the clot.
 * The terms recorded in the clot are ordered w.r.t. this ordering.
 * The field {\tt sens} indicates if the monomials are ordered increasingly
 * or decreasingly (the default) w.r.t. the lexicographic ordering
 * given by the ordering of the clot.
 * The field {\tt tgest} contained the term manager of the clot.
 * 
 * Monomials are stored in {\tt bap_mont_COEFF} which can be accessed through
 * the array of {\tt bap_mont_COEFF} called {\tt tab_mont}.
 * The entries of index less than to {\tt tab_mont.size} point to
 * {\tt bap_mont_COEFF}.  The ones of index with {\tt tab_mont.size}
 * and $\mbox{\tt tab_mont.alloc} - 1$ may be zero or point to an empty
 * {\tt bap_mont_COEFF} (i.e. the field {\tt size} of which is zero).
 * Apart possibly the last one, the {\tt bap_mont_COEFF} of index less
 * than {\tt tab_mont.size} are full.
 * Monomials are numbered from $0$  to $\mbox{\tt size} - 1$.
 * 
 * The {\tt bap_mont_COEFF} are arrays of monomials.
 * The field {\tt alloc} gives the number of allocated monomials
 * of the arrays {\tt coeff} and {\tt term}. The field {\tt size} 
 * gives the number of used monomials. All allocated coefficients are 
 * initialized.
 * The field {\tt coeff} points to the array of coefficients.
 * The field {\tt term} points to the array of bap_termkeys.
 * 
 * When creating {\tt bap_clot_COEFF}, two parameters are read
 * in the polynomial parameter stack.
 */

/****************************************************************************
 MONT
 ****************************************************************************/

/*
   Allocates a bap_mont_COEFF with a capacity of at least n monomials.
   Initializes the coefficients.
*/

static bap_mont_COEFF new_mont_COEFF (bap_termanager tgest, ba0_int_p n)
{   bap_mont_COEFF mont;
    ba0_int_p i;

    mont = (bap_mont_COEFF)ba0_alloc (sizeof (struct bap_mont_COEFF));
    ba0_t2_alloc (sizeof (COEFF_t), sizeof (bap_termkey), n, 
		(void**)&mont->coeff, (void**)&mont->tclef, 
		(unsigned ba0_int_p*)&mont->alloc);
    bap_init_termkey_array_termanager (tgest, mont->tclef, mont->alloc);
    mont->size = 0;
    for (i = 0; i < mont->alloc; i++)
    {	
	COEFF_init (mont->coeff [i]);
    }
    return mont;
}

/****************************************************************************
 The CLOT
 ****************************************************************************/

static void verifie_clot_COEFF (bap_clot_COEFF C)
{
#if defined (BA0_HEAVY_DEBUG)
    ba0_int_p cumul_size, cumul_alloc, i;

    if (C == (bap_clot_COEFF)0)
	return;

    if (C->tab_mont.size > C->tab_mont.alloc || C->size > C->alloc)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    
    if (C->size > 0 && C->tab_mont.tab == (bap_mont_COEFF*)0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    cumul_size = 0;
    cumul_alloc = 0;
    for (i = 0; i < C->tab_mont.size; i++)
    {	if (C->tab_mont.tab [i] == (bap_mont_COEFF)0)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	if (i < C->tab_mont.size - 1 && 
		C->tab_mont.tab [i]->size != C->tab_mont.tab [i]->alloc)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	cumul_size += C->tab_mont.tab [i]->size;
	cumul_alloc += C->tab_mont.tab [i]->alloc;
	if (cumul_size < 0 || cumul_alloc < 0)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
    }

    if (cumul_size != C->size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    while (i < C->tab_mont.alloc && C->tab_mont.tab [i] != (bap_mont_COEFF)0)
	cumul_alloc += C->tab_mont.tab [i++]->alloc;

    if (cumul_alloc != C->alloc)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
#else
    C = (bap_clot_COEFF)0;
#endif
}

/*
  The reference ordering attributed to a bap_clot_COEFF at its creation
  is the current ordering.
*/

/*
  Initializes all the fields. In particular the memory.
  Takes two implicit parameters on the top of bap_polpar_stack.
*/

static void init_clot_COEFF (bap_clot_COEFF C, bav_term T)
{
    C->alloc = 0;
    C->size = 0;
    bap_init_termanager (&C->tgest, bap_current_typeof_termanager (), T);
    ba0_init_table ((ba0_table)&C->tab_mont);
    C->ordering = bav_R_Iordering ();
    C->sens = bap_current_typeof_monom_sequence ();
}

/*
  Initializes C and stores D in it.
  The sub-arrays of bap_mont_COEFF are shared.
*/

static void init_set_clot_COEFF (bap_clot_COEFF C, bap_clot_COEFF D)
{
    C->alloc = D->alloc;
    C->size = D->size;
    bap_init_set_termanager (&C->tgest, &D->tgest);
    C->tab_mont = D->tab_mont;
    C->ordering = D->ordering;
    C->sens = D->sens;
}

    
/*
 * Allocates a clot and initializes it. The term~$T$ gives a bound on the
 * terms stored in the clot.
 */

BAP_DLL bap_clot_COEFF bap_new_clot_COEFF (bav_term T)
{   bap_clot_COEFF C;
 
    C = (bap_clot_COEFF)ba0_alloc (sizeof (struct bap_clot_COEFF));
    init_clot_COEFF (C, T);
    return C;
}

/*
   Returns true if $C$ is zero.
*/

BAP_DLL bool bap_is_zero_clot_COEFF (bap_clot_COEFF C)
{
    return C->size == 0;
}


/* 
 * Reverts the monomials of~$C$.  Modifies C->sens. 
 */

BAP_DLL void bap_reverse_clot_COEFF (bap_clot_COEFF C)
{   struct bap_itermon_clot_COEFF I, J;
    ba0_int_p i;
    struct ba0_mark M;

    ba0_record (&M);
    bap_begin_itermon_clot_COEFF (&I, C);
    bap_end_itermon_clot_COEFF (&J, C);
    for (i = 0; i < C->size / 2; i++)
    {	bap_swap_itermon_clot_COEFF (&I, &J);
	bap_next_itermon_clot_COEFF (&I);
	bap_prev_itermon_clot_COEFF (&J);
    }
    switch (C->sens)
    {	case bap_decreasing_monom_sequence:
	    C->sens = bap_increasing_monom_sequence;
	    break;
	case bap_increasing_monom_sequence:
	    C->sens = bap_decreasing_monom_sequence;
	    break;
	default:
	    BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    }
    ba0_restore (&M);
}

/*
 * Changes the ordering of~$C$. Take care.
 */

BAP_DLL void bap_change_ordering_clot_COEFF (bap_clot_COEFF C, bav_Iordering r)
{
    C->ordering = r;
}

/*
 * One wishes to read the monomials of a clot.
 * The first monomial of a clot is the one which has number zero (the first
 * physically occuring in the clot).
 * 
 * The field {\tt clot} points to the clot.
 * A monomial is referred to by a {\tt struct bap_composite_number}.
 * The field {\tt global} is the number of the monomial.
 * The field {\tt primaire} is the index in {\tt tab_mont.tab} of the
 * {\tt bap_mont_COEFF} which contains the monomial.
 * The field {\tt secondaire} is the index of the monomial in the 
 * {\tt bap_mont_COEFF}.
 */

/*
 * Initializes $\ast\mbox{\em iter}$ at the beginning of the clot.
 */

BAP_DLL void bap_begin_itermon_clot_COEFF 
			(bap_itermon_clot_COEFF iter, bap_clot_COEFF clot)
{ 
    verifie_clot_COEFF (clot);
    iter->clot = clot;
    iter->num.primaire = 0;
    iter->num.secondaire = 0;
    iter->num.global = 0;
}

/*
 * Initializes $\ast \mbox{\em iter}$ at the end of the clot.
 */

BAP_DLL void bap_end_itermon_clot_COEFF 
			(bap_itermon_clot_COEFF iter, bap_clot_COEFF clot)
{   
    verifie_clot_COEFF (clot);
    iter->clot = clot;
    if (bap_is_zero_clot_COEFF (clot))
    {	iter->num.primaire = -1;
	iter->num.global = -1;
    } else
    {   iter->num.primaire = clot->tab_mont.size - 1;
	iter->num.secondaire = 
			clot->tab_mont.tab [iter->num.primaire]->size - 1;
	iter->num.global = clot->size - 1;
    }
}

/*
 * Returns true if $\ast \mbox{\em iter}$ has exited from the clot.
 */

BAP_DLL bool bap_outof_itermon_clot_COEFF (bap_itermon_clot_COEFF iter)
{
    return iter->num.global < 0 || iter->num.global >= iter->clot->size;
}

/*
 * One step.
 */

BAP_DLL void bap_next_itermon_clot_COEFF (bap_itermon_clot_COEFF iter)
{   bap_mont_COEFF m;

    if (bap_outof_itermon_clot_COEFF (iter))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    m = iter->clot->tab_mont.tab [iter->num.primaire];
    if (++iter->num.secondaire == m->size)
    {	iter->num.primaire++;
	iter->num.secondaire = 0;
    }
    iter->num.global++;
}

/*
 * One step backwards.
 */

BAP_DLL void bap_prev_itermon_clot_COEFF (bap_itermon_clot_COEFF iter)
{   bap_mont_COEFF m;

    if (bap_outof_itermon_clot_COEFF (iter))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (--iter->num.secondaire < 0)
    {	iter->num.primaire--;
	if (iter->num.primaire >= 0)
	{   m = iter->clot->tab_mont.tab [iter->num.primaire];
	    iter->num.secondaire = m->size - 1;
	}
    }
    iter->num.global--;
}

/*
 * Returns the number of the current monomial.
 */

BAP_DLL ba0_int_p bap_number_itermon_clot_COEFF (bap_itermon_clot_COEFF iter)
{
    return iter->num.global;
}

/*
 * Stores the current term in~$T$.
 * This term is ordered w.r.t. the ordering of the clot.
 */

BAP_DLL void bap_term_itermon_clot_COEFF (bav_term T, bap_itermon_clot_COEFF iter)
{   bap_termkey c;

    if (bap_outof_itermon_clot_COEFF (iter))
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
    c = iter->clot->tab_mont.tab 
			[iter->num.primaire]->tclef [iter->num.secondaire];
    bap_set_term_from_termkey_termanager (&iter->clot->tgest, T, c);
}

/*
 * Returns the address of the current coefficient.
 */

BAP_DLL COEFF_t* bap_coeff_itermon_clot_COEFF (bap_itermon_clot_COEFF iter)
{
    if (bap_outof_itermon_clot_COEFF (iter))
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
    return &iter->clot->tab_mont.tab 
			[iter->num.primaire]->coeff [iter->num.secondaire];
}

/*
 * Swaps the monomials the iterators point to.
 */

BAP_DLL void bap_swap_itermon_clot_COEFF 
			(bap_itermon_clot_COEFF I, bap_itermon_clot_COEFF J)
{   bap_mont_COEFF mi, mj;
    if (bap_outof_itermon_clot_COEFF (I) || bap_outof_itermon_clot_COEFF (J))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    mi = I->clot->tab_mont.tab [I->num.primaire];
    mj = J->clot->tab_mont.tab [J->num.primaire];
    BA0_SWAP (bap_termkey, mi->tclef [I->num.secondaire], 
					mj->tclef [J->num.secondaire]);
    COEFF_swap (mi->coeff [I->num.secondaire], 
					mj->coeff [J->num.secondaire]);
}

/*
 * Moves the iterator to the monomial of nuber {\em numero}.
 */

BAP_DLL void bap_goto_itermon_clot_COEFF (bap_itermon_clot_COEFF iter, ba0_int_p numero)
{   bap_mont_COEFF *tab;
    ba0_int_p primaire, secondaire;

    if (numero < 0 || numero >= iter->clot->size)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);

    tab = (bap_mont_COEFF*)iter->clot->tab_mont.tab;
    primaire = 0;
    secondaire = numero; 
    while (secondaire >= tab [primaire]->size)
    {	secondaire -= tab [primaire]->size;
	primaire++;
    }
    iter->num.global = numero;
    iter->num.primaire = primaire;
    iter->num.secondaire = secondaire;
}

/*
 * \subsubsection{Creators of clot}
 * 
 * They are used to create clots monomial by monomial.
 * 
 * \typedef struct bap_creator_clot_COEFF bap_clot_COEFF.xh
 * 
 * The iterator points to the monomial to be rewritten.
 * The field nbmon_par_mont gives the size of the struct bap_mont_COEFF
 * to be allocated.
 */

/*
 * Initializes $\ast\mbox{\em crea}$ at the beginning of the clot.
 * If some {\tt bap_mont_COEFF} need to be allocated, they will have the
 * size {\em nbmon_par_mont}.
 */


BAP_DLL void bap_begin_creator_clot_COEFF 
			(bap_creator_clot_COEFF crea, bap_clot_COEFF clot, 
			 bav_term T, ba0_int_p nbmon_par_mont)
{   bap_mont_COEFF m;
    ba0_int_p n;
    bool reinit;

    bap_reset_termanager 
	(&clot->tgest, bap_current_typeof_termanager (), T, &reinit);
    if (reinit)
    {   for (n = 0; n < clot->tab_mont.size; n++)
	{   m = clot->tab_mont.tab [n];
	    bap_init_termkey_array_termanager 
					(&clot->tgest, m->tclef, m->alloc);
	}
	for (n = clot->tab_mont.size; n < clot->tab_mont.alloc; n++)
	{   m = clot->tab_mont.tab [n];
	    if (m != 0)
	 	bap_init_termkey_array_termanager 
					(&clot->tgest, m->tclef, m->alloc); 
	}
    }

    clot->ordering = bav_R_Iordering ();
    clot->sens = bap_current_typeof_monom_sequence ();

    crea->iter.clot = clot;

    crea->iter.num.global = 0;
    crea->iter.num.primaire = 0;
    crea->iter.num.secondaire = 0;

    crea->nbmon_par_mont = nbmon_par_mont;
}

/*
 * Sets the iterator just after the last monomial of the clot.
 */

/*
 * Initializes the creator at the end of the clot.
 * If some {\tt bap_mont_COEFF} need to be allocated, they will have the
 * size {\em nbmon_par_mont}.
 */

BAP_DLL void bap_append_creator_clot_COEFF (bap_creator_clot_COEFF crea, 
			bap_clot_COEFF clot, ba0_int_p nbmon_par_mont)
{   bap_mont_COEFF m;

    verifie_clot_COEFF (clot);

    crea->iter.clot = clot;
    crea->iter.num.global = clot->size;
/*
   On place l'iterateur juste derriere le dernier monome utilise de la clot.
   On n'en fait pas une fonction d'itermon puisque, par definition, on
	fait un debordement de ba0_table !
*/
    if (bap_is_zero_clot_COEFF (clot))
    {	crea->iter.num.primaire = 0;
	crea->iter.num.secondaire = 0;
    } else
    {   crea->iter.num.primaire = clot->tab_mont.size - 1;
	m = clot->tab_mont.tab [crea->iter.num.primaire];
	crea->iter.num.secondaire = m->size;
	if (crea->iter.num.secondaire == m->alloc)
	{   crea->iter.num.primaire++;
	    crea->iter.num.secondaire = 0;
	}
    }
    crea->nbmon_par_mont = nbmon_par_mont;
}

/*
  Returns the bap_mont_COEFF crea points to.
  Allocates a bap_mont_COEFF if there is no place left.
  If as many used monomials as allocated monomials
     one needs to allocate a new mont_COEFF
     Now, if tab_mont.tab is full then it must be
	resized before allocating the mont_COEFF.
  else
     there is no need to allocated a new mont_COEFF
     it may be useful to move to the next mont_COEFF.
*/

static bap_mont_COEFF mont_address_creator_clot_COEFF 
						(bap_creator_clot_COEFF crea)
{   bap_clot_COEFF clot;
    bap_mont_COEFF mont;

    clot = crea->iter.clot;

    if (clot->alloc == crea->iter.num.global)
    {   if (clot->tab_mont.alloc == crea->iter.num.primaire)
	{   bap_mont_COEFF* new_tab;
	    ba0_int_p n, new_alloc;

	    new_alloc = 2 * clot->tab_mont.alloc + 1;
	    new_tab = (bap_mont_COEFF*)ba0_alloc (sizeof (void*) * new_alloc);
	    memcpy (new_tab, clot->tab_mont.tab, 
			clot->tab_mont.alloc * sizeof (bap_mont_COEFF));
	    for (n = clot->tab_mont.alloc; n < new_alloc; n++)
		new_tab [n] = (bap_mont_COEFF)0;
	    clot->tab_mont.alloc = new_alloc;
	    clot->tab_mont.tab = (bap_mont_COEFF*)new_tab;
	}
	mont = new_mont_COEFF (&clot->tgest, crea->nbmon_par_mont);
	clot->tab_mont.tab [crea->iter.num.primaire] = mont;
	clot->alloc += mont->alloc;
    } else 
    	mont = clot->tab_mont.tab [crea->iter.num.primaire];

    return mont;
}

/*
   There is a slight difference with bap_next_itermon_clot_COEFF (comparison
   with alloc instead of size though it amounts to the same). There is no
   overflow test.
*/

static void next_creator_clot_COEFF (bap_creator_clot_COEFF crea)
{   bap_mont_COEFF mont;

    mont = crea->iter.clot->tab_mont.tab [crea->iter.num.primaire];
    if (++crea->iter.num.secondaire == mont->alloc)
    {	crea->iter.num.primaire++;
	crea->iter.num.secondaire = 0;
    }
    crea->iter.num.global++;
}

/*
 * Writes $c\,T$ on the clot.
 * The coefficient $c$ may be zero (then nothing is done).
 */

void bap_write_creator_clot_COEFF 
			(bap_creator_clot_COEFF crea, bav_term T, COEFF_t c)
{   bap_mont_COEFF mont;

    if (COEFF_sgn (c) == 0)
	return;

    mont = mont_address_creator_clot_COEFF (crea);

    bap_set_termkey_from_term_termanager 
	(&crea->iter.clot->tgest, &mont->tclef [crea->iter.num.secondaire], T);
    COEFF_set (mont->coeff [crea->iter.num.secondaire], c);

    next_creator_clot_COEFF (crea);
}

/*
 * Writes $- c\,T$ on the clot.
 */

BAP_DLL void bap_write_neg_creator_clot_COEFF 
			(bap_creator_clot_COEFF crea, bav_term T, COEFF_t c)
{   bap_mont_COEFF mont;

    if (COEFF_sgn (c) == 0)
	return;

    mont = mont_address_creator_clot_COEFF (crea);

    bap_set_termkey_from_term_termanager 
	(&crea->iter.clot->tgest, &mont->tclef [crea->iter.num.secondaire], T);
    COEFF_neg (mont->coeff [crea->iter.num.secondaire], c);

    next_creator_clot_COEFF (crea);
}

/*
 * Rewrites the current term with~$T$.
 * It is assumed that the creator points to an existing monomial.
 */

BAP_DLL void bap_write_term_creator_clot_COEFF 
				(bap_creator_clot_COEFF crea, bav_term T)
{   bap_mont_COEFF mont;

    mont = mont_address_creator_clot_COEFF (crea);
    bap_set_termkey_from_term_termanager 
	(&crea->iter.clot->tgest, &mont->tclef [crea->iter.num.secondaire], T);
    next_creator_clot_COEFF (crea);
}

/*
 * Writes on $\ast\mbox{\em crea}$ all the monomials of {\em clot}
 * with number $l \leq n < r$.
 */

BAP_DLL void bap_write_all_creator_clot_COEFF (bap_creator_clot_COEFF crea, 
				bap_clot_COEFF clot, ba0_int_p l, ba0_int_p r)
{   struct bap_itermon_clot_COEFF iter;
    bap_mont_COEFF d_mont, s_mont;
    bap_termkey *d_clef, s_clef;
    ba0_int_p i;

    if (l == r)
	return;

    bap_begin_itermon_clot_COEFF (&iter, clot);
    bap_goto_itermon_clot_COEFF (&iter, l);

    for (i = l; i < r; i++)
    {	
	s_mont = iter.clot->tab_mont.tab [iter.num.primaire];
	s_clef = s_mont->tclef [iter.num.secondaire];
	d_mont = mont_address_creator_clot_COEFF (crea);
	d_clef = &d_mont->tclef [crea->iter.num.secondaire];
	bap_set_termkey_from_termkey_termanager 
		(&crea->iter.clot->tgest, d_clef, &iter.clot->tgest, s_clef);
	COEFF_set (d_mont->coeff [crea->iter.num.secondaire], 
				   s_mont->coeff [iter.num.secondaire]);
	bap_next_itermon_clot_COEFF (&iter);
	next_creator_clot_COEFF (crea);
    }
}

/* Same function as above. Changes the sign of the coefficients. */

BAP_DLL void bap_write_neg_all_creator_clot_COEFF 
		(bap_creator_clot_COEFF crea, bap_clot_COEFF clot, 
		ba0_int_p l, ba0_int_p r)
{   struct bap_itermon_clot_COEFF iter;
    bap_mont_COEFF d_mont, s_mont;
    bap_termkey *d_clef, s_clef;
    ba0_int_p i;

    if (l == r)
	return;

    bap_begin_itermon_clot_COEFF (&iter, clot);
    bap_goto_itermon_clot_COEFF (&iter, l);

    for (i = l; i < r; i++)
    {	
	s_mont = iter.clot->tab_mont.tab [iter.num.primaire];
	s_clef = s_mont->tclef [iter.num.secondaire];
	d_mont = mont_address_creator_clot_COEFF (crea);
	d_clef = &d_mont->tclef [crea->iter.num.secondaire];
	bap_set_termkey_from_termkey_termanager 
		(&crea->iter.clot->tgest, d_clef, &iter.clot->tgest, s_clef);
	COEFF_neg (d_mont->coeff [crea->iter.num.secondaire], 
				   s_mont->coeff [iter.num.secondaire]);
	bap_next_itermon_clot_COEFF (&iter);
	next_creator_clot_COEFF (crea);
    }
}

/* Same function as above.  Coefficients are multiplied by~$c$. */

BAP_DLL void bap_write_mul_all_creator_clot_COEFF 
	(bap_creator_clot_COEFF crea, bap_clot_COEFF clot, 
	 COEFF_t c, ba0_int_p l, ba0_int_p r)
{   struct bap_itermon_clot_COEFF iter;
    bap_mont_COEFF d_mont, s_mont;
    bap_termkey *d_clef, s_clef;
    ba0_int_p i;

    if (l == r || COEFF_sgn (c) == 0)
	return;

    bap_begin_itermon_clot_COEFF (&iter, clot);
    bap_goto_itermon_clot_COEFF (&iter, l);

    for (i = l; i < r; i++)
    {	
	s_mont = iter.clot->tab_mont.tab [iter.num.primaire];
	s_clef = s_mont->tclef [iter.num.secondaire];
	d_mont = mont_address_creator_clot_COEFF (crea);
	d_clef = &d_mont->tclef [crea->iter.num.secondaire];
	bap_set_termkey_from_termkey_termanager 
		(&crea->iter.clot->tgest, d_clef, &iter.clot->tgest, s_clef);
	COEFF_mul (d_mont->coeff [crea->iter.num.secondaire], 
			   s_mont->coeff [iter.num.secondaire], c);
	if (COEFF_sgn (d_mont->coeff [crea->iter.num.secondaire]) == 0)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	bap_next_itermon_clot_COEFF (&iter);
	next_creator_clot_COEFF (crea);
    }
}

/* Same function as above.  Coefficients are divided by~$c$. */

#if defined (BAD_FLAG_mpz)
BAP_DLL void bap_write_exquo_all_creator_clot_COEFF 
	(bap_creator_clot_COEFF crea, bap_clot_COEFF clot, 
					COEFF_t c, ba0_int_p l, ba0_int_p r)
{   struct bap_itermon_clot_COEFF iter;
    bap_mont_COEFF d_mont, s_mont;
    bap_termkey *d_clef, s_clef;
    ba0_int_p i;

    if (l == r || COEFF_is_one (c))
	return;

    bap_begin_itermon_clot_COEFF (&iter, clot);
    bap_goto_itermon_clot_COEFF (&iter, l);

    for (i = l; i < r; i++)
    {	
	s_mont = iter.clot->tab_mont.tab [iter.num.primaire];
	s_clef = s_mont->tclef [iter.num.secondaire];
	d_mont = mont_address_creator_clot_COEFF (crea);
	d_clef = &d_mont->tclef [crea->iter.num.secondaire];
	bap_set_termkey_from_termkey_termanager 
		(&crea->iter.clot->tgest, d_clef, &iter.clot->tgest, s_clef);
	mpz_divexact (d_mont->coeff [crea->iter.num.secondaire], 
			s_mont->coeff [iter.num.secondaire], c);
	bap_next_itermon_clot_COEFF (&iter);
	next_creator_clot_COEFF (crea);
    }
}
#endif

/* Close the creation. */

BAP_DLL void bap_close_creator_clot_COEFF (bap_creator_clot_COEFF crea)
{   bap_clot_COEFF clot;
    bap_mont_COEFF m;
    ba0_int_p i;

    clot = crea->iter.clot;

    clot->size = crea->iter.num.global;

    for (i = 0; i < crea->iter.num.primaire; i++)
    {	m = clot->tab_mont.tab [i];
	m->size = m->alloc;
    }

    if (crea->iter.num.secondaire != 0)
    {	m = clot->tab_mont.tab [crea->iter.num.primaire];
	m->size = crea->iter.num.secondaire;
	clot->tab_mont.size = crea->iter.num.primaire + 1;
    } else
	clot->tab_mont.size = crea->iter.num.primaire;

    for (i = clot->tab_mont.size; i < clot->tab_mont.alloc; i++)
    {	m = clot->tab_mont.tab [i];
	if (m != (bap_mont_COEFF)0)
	    m->size = 0;
    }
    verifie_clot_COEFF (clot);
}

/*
 * Moves the creator to the monomial with number {\em numero}.
 * No allocation is performed. It is assumed that the monomial exists.
 */

BAP_DLL void bap_goto_creator_clot_COEFF 
			(bap_creator_clot_COEFF crea, ba0_int_p numero)
{
    bap_goto_itermon_clot_COEFF (&crea->iter, numero);
}

/****************************************************************************
 REORDONNE
 ****************************************************************************/

/*
 * Type declaration used for a local variable of quicksort.
 */

struct quicksort_data {
    struct bap_itermon_clot_COEFF l; 
    struct bap_itermon_clot_COEFF r; 
    struct bap_itermon_clot_COEFF i; 
    struct bap_itermon_clot_COEFF j; 
    struct bap_itermon_clot_COEFF k;
    struct bav_term Tl; 
    struct bav_term Tr; 
    struct bav_term Ti; 
    struct bav_term Tj; 
    struct bav_term Tk; 
    struct bav_term pivot;
    unsigned ba0_int_p zi;
    unsigned ba0_int_p zj;
    unsigned ba0_int_p zk;
};

static void bap_init_quicksort_data 
			(struct quicksort_data* qs, bap_clot_COEFF clot)
{
    bap_begin_itermon_clot_COEFF (&qs->l, clot);
    bap_begin_itermon_clot_COEFF (&qs->r, clot);
    bap_begin_itermon_clot_COEFF (&qs->i, clot);
    bap_begin_itermon_clot_COEFF (&qs->j, clot);
    bap_begin_itermon_clot_COEFF (&qs->k, clot);
    bav_init_term (&qs->Tl);
    bav_init_term (&qs->Tr);
    bav_init_term (&qs->Ti);
    bav_init_term (&qs->Tj);
    bav_init_term (&qs->Tk);
    bav_init_term (&qs->pivot);
    qs->zi = qs->zj = qs->zk = 0;
}

static void quicksort_clot_COEFF 
			(ba0_int_p l, ba0_int_p r, struct quicksort_data* qs)
{   ba0_int_p i, j, k;
    enum ba0_compare_code code;
/*
 * Readonly static data structure
 */
    static ba0_int_p alpha [] = { 1, 2, 2, 3, 4, 1, 3 };
    /*                      -  -  -  -  -  -  -         */
    static ba0_int_p beta []  = { 1, 3, 1, 4, 3, 2, 2 };

    if (r - l > 8)
    {
        i = (alpha[qs->zi] * (l+2) + beta[qs->zi] * (r-2)) / 
						(alpha[qs->zi] + beta[qs->zi]);
        qs->zi = (qs->zi + 1) % 7;
        j = (alpha[qs->zj] * (l+1) + beta[qs->zj] * (i-1)) / 
						(alpha[qs->zj] + beta[qs->zj]);
        qs->zj = (qs->zj + 3) % 7;
        k = (alpha[qs->zk] * (i+1) + beta[qs->zk] * (r-1)) / 
						(alpha[qs->zk] + beta[qs->zk]);
        qs->zk = (qs->zk + 5) % 7;
        if (i == j || i == k || j == k)
            BA0_RAISE_EXCEPTION (BA0_ERRALG);
        bap_goto_itermon_clot_COEFF (&qs->l, l);
        bap_term_itermon_clot_COEFF (&qs->Tl, &qs->l);
        bap_goto_itermon_clot_COEFF (&qs->r, r);
        bap_term_itermon_clot_COEFF (&qs->Tr, &qs->r);
        bap_goto_itermon_clot_COEFF (&qs->i, i);
        bap_term_itermon_clot_COEFF (&qs->Ti, &qs->i);
        bap_goto_itermon_clot_COEFF (&qs->j, j);
        bap_term_itermon_clot_COEFF (&qs->Tj, &qs->j);
        bap_goto_itermon_clot_COEFF (&qs->k, k);
        bap_term_itermon_clot_COEFF (&qs->Tk, &qs->k);
/*
   On souhaite que le pivot soit le median de T [i], T [j], T [k].
   On doit echanger T [l] avec le median et T [r] avec le max.
*/
        code = bav_compare_term (&qs->Ti, &qs->Tj);
        if (code == ba0_gt)
        {   code = bav_compare_term (&qs->Tj, &qs->Tk);
            if (code == ba0_gt)
	    {	bap_swap_itermon_clot_COEFF (&qs->l, &qs->j);
		bap_swap_itermon_clot_COEFF (&qs->r, &qs->i);
/* T [i] > T [j] > T [k] */
                bav_set_term (&qs->pivot, &qs->Tj);
            } else
            {   code = bav_compare_term (&qs->Ti, &qs->Tk);
                if (code == ba0_gt)
		{   bap_swap_itermon_clot_COEFF (&qs->l, &qs->k);
		    bap_swap_itermon_clot_COEFF (&qs->r, &qs->i);
/* T [i] > T [k] > T [j] */
		    bav_set_term (&qs->pivot, &qs->Tk);
		} else
		{   bap_swap_itermon_clot_COEFF (&qs->l, &qs->i);
		    bap_swap_itermon_clot_COEFF (&qs->r, &qs->k);
/* T [k] > T [i] > T [j] */
                    bav_set_term (&qs->pivot, &qs->Ti);
                }
	    }
	} else
	{   code = bav_compare_term (&qs->Tk, &qs->Tj);
            if (code == ba0_gt)
	    {   bap_swap_itermon_clot_COEFF (&qs->l, &qs->j);
		bap_swap_itermon_clot_COEFF (&qs->r, &qs->k);
/* T [k] > T [j] > T [i] */
                bav_set_term (&qs->pivot, &qs->Tj);
	    } else
	    {	code = bav_compare_term (&qs->Tk, &qs->Ti);
		if (code == ba0_gt)
		{   bap_swap_itermon_clot_COEFF (&qs->l, &qs->k);
		    bap_swap_itermon_clot_COEFF (&qs->r, &qs->j);
/* T [j] > T [k] > T [i] */
		    bav_set_term (&qs->pivot, &qs->Tk);
		} else
		{   bap_swap_itermon_clot_COEFF (&qs->l, &qs->i);
		    bap_swap_itermon_clot_COEFF (&qs->r, &qs->j);
/* T [j] > T [i] > T [k] */
                    bav_set_term (&qs->pivot, &qs->Ti);
	 	}
	    }
	}
/*
   Invariants (sauf avant le 1er tour).
   i, j in [l, r]
   T [i] < pivot (sauf 1er tour ou on a egalite)
   T [j] > pivot
   Les T [k < i] sont > pivot
   Les T [k > j] sont < pivot
*/
	i = l;
	j = r;
	qs->i = qs->l;
	qs->j = qs->r;
	do
	{   bap_swap_itermon_clot_COEFF (&qs->i, &qs->j);
	    do
	    {	i += 1;
		bap_next_itermon_clot_COEFF (&qs->i);
		bap_term_itermon_clot_COEFF (&qs->Ti, &qs->i);
		code = bav_compare_term (&qs->Ti, &qs->pivot);
	    } while (code == ba0_gt);
	    do
	    {	j -= 1;
		bap_prev_itermon_clot_COEFF (&qs->j);
		bap_term_itermon_clot_COEFF (&qs->Tj, &qs->j);
		code = bav_compare_term (&qs->Tj, &qs->pivot);
	    } while (code == ba0_lt);
	} while (i < j);
	bap_swap_itermon_clot_COEFF (&qs->i, &qs->r);
	quicksort_clot_COEFF (l, i - 1, qs);
	quicksort_clot_COEFF (i + 1, r, qs);
    } else
    {	bool sorted;
	bav_term cour, suiv;

	sorted = false;
	cour = &qs->Ti;
	suiv = &qs->Tj;
	for (i = r - 1; ! sorted && i >= l; i--)
	{   sorted = true;
	    bap_goto_itermon_clot_COEFF (&qs->i, l);
	    bap_term_itermon_clot_COEFF (cour, &qs->i);
	    bap_goto_itermon_clot_COEFF (&qs->j, l + 1);
	    bap_term_itermon_clot_COEFF (suiv, &qs->j);
	    code = bav_compare_term (cour, suiv);
	    if (code == ba0_lt)
	    {	BA0_SWAP (bav_term, cour, suiv);
		bap_swap_itermon_clot_COEFF (&qs->i, &qs->j);
		sorted = false;
	    }
	    for (j = l + 1; j <= i; j++)
	    {	qs->i = qs->j;
		bap_next_itermon_clot_COEFF (&qs->j);
		BA0_SWAP (bav_term, cour, suiv);
		bap_term_itermon_clot_COEFF (suiv, &qs->j);
		code = bav_compare_term (cour, suiv);
		if (code == ba0_lt)
		{   BA0_SWAP (bav_term, cour, suiv);
		    bap_swap_itermon_clot_COEFF (&qs->i, &qs->j);
		    sorted = false;
		}
	    }
	}
    }
}

/*
 * Sorts the monomials of the clot with numbers $l \leq n < r$ w.r.t.
 * the current ordering and the {\em sens} on the top of the polynomial
 * parameter stack.
 */


BAP_DLL void bap_sort_clot_COEFF (bap_clot_COEFF clot, ba0_int_p l, ba0_int_p r)
{   struct bap_clot_COEFF old_clot;
    struct bap_creator_clot_COEFF crea;
    struct bap_itermon_clot_COEFF iter;
    struct quicksort_data qs;
    struct bav_term T;
    enum bap_typeof_monom_sequence sens;
    ba0_int_p i;
    struct ba0_mark M;

    if (l >= r)
	return;

    ba0_record (&M);

    sens = bap_current_typeof_monom_sequence ();
    if (sens != bap_increasing_monom_sequence && 
					sens != bap_decreasing_monom_sequence)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
 
    init_set_clot_COEFF (&old_clot, clot);
    bap_begin_itermon_clot_COEFF (&iter, &old_clot);
    bap_goto_itermon_clot_COEFF (&iter, l);

    bav_init_term (&T);
    bav_set_term (&T, &clot->tgest.rang_total);
    bav_sort_term (&T);

    bap_push_clot_parameters_stack 
	(bap_current_typeof_termanager (), bap_decreasing_monom_sequence);
    bap_begin_creator_clot_COEFF (&crea, clot, &T, 0);
    bap_pull_clot_parameters_stack ();

    bap_goto_creator_clot_COEFF (&crea, l);

    for (i = l; i < r; i++)
    {	bap_term_itermon_clot_COEFF (&T, &iter);
	bav_sort_term (&T);
	bap_write_term_creator_clot_COEFF (&crea, &T);
	bap_next_itermon_clot_COEFF (&iter);
    }

    bap_init_quicksort_data (&qs, clot);
    quicksort_clot_COEFF (l, r - 1, &qs);

    if (sens == bap_increasing_monom_sequence)
	bap_reverse_clot_COEFF (clot);

    ba0_restore (&M);
}

/****************************************************************************
 GARBAGE COLLECTOR AND RELATED FUNCTIONS
 ****************************************************************************/

/*
 * Readonly static data
 */

static char _struct_mont [] = "struct bap_mont_COEFF";
static char _struct_mont_tclef [] = "bap_mont_COEFF->tclef";
static char _struct_mont_coeff [] = "bap_mont_COEFF->coeff";

static ba0_int_p garbage1_mont_COEFF 
	(bap_termanager tgest, bap_mont_COEFF A, enum ba0_garbage_code code)
{
    ba0_int_p n = 0;

    if (A == (bap_mont_COEFF)0)
	return 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info 
		(A, sizeof (struct bap_mont_COEFF), _struct_mont);
    n += ba0_new_gc_info 
	(A->tclef, sizeof(bap_termkey) * A->alloc, _struct_mont_tclef);
    n += ba0_new_gc_info 
	(A->coeff, sizeof(COEFF_t) * A->alloc, _struct_mont_coeff);
    if (bap_worth_garbage_termkey_termanager (tgest))
    {   ba0_int_p i;
	for (i = 0; i < A->alloc; i++)
	    n += bap_garbage1_termkey_termanager 
					(tgest, A->tclef [i], ba0_embedded);
    }

#if ! defined (BAD_FLAG_mint_hp)
    {	ba0_int_p i;
	for (i = 0; i < A->alloc; i++)
	    n += ba0_garbage1_COEFF (&A->coeff [i], ba0_embedded);
    }
#endif
    return n;
}

/*
 * Readonly static data
 */

static char _struct_clot [] = "struct bap_clot_COEFF";
static char _struct_clot_tab [] = "bap_clot_COEFF->tab_mont.tab";

BAP_DLL ba0_int_p bap_garbage1_clot_COEFF (void* AA, enum ba0_garbage_code code)
{   bap_clot_COEFF A = (bap_clot_COEFF)AA;
    ba0_int_p i, n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info 
		(A, sizeof (struct bap_clot_COEFF), _struct_clot);

    if (A->tab_mont.tab)
        n += ba0_new_gc_info
		(A->tab_mont.tab, sizeof(bap_mont_COEFF) * A->tab_mont.alloc,
		 _struct_clot_tab);

    n += bap_garbage1_termanager (&A->tgest, ba0_embedded);

    for (i = 0; i < A->tab_mont.alloc; i++)
	n += garbage1_mont_COEFF (&A->tgest, A->tab_mont.tab [i], ba0_isolated);

    return n;
}

static bap_mont_COEFF garbage2_mont_COEFF 
	(bap_termanager tgest, bap_mont_COEFF A, enum ba0_garbage_code code)
{

    if (A == (bap_mont_COEFF)0)
	return A;

    if (code == ba0_isolated)
	A = (bap_mont_COEFF)ba0_new_addr_gc_info (A, _struct_mont);

    A->tclef = (bap_termkey*)ba0_new_addr_gc_info 
			(A->tclef, _struct_mont_tclef);

    A->coeff = (COEFF_t*)ba0_new_addr_gc_info
			(A->coeff, _struct_mont_coeff);

    if (bap_worth_garbage_termkey_termanager (tgest))
    {   ba0_int_p i;
	for (i = 0; i < A->alloc; i++)
	    A->tclef [i] = bap_garbage2_termkey_termanager 
					(tgest, A->tclef [i], ba0_embedded);
    }

#if ! defined (BAD_FLAG_mint_hp)
    {	ba0_int_p i;
	for (i = 0; i < A->alloc; i++)
	    ba0_garbage2_COEFF (&A->coeff [i], ba0_embedded);
    }
#endif
    return A;
}

BAP_DLL void* bap_garbage2_clot_COEFF (void* AA, enum ba0_garbage_code code)
{   bap_clot_COEFF A;
    ba0_int_p i;

    if (code == ba0_isolated)
	A = (bap_clot_COEFF)ba0_new_addr_gc_info (AA, _struct_clot);
    else
	A = (bap_clot_COEFF)AA;

    if (A->tab_mont.tab)
	A->tab_mont.tab = (bap_mont_COEFF*)ba0_new_addr_gc_info 
		(A->tab_mont.tab, _struct_clot_tab);

    bap_garbage2_termanager (&A->tgest, ba0_embedded); 

    for (i = 0; i < A->tab_mont.alloc; i++)
	A->tab_mont.tab [i] = garbage2_mont_COEFF 
				(&A->tgest, A->tab_mont.tab [i], ba0_isolated);

    return A;
}

BAP_DLL void bap_switch_ring_clot_COEFF 
			(bap_clot_COEFF clot, bav_differential_ring R)
{
    bap_switch_ring_termanager (&clot->tgest, R);
}

#endif
#undef BAD_FLAG_COEFF
