#include "bap_iterator_index.h"

/*-
\subsubsection{Index creators}

\typedef struct bap_iterator_index bap_common.h

They permit us to create array of index index by index.
The index in this array must be decomposed since arrays of index
actually are arrays of arrays of index.
An index {\tt num.global} is thus decomposed as an index
{\tt num.primaire} in {\tt tab_index} and an index {\tt num.secondaire}
in {\tt tab_index [num.primaire]}.
The creators are designed so that thay can be used in same time as
index iterators (when one wishes to read and rewrite an array of index
in the same time).
-*/

/*-
\paragraph{void bap_begin_creator_index 
			(bap_creator_index crea, bap_indexed_access ind)}
Initializes $\ast\mbox{\em crea}$ at the beginning of $\ast\mbox{\em ind}$.
-*/

BAP_DLL void bap_begin_creator_index 
			(bap_creator_index crea, bap_indexed_access ind)
{
    crea->ind = ind;
    crea->num.primaire = 0;
    crea->num.secondaire = 0;
    crea->num.global = 0;
}

/*-
\paragraph{void bap_write_creator_index (bap_creator_index crea, ba0_int_p n)}
Writes index~$n$ in $\ast\mbox{\em crea}$.
Assumes that no allocation needs being performed.
-*/

BAP_DLL void bap_write_creator_index (bap_creator_index crea, ba0_int_p n)
{   ba0_tableof_int_p tab;

    if (crea->num.global >= crea->ind->alloc)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    tab = crea->ind->tab_index.tab [crea->num.primaire];
    tab->tab [crea->num.secondaire++] = n;
    if (crea->num.secondaire == tab->alloc)
    {	crea->num.primaire++;
	crea->num.secondaire = 0;
    } 
    crea->num.global++;
}

/*-
\paragraph{void bap_close_creator_index (bap_creator_index crea)}
Closes the index construction. Do not forget.
-*/

BAP_DLL void bap_close_creator_index (bap_creator_index crea)
{   ba0_tableof_int_p tab;
    ba0_int_p i;

    for (i = 0; i < crea->num.primaire; i++)
    {	tab = crea->ind->tab_index.tab [i];
	tab->size = tab->alloc;
    }
    if (crea->num.secondaire != 0)
    {	tab = crea->ind->tab_index.tab [crea->num.primaire];
	tab->size = crea->num.secondaire;
	crea->ind->tab_index.size = crea->num.primaire + 1;
    } else
	crea->ind->tab_index.size = crea->num.primaire;
    for (i = crea->ind->tab_index.size; i < crea->ind->tab_index.alloc; i++)
    {	tab = crea->ind->tab_index.tab [i];
	tab->size = 0;
    }
    crea->ind->size = crea->num.global;
}

/*-
\subsubsection{Index iterators}

\typedef struct bap_iterator_index bap_common.h

The same structure as creators.
Here, one wishes to read the index.
-*/

/*-
\paragraph{void bap_begin_iterator_index 
			(bap_iterator_index iter, bap_indexed_access ind)}
Initializes $\ast\mbox{\em iter}$ at the beginning of $\ast\mbox{\em ind}$.
-*/

BAP_DLL void bap_begin_iterator_index 
			(bap_iterator_index iter, bap_indexed_access ind)
{
    iter->ind = ind;
    iter->num.primaire = 0;
    iter->num.secondaire = 0;
    iter->num.global = 0;
}

/*-
\paragraph{void bap_end_iterator_index 
			(bap_iterator_index iter, bap_indexed_access ind)}
Initializes $\ast\mbox{\em iter}$ at the and of $\ast\mbox{\em ind}$.
-*/

BAP_DLL void bap_end_iterator_index 
			(bap_iterator_index iter, bap_indexed_access ind)
{   ba0_tableof_int_p tab;

    iter->ind = ind;
    iter->num.primaire = ind->tab_index.size - 1;
    tab = ind->tab_index.tab [iter->num.primaire];
    iter->num.secondaire = tab->size - 1;
    iter->num.global = ind->size - 1;
}

/*-
\paragraph{bool bap_outof_iterator_index (bap_iterator_index iter)}
Returns true if the iterator has exited from the array of index.
-*/

BAP_DLL bool bap_outof_iterator_index (bap_iterator_index iter)
{
    return iter->num.global < 0 || iter->num.global >= iter->ind->size;
}

/*-
\paragraph{void bap_next_iterator_index (bap_iterator_index iter)}
One step.
-*/

BAP_DLL void bap_next_iterator_index (bap_iterator_index iter)
{   ba0_tableof_int_p tab;

    if (bap_outof_iterator_index (iter))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    tab = iter->ind->tab_index.tab [iter->num.primaire];
    if (++iter->num.secondaire >= tab->alloc)
    {	iter->num.primaire++;
	iter->num.secondaire = 0;
    }
    iter->num.global++;
}

/*-
\paragraph{void bap_prev_iterator_index (bap_iterator_index iter)}
One step backwards.
-*/

BAP_DLL void bap_prev_iterator_index (bap_iterator_index iter)
{   ba0_tableof_int_p tab;

    if (bap_outof_iterator_index (iter))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (--iter->num.secondaire < 0)
    {	iter->num.primaire--;
	if (iter->num.primaire >= 0)
	{   tab = iter->ind->tab_index.tab [iter->num.primaire];
	    iter->num.secondaire = tab->alloc - 1;
	}
    }
    iter->num.global--;
}

/*-
\paragraph{void bap_goto_iterator_index (bap_iterator_index iter, ba0_int_p n)}
Moves the iterator to the~$n$th element of the array of index.
-*/

BAP_DLL void bap_goto_iterator_index (bap_iterator_index iter, ba0_int_p n)
{   ba0_tableof_int_p *tab;
    ba0_int_p i, j;

    if (n < 0 || n >= iter->ind->size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    tab = iter->ind->tab_index.tab;
    for (i = 0, j = n; j >= tab [i]->size; i++)
	j -= tab [i]->size;
    iter->num.global = n;
    iter->num.primaire = i;
    iter->num.secondaire = j;
}

/*-
\paragraph{ba0_int_p bap_index_iterator_index (bap_iterator_index iter)}
Returns the index the iterator points to.
-*/

BAP_DLL ba0_int_p bap_index_iterator_index (bap_iterator_index iter)
{   ba0_tableof_int_p tab;

    if (bap_outof_iterator_index (iter))
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
    tab = iter->ind->tab_index.tab [iter->num.primaire];
    return tab->tab [iter->num.secondaire];
}
 
/*-
\paragraph{ba0_int_p bap_read_iterator_index (bap_iterator_index iter)}
Returns the index the iterator points to and performs one step.
-*/

BAP_DLL ba0_int_p bap_read_iterator_index (bap_iterator_index iter)
{   ba0_int_p index;

    index = bap_index_iterator_index (iter);
    bap_next_iterator_index (iter);
    return index;
}

/*-
\paragraph{void bap_swapindex_iterator_index 
				(bap_iterator_index I, bap_iterator_index J)}
Swaps the index the iterators point to.
-*/

BAP_DLL void bap_swapindex_iterator_index 
				(bap_iterator_index I, bap_iterator_index J)
{   ba0_tableof_int_p ti, tj;

    if (bap_outof_iterator_index (I) || bap_outof_iterator_index (J))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    ti = I->ind->tab_index.tab [I->num.primaire];
    tj = J->ind->tab_index.tab [J->num.primaire];
    BA0_SWAP (ba0_int_p, 
		ti->tab [I->num.secondaire], tj->tab [J->num.secondaire]);
}

/*-
\paragraph{void bap_set_iterator_index 
			(bap_iterator_index I, bap_iterator_index J)}
Stores $\ast J$ in $\ast I$.
-*/

BAP_DLL void bap_set_iterator_index 
				(bap_iterator_index I, bap_iterator_index J)
{
    *I = *J;
}

