#include "bap_polynom_COEFF.h"
#include "bap_add_polynom_COEFF.h"
#include "bap_creator_COEFF.h"
#include "bap_itermon_COEFF.h"
#include "bap_itercoeff_COEFF.h"
#include "bap__check_COEFF.h"
#include "bap_product_COEFF.h"
#include "bap_geobucket_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/*
 * A differential polynomial A is a selection in a clot
 * w.r.t. some ordering, called ordering of A.
 * The goal is to be able to manipulate the coefficients, the reductum \dots
 * of a polynomial without having to reconstruct a clot.
 * And to be able to manipulate them w.r.t. orderings which are not
 * necessary the ordering of the clot.
 * 
 * The field acces indicates if the access to the monomials is
 * sequential or through an index table. If the access is sequential
 * then the fields seq.first and seq.after provides
 * the numbers of the first monomial of A and of the first monomial
 * following A. The difference between these two fields gives the
 * number of monomials of A.
 * If the access is indexed, the structure ind provides an array
 * of index (numbers) of the monomials of A.
 * The zero polynomial is any polynomial made of no monomials.
 * 
 * The structure select permits to compose access to coefficients
 * and changes of orderings. Let the current term T to be determined.
 * Using its bap_termkey, one gets a term T'. This term T' is a multiple of
 * T. One gets T by removing some ranks in T'. The ordering on the
 * top of the selection stack, which is the ordering of A, is then
 * used to sort T.
 * 
 * The boolean coeff indicates if A is readonly or not.
 * The readonly polynomials are polynomials sharing datas with
 * some other polynomial. They are obtained either by readonly constructors
 * or by iterators of coefficients.
 * 
 * The generic format %A_ denotes a differential polynomial.
 * The generic format %a_ denotes a coefficient.
 */

/*************************************************************************
 :CONSTRUCTEURS
 INIT

 All functions parameterized by a clot are static
 *************************************************************************/

/*
   Initialise A avec C.
   Cette fonction n'est appelee que dans le cas de polynomes qui ne 
	sont pas des coefficients.
*/

static void bap_init_polynom_clot_COEFF (bap_polynom_COEFF A, bap_clot_COEFF C)
{
    A->clot = C;
    bav_init_term (&A->rang_total);
    A->acces = bap_sequential_monom_access;
    A->seq.first = 0;
    A->seq.after = C->size;
    bap_init_indexed_access (&A->ind);
    bap_init_selection (&A->select, (bav_variable)-1, C->ordering);
    A->readonly = false;
}

/* Initializes A to zero.  */

BAP_DLL void bap_init_polynom_COEFF (bap_polynom_COEFF A)
{   struct bav_term T;
    bap_clot_COEFF C;

    bav_init_term (&T);
    C = bap_new_clot_COEFF (&T);
    bap_init_polynom_clot_COEFF (A, C);
}

/* Initializes A to zero as a readonly polynomial.  */

BAP_DLL void bap_init_readonly_polynom_COEFF (bap_polynom_COEFF A)
{
    A->clot = (bap_clot_COEFF)0;
    bav_init_term (&A->rang_total);
    A->acces = bap_sequential_monom_access;
    A->seq.first = 0;
    A->seq.after = 0;
    bap_init_indexed_access (&A->ind);
    bap_init_selection (&A->select, (bav_variable)-1, bav_R_Iordering ());
    A->readonly = true;
}

/* Initializes A to 1.  */

BAP_DLL void bap_init_polynom_one_COEFF (bap_polynom_COEFF A)
{   struct bap_creator_COEFF crea;
    struct bav_term T;
    bap_clot_COEFF C;
    COEFF_t un;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init_set_ui (un, (unsigned long int)1);
    ba0_pull_stack ();

    bav_init_term (&T);
    C = bap_new_clot_COEFF (&T);
    bap_init_polynom_clot_COEFF (A, C);
    bap_begin_creator_COEFF (&crea, A, &T, bap_exact_total_rank, 1);
    bap_write_creator_COEFF (&crea, &T, un);
    bap_close_creator_COEFF (&crea);
    ba0_restore (&M);
}

/* Initializes A to the product c*rg.  */

BAP_DLL void bap_init_polynom_crk_COEFF 
				(bap_polynom_COEFF A, COEFF_t c, bav_rank rg)
{   struct bav_term T;
    struct bap_creator_COEFF crea;
    bap_clot_COEFF C;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&T);
    if (! bav_is_zero_rank (rg))
	bav_set_term_rank (&T, rg);
    ba0_pull_stack ();

    C = bap_new_clot_COEFF (&T);
    bap_init_polynom_clot_COEFF (A, C);
    bap_begin_creator_COEFF (&crea, A, &T, bap_exact_total_rank, 1);
    if (COEFF_sgn (c) != 0 || ! bav_is_zero_rank (rg))
	bap_write_creator_COEFF (&crea, &T, c);

    ba0_restore (&M);
    bap_close_creator_COEFF (&crea);
}

/****************************************************************************
 NEW

 Toutes les fonctions parametrees par une clot sont static.
 ****************************************************************************/

static bap_polynom_COEFF bap_new_polynom_clot_COEFF (bap_clot_COEFF C)
{   bap_polynom_COEFF A;

    A = (bap_polynom_COEFF)ba0_alloc (sizeof (struct bap_polynom_COEFF));
    bap_init_polynom_clot_COEFF (A, C);
    return A;
}

/* Allocates a differential polynomial and initializes it to zero.  */

BAP_DLL bap_polynom_COEFF bap_new_polynom_COEFF (void)
{   struct bav_term T;
    bap_clot_COEFF C;

    bav_init_term (&T);
    C = bap_new_clot_COEFF (&T);
    return bap_new_polynom_clot_COEFF (C);
}

/*
 * Allocates a differential polynomial and initializes it to zero as
 * in the readonly mode.
 */

BAP_DLL bap_polynom_COEFF bap_new_readonly_polynom_COEFF (void)
{   bap_polynom_COEFF A;

    A = (bap_polynom_COEFF)ba0_alloc (sizeof (struct bap_polynom_COEFF));
    bap_init_readonly_polynom_COEFF (A);
    return A;
}

/*
 * Allocates a differential polynomial and initializes it with the
 * product c*rg.
 */

BAP_DLL bap_polynom_COEFF bap_new_polynom_crk_COEFF (COEFF_t c, bav_rank rg)
{   struct bav_term T;
    struct bap_creator_COEFF crea;
    bap_polynom_COEFF A;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&T);
    if (! bav_is_zero_rank (rg))
	bav_set_term_rank (&T, rg);
    ba0_pull_stack ();

    A = bap_new_polynom_COEFF ();
    bap_begin_creator_COEFF (&crea, A, &T, bap_exact_total_rank, 1);
    if (COEFF_sgn (c) != 0 || ! bav_is_zero_rank (rg))
	bap_write_creator_COEFF (&crea, &T, c);

    ba0_restore (&M);

    bap_close_creator_COEFF (&crea);
    return A;
}

/* Allocates a differential polynomial and initializes it to 1.  */

BAP_DLL bap_polynom_COEFF bap_new_polynom_one_COEFF (void)
{   bap_polynom_COEFF A;
    COEFF_t c;
    struct ba0_mark M;
    struct bav_rank rg;

    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init_set_ui (c, 1);
    ba0_pull_stack ();

    rg = bav_constant_rank ();
    A = bap_new_polynom_crk_COEFF (c, &rg);
    ba0_restore (&M);
    return A;
}

/****************************************************************************
 SET
 ****************************************************************************/

/*
   Sets A to zero. This function can be applied to readonly polynomials too.
 */

BAP_DLL void bap_set_polynom_zero_COEFF (bap_polynom_COEFF A)
{
    if (A->acces == bap_sequential_monom_access)
	A->seq.after = A->seq.first = 0;
    else
	A->ind.size = 0;
    bav_set_term_one (&A->rang_total);
}

/* Sets A to 1.  */

BAP_DLL void bap_set_polynom_one_COEFF (bap_polynom_COEFF A)
{   struct bap_creator_COEFF crea;
    struct bav_term T;
    COEFF_t c;
    struct ba0_mark M;

    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init_set_ui (c, 1);
    bav_init_term (&T);
    ba0_pull_stack ();

    bap_begin_creator_COEFF (&crea, A, &T, bap_exact_total_rank, 1);
    bap_write_creator_COEFF (&crea, &T, c);
    bap_close_creator_COEFF (&crea);

    ba0_restore (&M);
} 

/*
 * Sets the product c*rg to A.
 * The clot of A is modified even if the the result is zero.
 */

BAP_DLL void bap_set_polynom_crk_COEFF 
			(bap_polynom_COEFF A, COEFF_t c, bav_rank rg)
{   struct bav_term T;
    struct bap_creator_COEFF crea;
    struct ba0_mark M;

    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&T);
    if (! bav_is_zero_rank (rg))
	bav_set_term_rank (&T, rg);
    ba0_pull_stack ();

    bap_begin_creator_COEFF (&crea, A, &T, bap_exact_total_rank, 1);
    if (COEFF_sgn (c) != 0 && ! bav_is_zero_rank (rg))
	bap_write_creator_COEFF (&crea, &T, c);

    bap_close_creator_COEFF (&crea);

    ba0_restore (&M);
}

/* Sets v^d to A.  */

BAP_DLL void bap_set_polynom_variable_COEFF 
			(bap_polynom_COEFF A, bav_variable v, bav_Idegree d)
{   struct bav_term T;
    struct bap_creator_COEFF crea;
    COEFF_t c;
    struct ba0_mark M;

    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&T);
    bav_set_term_variable (&T, v, d);
    COEFF_init_set_ui (c, 1);
    ba0_pull_stack ();

    bap_begin_creator_COEFF (&crea, A, &T, bap_exact_total_rank, 1);
    if (! bav_is_one_term (&T))
	bap_write_creator_COEFF (&crea, &T, c);
    bap_close_creator_COEFF (&crea);

    ba0_restore (&M);
}

/* Sets T to A.  */

BAP_DLL void bap_set_polynom_term_COEFF (bap_polynom_COEFF A, bav_term T)
{   struct bap_creator_COEFF crea;
    COEFF_t c;
    struct ba0_mark M;

    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init_set_si (c, 1);
    ba0_pull_stack ();

    bap_begin_creator_COEFF (&crea, A, T, bap_exact_total_rank, 1);
    bap_write_creator_COEFF (&crea, T, c);
    bap_close_creator_COEFF (&crea);
    ba0_restore (&M);
}

/* Sets the monomial c\,T to A. The coefficient c may be zero.  */

BAP_DLL void bap_set_polynom_monom_COEFF 
				(bap_polynom_COEFF A, COEFF_t c, bav_term T)
{   struct bap_creator_COEFF crea;

    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    if (COEFF_sgn (c) == 0)
	bap_set_polynom_zero_COEFF (A);
    else
    {   bap_begin_creator_COEFF (&crea, A, T, bap_exact_total_rank, 1);
	bap_write_creator_COEFF (&crea, T, c);
	bap_close_creator_COEFF (&crea);
    }
}

/*
 * Sets B to A. The data is completely copied.
 * Setting A to A raises an error. 
 */

BAP_DLL void bap_set_polynom_COEFF (bap_polynom_COEFF A, bap_polynom_COEFF B)
{   struct bap_creator_COEFF crea;
    struct bap_itermon_COEFF iter;
    struct bap_polynom_COEFF C;
    struct bav_term T;
    COEFF_t *c;
    ba0_int_p nbmon;
    struct ba0_mark M;

    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    if (bap_is_zero_polynom_COEFF (B))
	bap_set_polynom_zero_COEFF (A);
    else if (! bap_are_disjoint_polynom_COEFF (A, B))
    {	if (A == B)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	else if (B->acces == bap_sequential_monom_access)
	{   A->acces = bap_sequential_monom_access;
	    A->seq = B->seq;
	    bav_set_term (&A->rang_total, &B->rang_total);
	    bap_set_selection (&A->select, &B->select);
	} else
	{   
	    ba0_push_another_stack ();
	    ba0_record (&M);

	    bap_init_polynom_COEFF (&C);
	    bap_set_polynom_COEFF (&C, B);

	    ba0_pull_stack ();

	    bap_set_polynom_COEFF (A, &C);

	    ba0_restore (&M);
	}
    } else   
    {
	nbmon = bap_nbmon_polynom_COEFF (B) - A->clot->alloc;
	nbmon = nbmon > 0 ? nbmon : 0;
    
	ba0_push_another_stack ();
	ba0_record (&M);
	bav_init_term (&T);
	bav_realloc_term (&T, B->rang_total.size);
	ba0_pull_stack ();
    
	bap_begin_creator_COEFF 
		(&crea, A, &B->rang_total, bap_exact_total_rank, nbmon);
    
	if (bap_is_write_allable_creator_COEFF (&crea, B))
	    bap_write_all_creator_COEFF (&crea, B);
	else
	{   bap_begin_itermon_COEFF (&iter, B);
	    while (! bap_outof_itermon_COEFF (&iter))
	    {   c = bap_coeff_itermon_COEFF (&iter);
		bap_term_itermon_COEFF (&T, &iter);
		bap_write_creator_COEFF (&crea, &T, *c);
		bap_next_itermon_COEFF (&iter);
	    }
	}
	bap_close_creator_COEFF (&crea);
	ba0_restore (&M);
    }
}

/* Sets A to R in the readonly mode.  Some datas are shared.  */

BAP_DLL void bap_set_readonly_polynom_COEFF 
				(bap_polynom_COEFF R, bap_polynom_COEFF A)
{
    if (R == A)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    R->clot = A->clot;
    bav_set_term (&R->rang_total, &A->rang_total);
    if (A->acces == bap_sequential_monom_access)
    {	R->acces = bap_sequential_monom_access;
	R->seq = A->seq;
    } else
    {	struct bap_creator_index crea;
	struct bap_iterator_index iter;
	ba0_int_p i, nbmon;

	R->acces = bap_indexed_monom_access;
	nbmon = bap_nbmon_polynom_COEFF (A);
	bap_realloc_indexed_access (&R->ind, nbmon);
	bap_begin_creator_index (&crea, &R->ind);
	bap_begin_iterator_index (&iter, &A->ind);
	for (i = 0; i < nbmon; i++)
	    bap_write_creator_index (&crea, bap_read_iterator_index (&iter));
	bap_close_creator_index (&crea);
    }
    bap_set_selection (&R->select, &A->select);
    R->readonly = true;
}

/*
Sets the orderings of A and of its clot to r.
Take care.

void bap_set_ordering_polynom_and_clot_COEFF 
				(bap_polynom_COEFF A, bav_Iordering r)
{
    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    bap_change_ordering_selection (&A->select, r);
    bap_change_ordering_clot_COEFF (A->clot, r);
}
*/

/*
   PREDICATES
 */

/* Returns true if A is zero.  */

BAP_DLL bool bap_is_zero_polynom_COEFF (bap_polynom_COEFF A)
{
    return bap_nbmon_polynom_COEFF (A) == 0;
}

/* Returns true if A is one.  */ 

BAP_DLL bool bap_is_one_polynom_COEFF (bap_polynom_COEFF A)
{   struct bap_itermon_COEFF iter;
    COEFF_t *lc;

    if (! bav_is_one_term (&A->rang_total))
	return false;
    bap_begin_itermon_COEFF (&iter, A);
    lc = bap_coeff_itermon_COEFF (&iter);
    return COEFF_is_one (*lc);
}

/* Returns true if A is a constant.  */

BAP_DLL bool bap_is_numeric_polynom_COEFF (bap_polynom_COEFF A)
{
    return bap_is_zero_polynom_COEFF (A) || bav_is_one_term (&A->rang_total);
}

/*
 * Returns true if A does not depend on dependent variables (i.e. does
 * not involve derivatives).
 */

BAP_DLL bool bap_is_independent_polynom_COEFF (bap_polynom_COEFF A)
{   bool b;
    ba0_int_p i;
    enum bav_typeof_symbol type;

    if (bap_is_zero_polynom_COEFF (A))
	b = true;
    else
    {	b = true;
	for (i = 0; b && i < A->rang_total.size; i++)
	{   type = bav_symbol_type_variable (A->rang_total.rg [i].var);
	    b = type == bav_independent_symbol;
	}
    }
    return b;
}

/*
 * Returns true if A is differential polynomial linear, homogeneous,
 * with constant coefficients and depending only on derivatives of
 * only one differential indeterminate.
 */

/*
bool bap_is_lhoc_polynom_COEFF (bap_polynom_COEFF A)
{   struct bap_itermon_COEFF iter;
    struct bav_term T;
    struct bav_rank rg;
    bav_symbol s;
    bool b;
    ba0_int_p i;
    struct ba0_mark M;

    if (bap_is_numeric_polynom_COEFF (A))
	return bap_is_zero_polynom_COEFF (A);

    rg = A->rang_total.rg [0];
    if (rg.var->root->type != bav_dependent_symbol || rg.deg != 1)
	return false;

    s = rg.var->root;

    for (i = 1; i < A->rang_total.size; i++)
    {   rg = A->rang_total.rg [i];
	if (rg.var->root != s || rg.deg != 1)
	    return false;
    }
	
    ba0_record (&M);
    bav_init_term (&T);
    bap_begin_itermon_COEFF (&iter, A);
    b = true;
    while (b && ! bap_outof_itermon_COEFF (&iter))
    {	bap_term_itermon_COEFF (&T, &iter);
	if (T.size == 0 || T.rg [0].var->root != s || T.rg [0].deg != 1)
	    b = false;
	else
	{   for (i = 1; b && i < T.size; i++)
		b = T.rg [i].var->root->type == bav_constant_symbol;
	}
	bap_next_itermon_COEFF (&iter);
    }
    ba0_restore (&M);
    return b;
}
*/

/* Returns true if A only involves one variable.  */

BAP_DLL bool bap_is_univariate_polynom_COEFF (bap_polynom_COEFF A)
{
    return A->rang_total.size <= 1;
}

/* Returns true if A is reduced to a mere variable.  */

BAP_DLL bool bap_is_variable_polynom_COEFF (bap_polynom_COEFF A)
{
    return bav_total_degree_term (&A->rang_total) == 1 &&
					bap_nbmon_polynom_COEFF (A) == 1;
}

/*
 * returns true if A = variable - something
 */

BAP_DLL bool bap_is_solved_polynom_COEFF (bap_polynom_COEFF A)
{   struct bap_polynom_COEFF init;
    struct ba0_mark M;
    bool b;

    if (bap_is_numeric_polynom_COEFF (A))
	return false;

    if (bap_leading_degree_polynom_COEFF (A) != 1)
	return false;

    ba0_record (&M);
    bap_init_readonly_polynom_COEFF (&init);
    bap_initial_polynom_COEFF (&init, A);
    b = bap_is_numeric_polynom_COEFF (&init);
    ba0_restore (&M);

    return b;
}

/*
 * Returns true if A has the form v - p where v is a derivative
 * and p is a possibly zero polynomial which does not involve any derivative.
 */

BAP_DLL bool bap_is_derivative_minus_independent_polynom_COEFF 
							(bap_polynom_COEFF A)
{   ba0_int_p i;

    if (! bap_is_solved_polynom_COEFF (A))
	return false;

    if (bav_symbol_type_variable (A->rang_total.rg [0].var)
						!= bav_dependent_symbol)
	return false;
    for (i = 1; i < A->rang_total.size; i++)
	if (bav_symbol_type_variable (A->rang_total.rg [i].var) 
						== bav_dependent_symbol)
	    return false;
    return true;
}

BAP_DLL bool bap_is_rank_minus_monom_polynom_COEFF (bap_polynom_COEFF A)
{   struct bap_polynom_COEFF init;

    if (bap_nbmon_polynom_COEFF (A) > 2)
	return false;
    bap_init_readonly_polynom_COEFF (&init);
    bap_initial_polynom_COEFF (&init, A);
    return bap_is_one_polynom_COEFF (&init);
}

/* Returns true if A depends on v.  */

BAP_DLL bool bap_depend_polynom_COEFF (bap_polynom_COEFF A, bav_variable v)
{   ba0_int_p i;

    for (i = 0; i < A->rang_total.size; i++)
	if (A->rang_total.rg [i].var == v)
	    return true;
    return false;
}

/*
   It is assumed that T is sorted in increasing order
*/

BAP_DLL bool bap_depend_only_polynom_COEFF 
				(bap_polynom_COEFF A, bav_tableof_variable T)
{   bav_Inumber n = 0, m = 0;
    ba0_int_p i, j;

    i = 0;
    if (i < A->rang_total.size)
	n = bav_variable_number (A->rang_total.rg [i].var);
    j = T->size - 1;
    if (j >= 0)
	m = bav_variable_number (T->tab [j]);
    while (i < A->rang_total.size && j >= 0)
    {	if (n > m)
	    return false;
	else if (n < m)
	{   j -= 1;
	    m = bav_variable_number (T->tab [j]);
	} else
	{   i += 1;
	    j -= 1;
	    n = bav_variable_number (A->rang_total.rg [i].var);
	    m = bav_variable_number (T->tab [j]);
	}
    }
    return i == A->rang_total.size;
}

/***************************************************************************
 FONCTIONS DE :COMPARAISON
 ***************************************************************************/

/*
 * Returns ba0_lt if A < B, ba0_gt if A > B,
 * ba0_eq if A = B and ba0_equiv if A and B have the
 * same terms but not the same coefficients.
 */

BAP_DLL enum ba0_compare_code bap_compare_polynom_COEFF 
				(bap_polynom_COEFF A, bap_polynom_COEFF B)
{   struct bap_itermon_COEFF iterA, iterB;
    enum ba0_compare_code code, resultat = ba0_eq;
    struct bav_term TA, TB;
    struct ba0_mark M;

    bap__check_compatible_COEFF (A, B);

    if (A == B)
	return resultat;

    ba0_record (&M);
    bav_init_term (&TA);
    bav_init_term (&TB);
    bap_begin_itermon_COEFF (&iterA, A);
    bap_begin_itermon_COEFF (&iterB, B);
    while (! bap_outof_itermon_COEFF (&iterA) && 
					! bap_outof_itermon_COEFF (&iterB))
    {	bap_term_itermon_COEFF (&TA, &iterA);
	bap_term_itermon_COEFF (&TB, &iterB);
	code = bav_compare_term (&TA, &TB);
	if (code == ba0_lt || code == ba0_gt)
	{   resultat = code;
	    break;
	} 
	if (resultat == ba0_eq && 
			COEFF_cmp (*bap_coeff_itermon_COEFF (&iterA), 
				   *bap_coeff_itermon_COEFF (&iterB)) != 0)
	    resultat = ba0_equiv;
	bap_next_itermon_COEFF (&iterA);
	bap_next_itermon_COEFF (&iterB);
    }
    if (resultat == ba0_eq || resultat == ba0_equiv)
    {	if (! bap_outof_itermon_COEFF (&iterA))
	    resultat = ba0_gt;
	else if (! bap_outof_itermon_COEFF (&iterB))
	    resultat = ba0_lt;
    }
    ba0_restore (&M);
    return resultat;
}

/* Returns true if A = B.  */

BAP_DLL bool bap_equal_polynom_COEFF (bap_polynom_COEFF A, bap_polynom_COEFF B)
{   struct bap_itermon_COEFF iterA, iterB;
    struct bav_term TA, TB;
    struct ba0_mark M;

    bap__check_compatible_COEFF (A, B);

    if (A == B)
	return true;
    if (bap_nbmon_polynom_COEFF (A) != bap_nbmon_polynom_COEFF (B))
	return false;
    if (! bav_equal_term (&A->rang_total, &B->rang_total))
	return false;

    ba0_record (&M);
    bav_init_term (&TA);
    bav_init_term (&TB);
    bap_begin_itermon_COEFF (&iterA, A);
    bap_begin_itermon_COEFF (&iterB, B);
    while (! bap_outof_itermon_COEFF (&iterA))
    {	bap_term_itermon_COEFF (&TA, &iterA);
	bap_term_itermon_COEFF (&TB, &iterB);
	if (! bav_equal_term (&TA, &TB))
	{   ba0_restore (&M);
	    return false;
	}
	if (COEFF_cmp (*bap_coeff_itermon_COEFF (&iterA), 
				   *bap_coeff_itermon_COEFF (&iterB)) != 0)
	{   ba0_restore (&M);
	    return false;
	}
	bap_next_itermon_COEFF (&iterA);
	bap_next_itermon_COEFF (&iterB);
    }
    ba0_restore (&M);
    return true;
}

/* Returns true if P has greater rank than Q.  */

BAP_DLL bool bap_gt_rank_polynom_COEFF (void* P, void* Q)
{   bap_polynom_COEFF A = (bap_polynom_COEFF)P;
    bap_polynom_COEFF B = (bap_polynom_COEFF)Q;
    struct bav_rank Arg, Brg;

    Arg = bap_rank_polynom_COEFF (A);
    Brg = bap_rank_polynom_COEFF (B);
    return bav_gt_rank (&Arg, &Brg);
}

/* Returns true if P has lower rank than Q.  */

BAP_DLL bool bap_lt_rank_polynom_COEFF (void* P, void* Q)
{   bap_polynom_COEFF A = (bap_polynom_COEFF)P;
    bap_polynom_COEFF B = (bap_polynom_COEFF)Q;
    struct bav_rank Arg, Brg;

    Arg = bap_rank_polynom_COEFF (A);
    Brg = bap_rank_polynom_COEFF (B);

    return bav_lt_rank (&Arg, &Brg);
}

/* Returns true if P has the same rank as Q.  */

BAP_DLL bool bap_equal_rank_polynom_COEFF 
				(bap_polynom_COEFF P, bap_polynom_COEFF Q)
{   struct bav_rank Prg, Qrg;

    Prg = bap_rank_polynom_COEFF (P);
    Qrg = bap_rank_polynom_COEFF (Q);

    return bav_equal_rank (&Prg, &Qrg);
}

/*
 * Returns true if A and B are disjoint in the sense they do not
 * share any piece of data.
 */

BAP_DLL bool bap_are_disjoint_polynom_COEFF 
				(bap_polynom_COEFF A, bap_polynom_COEFF B)
{
    return A->clot != B->clot;
}

/* Returns the field sens of the clot of A.  */

BAP_DLL enum bap_typeof_monom_sequence 
		bap_monom_sequence_polynom_COEFF (bap_polynom_COEFF A)
{
    return A->clot->sens;
}

/* Sets to true the field hack of all the variables occuring in A.  */

BAP_DLL void bap_mark_indets_polynom_COEFF (bap_polynom_COEFF A)
{   ba0_int_p i;

    for (i = 0; i < A->rang_total.size; i++)
	A->rang_total.rg [i].var->hack = true;
}

BAP_DLL void bap_involved_parameters_polynom_COEFF
        (bav_tableof_parameter R, bap_polynom_COEFF A, bav_tableof_parameter P)
{   bav_variable v;
    ba0_int_p i, j;

    ba0_realloc_table ((ba0_table)R, P->size);
    for (i = 0; i < A->rang_total.size; i++)
    {   v = A->rang_total.rg [i].var;
        if (bav_is_a_parameter (v->root, &j, P) &&
                                        ! bav_is_a_parameter (v->root, 0, R))
        {   ba0_realloc_table ((ba0_table)R, R->size+1);
	    R->tab [R->size] = P->tab [j];
            R->size += 1;
        }
    }
}

/*
 * Sets the field hack of the variables occuring in A to the
 * degree of A in that variable (provided that the existing value
 * of that field is less than this degree).
 */

BAP_DLL void bap_mark_ranks_polynom_COEFF (bap_polynom_COEFF A)
{   ba0_int_p i;

    for (i = 0; i < A->rang_total.size; i++)
	A->rang_total.rg [i].var->hack = A->rang_total.rg [i].deg;
}

/*
  Internal function. Computes the field rang_total of A.
*/

BAP_DLL void bap_set_total_rank_polynom_COEFF (bap_polynom_COEFF A)
{   struct bap_itermon_COEFF iter;
    struct bav_term T;
    bav_variable v;
    ba0_int_p i;
    struct ba0_mark M;

    ba0_record (&M);
    bav_init_term (&T);
    bav_R_mark_variables (0);
    bap_begin_itermon_COEFF (&iter, A);
    while (! bap_outof_itermon_COEFF (&iter))
    {	bap_term_itermon_COEFF (&T, &iter);
	for (i = 0; i < T.size; i++)
	{   v = T.rg [i].var;
	    if (v->hack < T.rg [i].deg)
		v->hack = T.rg [i].deg;
	}
	bap_next_itermon_COEFF (&iter);
    }
    ba0_restore (&M);

    bav_marked_ranks_term (&A->rang_total);
    bav_sort_term (&A->rang_total);
}
     
/* Reverts the clot of A.  */

BAP_DLL void bap_reverse_polynom_COEFF (bap_polynom_COEFF A)
{
    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    bap_reverse_clot_COEFF (A->clot);
}

/*************************************************************************
 :REORDONNE
 *************************************************************************/

struct quicksort_data {
    struct bap_iterator_index l;
    struct bap_iterator_index r;
    struct bap_itermon_COEFF i;
    struct bap_itermon_COEFF j;
    struct bap_itermon_COEFF k;
    struct bav_term Ti;
    struct bav_term Tj;
    struct bav_term Tk;
    struct bav_term pivot;
    unsigned ba0_int_p zi;
    unsigned ba0_int_p zj;
    unsigned ba0_int_p zk;
};

static void bap_init_quicksort_data 
			(struct quicksort_data* qs, bap_polynom_COEFF R)
{
    bap_begin_iterator_index (&qs->l, &R->ind);
    bap_end_iterator_index (&qs->r, &R->ind);

    bap_begin_itermon_COEFF (&qs->i, R);
    bap_begin_itermon_COEFF (&qs->j, R);
    bap_begin_itermon_COEFF (&qs->k, R);

    bav_init_term (&qs->Ti);
    bav_init_term (&qs->Tj);
    bav_init_term (&qs->Tk);
    bav_init_term (&qs->pivot);
    qs->zi = qs->zj = qs->zk = 0;
}

/*
static ba0_int_p qs_depth;
static ba0_int_p qs_depth_max;
static ba0_int_p qs_n = 0;
*/

/*
   Cette variante suppose qu'il n'y a pas de doublons dans les
	\'el\'ements \`a trier (ce qui est vrai pour des termes).
   Le polyn\^ome est tri\'e par ordre d\'ecroissant.
   L'iterateur sert a se deplacer dans le polynome.
   Les struct bav_term sont des variables de travail.
*/

static void quicksort_COEFF 
			(ba0_int_p l, ba0_int_p r, struct quicksort_data* qs)
{   ba0_int_p i, j, k;
    enum ba0_compare_code code;
/*
 * Readonly static data
 */
    static ba0_int_p alpha [] = { 1, 2, 2, 3, 4, 1, 3 }; 
    /*                            -  -  -  -  -  -  - 	*/
    static ba0_int_p beta []  = { 1, 3, 1, 4, 3, 2, 2 }; 
/*
    ++qs->n;
    if (++qs->depth > qs->depth_max)
        qs->depth_max = qs->depth;
*/
    if (r - l > 8)
    {
        i = (alpha[qs->zi] * (l+2) + beta[qs->zi] * (r-2)) / 
						(alpha[qs->zi] + beta[qs->zi]);
        qs->zi = (qs->zi + 1) % 7;
        j = (alpha[qs->zj] * (l+1) + beta[qs->zj] * (i-1)) / 
						(alpha[qs->zj] + beta[qs->zj]);
        qs->zj = (qs->zj + 3) % 7;
        k = (alpha[qs->zk] * (i+1) + beta[qs->zk] * (r-1)) / 
						(alpha[qs->zk] + beta[qs->zk]);
        qs->zk = (qs->zk + 5) % 7;
	if (i == j || i == k || j == k)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	bap_goto_itermon_COEFF (&qs->i, i);
	bap_goto_itermon_COEFF (&qs->j, j);
	bap_goto_itermon_COEFF (&qs->k, k);

	bap_term_itermon_COEFF (&qs->Ti, &qs->i);
	bap_term_itermon_COEFF (&qs->Tj, &qs->j);
	bap_term_itermon_COEFF (&qs->Tk, &qs->k);
/*
   On souhaite que le pivot soit le median de T [i], T [j], T [k].
   On doit echanger T [l] avec le median et T [r] avec le max.
*/
	code = bav_compare_term (&qs->Ti, &qs->Tj);
	if (code == ba0_gt)
	{   code = bav_compare_term (&qs->Tj, &qs->Tk);
	    if (code == ba0_gt)
	    {	bap_swapindex_iterator_index (&qs->l, &qs->j.iter_ix);
		bap_swapindex_iterator_index (&qs->r, &qs->i.iter_ix);
/* T [i] > T [j] > T [k] */
		bav_set_term (&qs->pivot, &qs->Tj);
	    } else
	    {	code = bav_compare_term (&qs->Ti, &qs->Tk);
		if (code == ba0_gt)
		{   bap_swapindex_iterator_index (&qs->l, &qs->k.iter_ix);
		    bap_swapindex_iterator_index (&qs->r, &qs->i.iter_ix);
/* T [i] > T [k] > T [j] */
		    bav_set_term (&qs->pivot, &qs->Tk);
		} else
		{   bap_swapindex_iterator_index (&qs->l, &qs->i.iter_ix);
		    bap_swapindex_iterator_index (&qs->r, &qs->k.iter_ix);
/* T [k] > T [i] > T [j] */
		    bav_set_term (&qs->pivot, &qs->Ti);
		}
	    }
	} else
	{   code = bav_compare_term (&qs->Tk, &qs->Tj);
	    if (code == ba0_gt)
	    {   bap_swapindex_iterator_index (&qs->l, &qs->j.iter_ix);
		bap_swapindex_iterator_index (&qs->r, &qs->k.iter_ix);
/* T [k] > T [j] > T [i] */
		bav_set_term (&qs->pivot, &qs->Tj);
	    } else
	    {	code = bav_compare_term (&qs->Tk, &qs->Ti);
		if (code == ba0_gt)
		{   bap_swapindex_iterator_index (&qs->l, &qs->k.iter_ix);
		    bap_swapindex_iterator_index (&qs->r, &qs->j.iter_ix);
/* T [j] > T [k] > T [i] */
		    bav_set_term (&qs->pivot, &qs->Tk);
		} else
		{   bap_swapindex_iterator_index (&qs->l, &qs->i.iter_ix);
		    bap_swapindex_iterator_index (&qs->r, &qs->j.iter_ix);
/* T [j] > T [i] > T [k] */
		    bav_set_term (&qs->pivot, &qs->Ti);
		}
	    }
	}
/*
   Invariants (sauf avant le 1er tour).
   i, j in [l, r]
   T [i] < pivot (sauf 1er tour ou on a egalite)
   T [j] > pivot
   Les T [k < i] sont > pivot
   Les T [k > j] sont < pivot
*/
	i = l;
	j = r;
	bap_goto_itermon_COEFF (&qs->i, i);
	bap_goto_itermon_COEFF (&qs->j, j);
	do
	{   bap_swapindex_iterator_index (&qs->i.iter_ix, &qs->j.iter_ix);
	    do
	    {	i += 1;
/*
   Dans la mesure ou l'acces est indexe, il n'y aurait aucun avantage
   a utiliser un next_itermon au lieu d'un goto_itermon.
*/
		bap_next_itermon_COEFF (&qs->i);
		bap_term_itermon_COEFF (&qs->Ti, &qs->i);
		code = bav_compare_term (&qs->Ti, &qs->pivot);
	    } while (code == ba0_gt);
	    do
	    {	j -= 1;
		bap_prev_itermon_COEFF (&qs->j);
		bap_term_itermon_COEFF (&qs->Tj, &qs->j);
		code = bav_compare_term (&qs->Tj, &qs->pivot);
	    } while (code == ba0_lt);
	} while (i < j);
	bap_swapindex_iterator_index (&qs->i.iter_ix, &qs->r);
	bap_goto_iterator_index (&qs->r, i - 1);
	quicksort_COEFF (l, i - 1, qs);
	bap_goto_iterator_index (&qs->l, i + 1);
	bap_goto_iterator_index (&qs->r, r);
	quicksort_COEFF (i + 1, r, qs);
    } else
    {	bool sorted;
	bav_term cour, suiv;
	struct bap_iterator_index i_cour;

	sorted = false;
	cour = &qs->Ti;
	suiv = &qs->Tj;
	for (i = r - 1; ! sorted && i >= l; i--)
	{   sorted = true;
	    bap_goto_itermon_COEFF (&qs->j, l);
	    bap_term_itermon_COEFF (suiv, &qs->j);
	    for (j = l; j <= i; j++)
	    {	BA0_SWAP (bav_term, cour, suiv);
		bap_set_iterator_index (&i_cour, &qs->j.iter_ix);
		bap_next_itermon_COEFF (&qs->j);
		bap_term_itermon_COEFF (suiv, &qs->j);
		code = bav_compare_term (cour, suiv);
		if (code == ba0_lt)
		{   BA0_SWAP (bav_term, cour, suiv);
		    bap_swapindex_iterator_index (&i_cour, &qs->j.iter_ix);
		    sorted = false;
		}
	    }
	}
    }
/*
    qs->depth--;
*/
}

/*
 * Stores in R the polynomial A, sorted w.r.t. the current ordering.
 * The polynomial R is readonly and has indexed access.
 */

BAP_DLL void bap_sort_polynom_COEFF (bap_polynom_COEFF R, bap_polynom_COEFF A)
{   struct bap_creator_index crea;
    struct bap_iterator_index iter;
    struct quicksort_data qs;
    ba0_int_p i, offset, nbmonA;
    struct ba0_mark M;

    nbmonA = bap_nbmon_polynom_COEFF (A);
    if (nbmonA == 0)
    {	bap_set_polynom_zero_COEFF (R);
	return;
    }

    if (A == R)
    {	if (A->acces == bap_sequential_monom_access)
	{   offset = A->seq.first;
	    A->acces = bap_indexed_monom_access;
	    bap_realloc_indexed_access (&A->ind, nbmonA);
	    bap_begin_creator_index (&crea, &A->ind);
	    for (i = 0; i < nbmonA; i++)
		bap_write_creator_index (&crea, offset + i);
	    bap_close_creator_index (&crea);
	}
    } else
    {	R->clot = A->clot;
	R->acces = bap_indexed_monom_access;
	bap_realloc_indexed_access (&R->ind, nbmonA);
	bap_begin_creator_index (&crea, &R->ind);
	if (A->acces == bap_sequential_monom_access)
	{   offset = A->seq.first;
	    for (i = 0; i < nbmonA; i++)
		bap_write_creator_index (&crea, offset + i);
	} else
	{   bap_begin_iterator_index (&iter, &A->ind);
	    for (i = 0; i < nbmonA; i++)
		bap_write_creator_index 
				(&crea, bap_read_iterator_index (&iter));
	}
	bap_close_creator_index (&crea);
	bap_set_selection (&R->select, &A->select);
	bav_set_term (&R->rang_total, &A->rang_total);
    }
    bav_sort_term (&R->rang_total);
    if (bap_ordering_selection (&R->select) != bav_R_Iordering ())
	bap_push_selection (&R->select, (bav_variable)-1, bav_R_Iordering ());

    ba0_record (&M);
    bap_init_quicksort_data (&qs, R);
    quicksort_COEFF (0, nbmonA - 1, &qs);
/*
    {   int log2 = bap_ceil_log2 (nbmonA / 5);
        ba0_printf ("quicksort optimal = %d, reel = %d\n", log2, qs_depth_max);
    }
*/
    ba0_restore (&M);

    switch (bap_monom_sequence_polynom_COEFF (A))
    {	case bap_decreasing_monom_sequence:
	    break;
	case bap_increasing_monom_sequence:
	    bap_reverse_indexed_access (&R->ind);
	    break;
	default:
	    BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    }
    R->readonly = true;
}

/*
 * Sorts A by decreasing order w.r.t. the current ordering.
 * Sequential access only.
 * In the clot of A, only the monomials of A are concerned (take care).
 * 
 * This function has many different usages.
 * 
 * If one wishes to sort the result of a computation, it is a definitive
 * transform. If one wishes to sort the given data before a computation,
 * an invertible transform is performed, followed by the inverse transform
 * at the end of the computation.
 * The formal transform applies on a result which is known to be correctly
 * sorted w.r.t. the current ordering. Example. The content of a polynomial
 * w.r.t. some variable x. One may set x has maximal variable in some
 * new ordering. Since x does appear in the content, this polynomial is
 * already sorted w.r.t. the old ordering. The only transform to do is
 * to change the ordering number. This is what the formal transform does.
 */

BAP_DLL void bap_physort_polynom_COEFF 
			(bap_polynom_COEFF A, enum bap_typeof_transform type)
{
    if (A->acces == bap_indexed_monom_access)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    bap_push_clot_parameters_stack 
	(bap_current_typeof_termanager (), bap_decreasing_monom_sequence);
    bap_sort_clot_COEFF (A->clot, A->seq.first, A->seq.after);
    bap_pull_clot_parameters_stack ();
    bav_sort_term (&A->rang_total);
    switch (type)
    {	case bap_definitive_transform:
	    if (bap_variable_selection (&A->select) == (bav_variable)-1)
		bap_change_ordering_selection (&A->select, bav_R_Iordering ());
	    else
		bap_push_selection 
			(&A->select, (bav_variable)-1, bav_R_Iordering ());
	    break;
	case bap_invertible_transform:
	    bap_push_selection 
			(&A->select, (bav_variable)-1, bav_R_Iordering ());
	    break;
	case bap_inverse_transform:
	    bap_pull_selection (&A->select);
	    break;
	case bap_formal_transform:
	    bap_change_ordering_selection (&A->select, bav_R_Iordering ());
	    bap_change_ordering_clot_COEFF (A->clot, bav_R_Iordering ());
	    break;
    }
}

/* Returns the number of monomials of A.  */

BAP_DLL ba0_int_p bap_nbmon_polynom_COEFF (bap_polynom_COEFF A)
{
    if (A->acces == bap_sequential_monom_access)
	return A->seq.after - A->seq.first;
    else
	return A->ind.size;
}


/* Returns the leader of A.  */

BAP_DLL bav_variable bap_leader_polynom_COEFF (bap_polynom_COEFF A)
{
    if (bap_is_numeric_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRCST);
    return bav_leader_term (&A->rang_total);
}


/* Returns the rank of A.  */

BAP_DLL struct bav_rank bap_rank_polynom_COEFF (bap_polynom_COEFF A)
{   struct bav_rank rg;

    if (bap_is_zero_polynom_COEFF (A))
	rg = bav_zero_rank ();
    else if (bav_is_one_term (&A->rang_total))
	rg = bav_constant_rank ();
    else
	rg = bav_rank_term (&A->rang_total);
    return rg;
}

/* Returns the degree of A w.r.t. its leader.  */

BAP_DLL bav_Idegree bap_leading_degree_polynom_COEFF (bap_polynom_COEFF A)
{
    if (bap_is_numeric_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRCST);
    return bav_leading_degree_term (&A->rang_total);
}

BAP_DLL bav_Iorder bap_total_order_polynom_COEFF (bap_polynom_COEFF A)
{
    return bav_total_order_term (&A->rang_total);
}

/* Sets the gcd of the terms of A to T.  */

BAP_DLL void bap_minimal_total_rank_polynom_COEFF 
					(bav_term T, bap_polynom_COEFF A)
{   struct bap_itermon_COEFF iter;
    struct bav_term U;
    bav_variable v;
    ba0_int_p i, j;
    struct ba0_mark M;

    if (bap_is_zero_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRNUL);

    ba0_push_another_stack ();
    ba0_record (&M);

    bav_init_term (&U);

    switch (bap_monom_sequence_polynom_COEFF (A))
    {	case bap_decreasing_monom_sequence:
	    bap_end_itermon_COEFF (&iter, A);
	    bap_term_itermon_COEFF (&U, &iter);
	    if (bav_is_one_term (&U))
		goto fin;
	    break;
	case bap_increasing_monom_sequence:
	    bap_begin_itermon_COEFF (&iter, A);
	    bap_term_itermon_COEFF (&U, &iter);
	    if (bav_is_one_term (&U))
		goto fin;
	    break;
	default:
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
    }

    for (i = 0; i < A->rang_total.size; i++)
	A->rang_total.rg [i].var->hack = A->rang_total.rg [i].deg;

    bap_begin_itermon_COEFF (&iter, A);
    while (! bap_outof_itermon_COEFF (&iter))
    {	bap_term_itermon_COEFF (&U, &iter);
	i = 0;
	for (j = 0; j < U.size; j++)
	{   while (A->rang_total.rg [i].var != U.rg [j].var)
	    {	A->rang_total.rg [i].var->hack = 0;
		i++;
	    }
	    if (A->rang_total.rg [i].var->hack > U.rg [j].deg)
		A->rang_total.rg [i].var->hack = U.rg [j].deg;
	    i++;
	}
	while (i < A->rang_total.size)
	    A->rang_total.rg [i++].var->hack = 0;
	bap_next_itermon_COEFF (&iter);
    }

    U.size = 0;
    bav_realloc_term (&U, A->rang_total.size);
    for (i = j = 0; i < A->rang_total.size; i++)
    {   v = A->rang_total.rg [i].var;
	if (v->hack > 0)
	{   U.rg [j].var = v;
	    U.rg [j].deg = v->hack;
	    j++;
	}
    }
    U.size = j;
fin:
    bav_sort_term (&U);
    ba0_pull_stack ();
    bav_set_term (T, &U);
    ba0_restore (&M);
}

BAP_DLL void bap_leading_term_polynom_COEFF (bav_term T, bap_polynom_COEFF A)
{   struct bap_itermon_COEFF iter;

    if (bap_is_zero_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRNUL);

    bap_begin_itermon_COEFF (&iter, A);
    bap_term_itermon_COEFF (T, &iter);
}

/* 
 * Returns the address of the numerical coefficient of the leading 
 * monomial of A.
 */

BAP_DLL COEFF_t* bap_numeric_initial_polynom_COEFF (bap_polynom_COEFF A)
{   struct bap_itermon_COEFF iter;
    COEFF_t *c;

    if (bap_is_zero_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRNUL);

    bap_begin_itermon_COEFF (&iter, A);
    c = bap_coeff_itermon_COEFF (&iter);

    return c;
}

/*
 * Sets the initial of A to {\em initial} and the reductum of A
 * to {\em reductum} as readonly polynomials. These parameters may be zero.
 */

BAP_DLL void bap_initial_and_reductum_polynom_COEFF 
	(bap_polynom_COEFF initial, bap_polynom_COEFF reductum, 
	 bap_polynom_COEFF A)
{   struct bap_itercoeff_COEFF iter;

    if (bap_is_numeric_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRCST);
    if ((A == initial && reductum != BAP_NOT_A_POLYNOM_COEFF) || 
					(A == initial && A == reductum))
	BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    bap_begin_itercoeff_COEFF (&iter, A, bap_leader_polynom_COEFF (A));
    if (initial != BAP_NOT_A_POLYNOM_COEFF)
	bap_coeff_itercoeff_COEFF (initial, &iter);
    if (reductum != BAP_NOT_A_POLYNOM_COEFF)
    {	bap_next_itermon_COEFF (&iter.fin);
	bap_reductum_itermon_COEFF (&iter.fin, reductum);
    }
}

/* Sets the initial of A to R as a readonly polynomial.  */

BAP_DLL void bap_initial_polynom_COEFF 
				(bap_polynom_COEFF R, bap_polynom_COEFF A)
{
    bap_initial_and_reductum_polynom_COEFF (R, BAP_NOT_A_POLYNOM_COEFF, A);
}

/* Sets the reductum of A to R as a readonly polynomial.  */

BAP_DLL void bap_reductum_polynom_COEFF 
				(bap_polynom_COEFF R, bap_polynom_COEFF A)
{
    bap_initial_and_reductum_polynom_COEFF (BAP_NOT_A_POLYNOM_COEFF, R, A);
}

/*
 * Variant of an above function where the variable v w.r.t. which the
 * initial and the reductum are defined can be specified.
 * It is assumed that v is greater than or equal to the leader of A.
 * The resulting polynomials are readonly.
 */

BAP_DLL void bap_initial_and_reductum2_polynom_COEFF 
	(bap_polynom_COEFF initial, bap_polynom_COEFF reductum, 
	 bap_polynom_COEFF A, bav_variable v)
{
    if (bap_is_numeric_polynom_COEFF (A) || bap_leader_polynom_COEFF (A) != v)
    {	if (initial != BAP_NOT_A_POLYNOM_COEFF)
	    bap_set_readonly_polynom_COEFF (initial, A);
	if (reductum != BAP_NOT_A_POLYNOM_COEFF)
	    bap_set_polynom_zero_COEFF (reductum);
    } else
	bap_initial_and_reductum_polynom_COEFF (initial, reductum, A);
}

/* Variant of an above function.  */

BAP_DLL void bap_initial2_polynom_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF A, bav_variable v)
{
    bap_initial_and_reductum2_polynom_COEFF (R, BAP_NOT_A_POLYNOM_COEFF, A, v);
}

/* Variant of an above function.  */

BAP_DLL void bap_reductum2_polynom_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF A, bav_variable v)
{
    bap_initial_and_reductum2_polynom_COEFF (BAP_NOT_A_POLYNOM_COEFF, R, A, v);
}

/*
 * Sets to R the coefficient of v^d in A.
 * It is assumed that v is greater than or equal to the leader of A.
 */

BAP_DLL void bap_coeff2_polynom_COEFF (bap_polynom_COEFF R, bap_polynom_COEFF A, 
						bav_variable v, bav_Idegree d)
{
    if (bap_is_numeric_polynom_COEFF (A) || bap_leader_polynom_COEFF (A) != v)
    {	if (d == 0)
	    bap_set_readonly_polynom_COEFF (R, A);
	else
	    bap_set_polynom_zero_COEFF (R);
    } else
    {	struct bap_itercoeff_COEFF iter;
	struct bav_term T;
	struct ba0_mark M;

	ba0_push_another_stack ();
	ba0_record (&M);
	bav_init_term (&T);
	bav_set_term_variable (&T, v, d);
	ba0_pull_stack ();

	bap_begin_itercoeff_COEFF (&iter, A, v);
	bap_seek_coeff_itercoeff_COEFF (R, &iter, &T);
	ba0_restore (&M);
    }
}

/*
 * Computes the initial and the reductum of A w.r.t. v (w.r.t the leader
 * of A if v is zero). The variable v does not need to be greater
 * than or equal to the leader of A. The resulting polynomials do not
 * share any piece of data with A and are not readonly.
 */

BAP_DLL void bap_lcoeff_and_reductum_polynom_COEFF 
		(bap_polynom_COEFF lcoeff, bap_polynom_COEFF reductum, 
			bap_polynom_COEFF A, bav_variable v)
{   struct bap_polynom_COEFF AA, lc, red;
    struct ba0_mark M;

    if ((lcoeff != BAP_NOT_A_POLYNOM_COEFF && lcoeff->readonly) ||
		(reductum != BAP_NOT_A_POLYNOM_COEFF && reductum->readonly))
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    if (bap_is_numeric_polynom_COEFF (A) ||
		(v != BAV_NOT_A_VARIABLE && v != bap_leader_polynom_COEFF (A)))
    {
	if (! bap_depend_polynom_COEFF (A, v))
	{   if (lcoeff != BAP_NOT_A_POLYNOM_COEFF && lcoeff != A)
		bap_set_polynom_COEFF (lcoeff, A);
	    if (reductum != BAP_NOT_A_POLYNOM_COEFF)
		bap_set_polynom_zero_COEFF (reductum);
	} else
	{   bav_Iordering r;

	    ba0_push_another_stack ();
	    ba0_record (&M);

	    r = bav_R_copy_ordering (bav_R_Iordering ());
	    bav_R_push_ordering (r);
	    bav_R_set_maximal_variable (v);

	    bap_init_readonly_polynom_COEFF (&AA);
	    bap_init_readonly_polynom_COEFF (&lc);
	    bap_init_readonly_polynom_COEFF (&red);
	    bap_sort_polynom_COEFF (&AA, A);
	    bap_initial_and_reductum_polynom_COEFF 
		(lcoeff != BAP_NOT_A_POLYNOM_COEFF ? 
			&lc : BAP_NOT_A_POLYNOM_COEFF, 
		reductum != BAP_NOT_A_POLYNOM_COEFF ? 
			&red : BAP_NOT_A_POLYNOM_COEFF, 
		&AA);

	    bav_R_pull_ordering ();
	    ba0_pull_stack ();
	    if (lcoeff != BAP_NOT_A_POLYNOM_COEFF)
		bap_set_polynom_COEFF (lcoeff, &lc);
	    if (reductum != BAP_NOT_A_POLYNOM_COEFF)
		bap_set_polynom_COEFF (reductum, &red);

	    bav_R_free_ordering (r);
	    ba0_restore (&M);
	}
    } else
    {	ba0_push_another_stack ();
	ba0_record (&M);

	bap_init_readonly_polynom_COEFF (&lc);
	bap_init_readonly_polynom_COEFF (&red);
	bap_initial_and_reductum_polynom_COEFF 
		(lcoeff != BAP_NOT_A_POLYNOM_COEFF ? 
			&lc : BAP_NOT_A_POLYNOM_COEFF, 
		reductum != BAP_NOT_A_POLYNOM_COEFF ? 
			&red : BAP_NOT_A_POLYNOM_COEFF, A);

	ba0_pull_stack ();
	if (lcoeff != BAP_NOT_A_POLYNOM_COEFF)
	    bap_set_polynom_COEFF (lcoeff, &lc);
	if (reductum != BAP_NOT_A_POLYNOM_COEFF)
	    bap_set_polynom_COEFF (reductum, &red);

	ba0_restore (&M);
    }
}

/*
 * Variant of the above function. Only the initial is computed.
 */

BAP_DLL void bap_lcoeff_polynom_COEFF 
		(bap_polynom_COEFF lcoeff, bap_polynom_COEFF A, bav_variable v)
{
    bap_lcoeff_and_reductum_polynom_COEFF 
				(lcoeff, BAP_NOT_A_POLYNOM_COEFF, A, v);
}

/*
 * Sets to R the coefficient of v^d in A
 * (v is the leader of A if zero).
 * The variable v does not need to be greater than or equal to the leader
 * of v. The result does not share any piece of data with A.
 */

BAP_DLL void bap_coeff_polynom_COEFF (bap_polynom_COEFF R, bap_polynom_COEFF A, 
				bav_variable v, bav_Idegree d)
{   struct bap_itercoeff_COEFF iter;
    struct bap_polynom_COEFF AA, coeff;
    struct bav_term T;
    struct ba0_mark M;

    if (R->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);
    if (bap_is_numeric_polynom_COEFF (A) ||
		(v != BAV_NOT_A_VARIABLE && v != bap_leader_polynom_COEFF (A)))
    {	if (! bap_depend_polynom_COEFF (A, v))
	{   if (d == 0)
	    {   if (R != A)
		    bap_set_polynom_COEFF (R, A);
	    } else 
		bap_set_polynom_zero_COEFF (R);
	} else
	{   bav_Iordering r;

	    ba0_push_another_stack ();
	    ba0_record (&M);

	    r = bav_R_copy_ordering (bav_R_Iordering ());
	    bav_R_push_ordering (r);
	    bav_R_set_maximal_variable (v);

	    bap_init_readonly_polynom_COEFF (&AA);
	    bap_init_readonly_polynom_COEFF (&coeff);
	    bap_sort_polynom_COEFF (&AA, A);
	    bap_begin_itercoeff_COEFF (&iter, &AA, v);
	    bav_init_term (&T);
	    bav_set_term_variable (&T, v, d);
	    bap_seek_coeff_itercoeff_COEFF (&coeff, &iter, &T);

	    bav_R_pull_ordering ();
	    ba0_pull_stack ();
	    bap_set_polynom_COEFF (R, &coeff);

	    bav_R_free_ordering (r);
	    ba0_restore (&M);
	}
    } else
    {	ba0_push_another_stack ();
	ba0_record (&M);

	bap_init_readonly_polynom_COEFF (&coeff);
	bav_init_term (&T);
	bav_set_term_variable (&T, bap_leader_polynom_COEFF (A), d);
	bap_begin_itercoeff_COEFF (&iter, A, bap_leader_polynom_COEFF (A));
	bap_seek_coeff_itercoeff_COEFF (&coeff, &iter, &T);

	ba0_pull_stack ();
	bap_set_polynom_COEFF (R, &coeff);

	ba0_restore (&M);
    }
}

/*
   Write A = 

      a_e v^e + ... + a_{d+1} v^{d+1} + a_d v^d + a_{d-1} v^{d-1} + ... + a_0

   Assigns a_d to C and 
	   a_e v^e + ... + a_{d+1} v^{d+1} + a_{d-1} v^{d-1} + ... + a_0 to R.

   Polynomials C and R may be zero.
*/

BAP_DLL void bap_coeff_and_reductum_polynom_COEFF
        (bap_polynom_COEFF C, bap_polynom_COEFF R,
         bap_polynom_COEFF A, bav_variable v, bav_Idegree d)
{   struct bap_itermon_COEFF iter;
    struct bap_creator_COEFF creaC, creaR;
    struct bap_polynom_COEFF CC, RR;
    struct bav_term T, U;
    struct ba0_mark M;
    ba0_int_p i, nbmon;

    if (R == (bap_polynom_COEFF)0)
    {	bap_coeff_polynom_COEFF (C, A, v, d);
	return;
    } else if (R->readonly)
        BA0_RAISE_EXCEPTION (BAP_ERRCOE);
    if (C != (bap_polynom_COEFF)0 && C->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);
    if (C == R)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (bap_is_numeric_polynom_COEFF (A) || 
		(v != BAV_NOT_A_VARIABLE && ! bap_depend_polynom_COEFF (A, v)))
    {	if (d == 0)
	{   if (C != A && C != (bap_polynom_COEFF)0)
		bap_set_polynom_COEFF (C, A);
	     bap_set_polynom_zero_COEFF (R);
	} else
	{   if (C != (bap_polynom_COEFF)0)
		bap_set_polynom_zero_COEFF (C);
	    if (R != A)
		bap_set_polynom_COEFF (R, A);
	}
	return;
    }

    nbmon = bap_nbmon_polynom_COEFF (A);

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&T);
    bav_init_term (&U);

    if (C != A && R != A)
    {	ba0_pull_stack ();
	if (C != (bap_polynom_COEFF)0)
	    bap_begin_creator_COEFF 
	        (&creaC, C, &A->rang_total, bap_approx_total_rank, nbmon/2+1);
	bap_begin_creator_COEFF
	    (&creaR, R, &A->rang_total, bap_approx_total_rank, nbmon/2+1);
	ba0_push_another_stack ();
    } else
    {	bap_init_polynom_COEFF (&CC);
	bap_init_polynom_COEFF (&RR);
	if (C != (bap_polynom_COEFF)0)
	    bap_begin_creator_COEFF 
	       (&creaC, &CC, &A->rang_total, bap_approx_total_rank, nbmon/2+1);
	bap_begin_creator_COEFF
	    (&creaR, &RR, &A->rang_total, bap_approx_total_rank, nbmon/2+1);
    }

    bap_begin_itermon_COEFF (&iter, A);
    while (! bap_outof_itermon_COEFF (&iter))
    {	bap_term_itermon_COEFF (&T, &iter);
	i = 0;
	while (i < T.size && T.rg [i].var != v)
	    i++;
	if ((i < T.size && T.rg [i].deg == d) || (i == T.size && d == 0))
	{   bav_exquo_term_variable (&U, &T, v, d);
	    if (C != A && R != A)
	    {	ba0_pull_stack ();
		if (C != (bap_polynom_COEFF)0)
		    bap_write_creator_COEFF 
			(&creaC, &U, *bap_coeff_itermon_COEFF (&iter));
		ba0_push_another_stack ();
	    } else
		bap_write_creator_COEFF
			(&creaC, &U, *bap_coeff_itermon_COEFF (&iter));
	} else
	{   if (C != A && R != A)
	    {   ba0_pull_stack ();
		bap_write_creator_COEFF 
				(&creaR, &T, *bap_coeff_itermon_COEFF (&iter));
		ba0_push_another_stack ();
	    } else
		bap_write_creator_COEFF
				(&creaR, &T, *bap_coeff_itermon_COEFF (&iter));
	}
	bap_next_itermon_COEFF (&iter);
    }
    bap_close_itermon_COEFF (&iter);

    if (C != A && R != A)
    {	ba0_pull_stack ();
	if (C != (bap_polynom_COEFF)0)
	    bap_close_creator_COEFF (&creaC);
	bap_close_creator_COEFF (&creaR);
	ba0_push_another_stack ();
    } else
    {	if (C != (bap_polynom_COEFF)0)
	    bap_close_creator_COEFF (&creaC);
	bap_close_creator_COEFF (&creaR);
    }

    ba0_pull_stack ();
    if (C == A || R == A)
    {	if (C != (bap_polynom_COEFF)0)
	    bap_set_polynom_COEFF (C, &CC);
	bap_set_polynom_COEFF (R, &RR);
    }
    ba0_restore (&M);
}

/* Returns the degree of A in v.  Zero has degree -1.  */

BAP_DLL bav_Idegree bap_degree_polynom_COEFF (bap_polynom_COEFF A, bav_variable v)
{   ba0_int_p i;

    if (bap_is_zero_polynom_COEFF (A))
	return -1;
/*
   On cherche si v figure dans le rang_total de A (indice i)
*/
    i = 0;
    while (i < A->rang_total.size && A->rang_total.rg [i].var != v)
	i++;
    return i == A->rang_total.size ? 0 : A->rang_total.rg [i].deg;
}

/*
 * Returns the total degree of A.
 * Zero has total degree -1.
 */

BAP_DLL bav_Idegree bap_total_degree_polynom_COEFF (bap_polynom_COEFF A)
{
    return bav_total_degree_term (&A->rang_total);
}

/*
 * Sets to R the polynomial obtained by replacing the initial of A by C.
 */

BAP_DLL void bap_replace_initial_polynom_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF A, bap_polynom_COEFF C)
{   struct bap_creator_COEFF crea;
    struct bap_itercoeff_COEFF iter;
    struct bap_itermon_COEFF itermon;
    bap_polynom_COEFF P;
    struct bav_term T;
    struct bav_rank rg;
    COEFF_t *lc;
    struct ba0_mark M;

    if (bap_is_numeric_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRCST);
    if (R->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    ba0_push_another_stack ();
    ba0_record (&M);

    bav_init_term (&T);
    bav_set_term (&T, &A->rang_total);
    bav_lcm_term (&T, &T, &C->rang_total);

    P = bap_new_polynom_COEFF ();
    bap_begin_creator_COEFF (&crea, P, &T, bap_approx_total_rank,
		bap_nbmon_polynom_COEFF (A) + bap_nbmon_polynom_COEFF (C));

    rg = bap_rank_polynom_COEFF (A);
    bap_begin_itermon_COEFF (&itermon, C);
    while (! bap_outof_itermon_COEFF (&itermon))
    {	bap_term_itermon_COEFF (&T, &itermon);
	bav_mul_term_rank (&T, &T, &rg);
	lc = bap_coeff_itermon_COEFF (&itermon);
	bap_write_creator_COEFF (&crea, &T, *lc);
	bap_next_itermon_COEFF (&itermon);
    }
    bap_begin_itercoeff_COEFF (&iter, A, bap_leader_polynom_COEFF (A));
    bap_next_itermon_COEFF (&iter.fin);
    while (! bap_outof_itermon_COEFF (&iter.fin))
    {	bap_term_itermon_COEFF (&T, &iter.fin);
	lc = bap_coeff_itermon_COEFF (&iter.fin);
	bap_write_creator_COEFF (&crea, &T, *lc);
	bap_next_itermon_COEFF (&iter.fin);
    }
    bap_close_creator_COEFF (&crea);
    ba0_pull_stack ();
    bap_set_polynom_COEFF (R, P);
    ba0_restore (&M);
}

/*
 * Sets the separant of A to R.
 */

BAP_DLL void bap_separant_polynom_COEFF (bap_polynom_COEFF R, bap_polynom_COEFF A)
{   struct bap_itermon_COEFF iter;
    struct bap_creator_COEFF crea;
    bap_polynom_COEFF P;
    COEFF_t c;
    struct bav_term T;
    struct bav_rank rg;
    struct ba0_mark M;

    if (bap_is_numeric_polynom_COEFF (A))
        BA0_RAISE_EXCEPTION (BAP_ERRCST);
    if (bap_monom_sequence_polynom_COEFF (A) != bap_decreasing_monom_sequence)
	BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    if (R->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);
    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init (c);
    bav_init_term (&T);
    bav_set_term (&T, &A->rang_total);

    rg = bav_rank_term (&T);

    if (T.rg [0].deg == 1)
	bav_shift_term (&T, &T);
    else
	T.rg [0].deg -= 1;

    P = bap_new_polynom_COEFF ();
    bap_begin_creator_COEFF (&crea, P, &T, 
			bap_approx_total_rank, bap_nbmon_polynom_COEFF (A));

    bap_begin_itermon_COEFF (&iter, A);
    bap_term_itermon_COEFF (&T, &iter);
    for (;;)
    {   if (T.rg [0].deg == 1)
	{   bav_shift_term (&T, &T);
	    bap_write_creator_COEFF 
			(&crea, &T, *bap_coeff_itermon_COEFF (&iter));
	} else
	{   COEFF_mul_ui (c, *bap_coeff_itermon_COEFF (&iter), T.rg [0].deg);
	    T.rg [0].deg -= 1;
	    bap_write_creator_COEFF (&crea, &T, c);
	}
	bap_next_itermon_COEFF (&iter);
	if (bap_outof_itermon_COEFF (&iter))
	    break;
	bap_term_itermon_COEFF (&T, &iter);
	if (bav_is_one_term (&T) || bav_leader_term (&T) != rg.var)
	    break;
    }
    bap_close_creator_COEFF (&crea);
    ba0_pull_stack ();
    bap_set_polynom_COEFF (R, P);
    ba0_restore (&M);
}

BAP_DLL void bap_separant2_polynom_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF A, bav_variable v)
{   struct bap_polynom_COEFF B;
    bav_Iordering r;
    struct ba0_mark M;

    if (! bap_depend_polynom_COEFF (A, v))
    	bap_set_polynom_zero_COEFF (R);
    else if (bap_leader_polynom_COEFF (A) == v)
	bap_separant_polynom_COEFF (R, A);
    else
    {
	ba0_push_another_stack ();
	ba0_record (&M);
	r = bav_R_copy_ordering (bav_R_Iordering ());
	bav_R_push_ordering (r);
	bav_R_set_maximal_variable (v);
	bap_init_readonly_polynom_COEFF (&B);
	bap_sort_polynom_COEFF (&B, A);
	ba0_pull_stack ();
	bap_separant_polynom_COEFF (R, &B);
	bav_R_pull_ordering ();
	bap_physort_polynom_COEFF (R, bap_definitive_transform);
	bav_R_free_ordering (r);
	ba0_restore (&M);
    }
}

/*
 * Sorting tables of polynomials
 */
/*
 * For qsort. See below
 */

static int comp_polynom_ascending (const void* x, const void* y)
{   bap_polynom_COEFF A = *(bap_polynom_COEFF*)x;
    bap_polynom_COEFF B = *(bap_polynom_COEFF*)y;
    struct bav_term TA, TB;
    struct ba0_mark M;
    enum ba0_compare_code code;

    ba0_record (&M);
    bav_init_term (&TA);
    bav_init_term (&TB);
    bap_leading_term_polynom_COEFF (&TA, A);
    bap_leading_term_polynom_COEFF (&TB, B);
    code = bav_compare_term (&TA, &TB);
    ba0_restore (&M);

    if (code == ba0_lt)
        return -1;
    else if (code == ba0_eq)
        return 0;
    else
        return 1;
}

static int comp_polynom_descending (const void* x, const void* y)
{   bap_polynom_COEFF A = *(bap_polynom_COEFF*)x;
    bap_polynom_COEFF B = *(bap_polynom_COEFF*)y;
    struct bav_term TA, TB;
    struct ba0_mark M;
    enum ba0_compare_code code;

    ba0_record (&M);
    bav_init_term (&TA);
    bav_init_term (&TB);
    bap_leading_term_polynom_COEFF (&TA, A);
    bap_leading_term_polynom_COEFF (&TB, B);
    code = bav_compare_term (&TA, &TB);
    ba0_restore (&M);

    if (code == ba0_gt)
        return -1;
    else if (code == ba0_eq)
        return 0;
    else
        return 1;
}

BAP_DLL void bap_sort_tableof_polynom_COEFF
		(bap_tableof_polynom_COEFF T, enum ba0_sort_mode mode)
{
    switch (mode)
    {   case ba0_descending_mode:
	    qsort (T->tab, T->size, sizeof (bap_polynom_COEFF),
						&comp_polynom_descending);
	    break;
	case ba0_ascending_mode:
	    qsort (T->tab, T->size, sizeof (bap_polynom_COEFF),
						&comp_polynom_ascending);
	    break;
    }
}

/****************************************************************************
 GARBAGE COLLECTOR
 ****************************************************************************/

/*
 * Readonly static data
 */

static char _struct_polynom [] = "struct bap_polynom_COEFF";
static char _struct_polynom_rang [] = "bap_polynom_COEFF->rang_total";

BAP_DLL ba0_int_p bap_garbage1_polynom_COEFF 
				(void* AA, enum ba0_garbage_code code)
{   bap_polynom_COEFF A = (bap_polynom_COEFF)AA;
    ba0_int_p n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info 
		    (A, sizeof (struct bap_polynom_COEFF), _struct_polynom);

    n += bap_garbage1_clot_COEFF (A->clot, ba0_isolated);

    if (A->rang_total.rg != (struct bav_rank*)0)
	n += ba0_new_gc_info
	    (A->rang_total.rg, sizeof(struct bav_rank) * A->rang_total.alloc,
	     _struct_polynom_rang);

    n += bap_garbage1_indexed_access (&A->ind, ba0_embedded);

    return n;
}

BAP_DLL void* bap_garbage2_polynom_COEFF (void* AA, enum ba0_garbage_code code)
{   bap_polynom_COEFF A;

    if (code == ba0_isolated)
	A = (bap_polynom_COEFF)ba0_new_addr_gc_info (AA, _struct_polynom);
    else
	A = (bap_polynom_COEFF)AA;

    A->clot = (bap_clot_COEFF)bap_garbage2_clot_COEFF (A->clot, ba0_isolated);

    if (A->rang_total.rg != (struct bav_rank*)0)
	A->rang_total.rg = (struct bav_rank*)ba0_new_addr_gc_info
			(A->rang_total.rg, _struct_polynom_rang);

    bap_garbage2_indexed_access (&A->ind, ba0_embedded);

    return A;
}

BAP_DLL void* bap_copy_polynom_COEFF (void* AA)
{   bap_polynom_COEFF A = (bap_polynom_COEFF)AA;
    bap_polynom_COEFF B;

    B = bap_new_polynom_COEFF ();
    bap_set_polynom_COEFF (B, A);
    return B;
}

BAP_DLL unsigned ba0_int_p bap_sizeof_polynom_COEFF (bap_polynom_COEFF F)
{   bap_polynom_COEFF P;
    struct ba0_mark A;
    struct ba0_mark B;
    unsigned ba0_int_p size;

    ba0_record (&A);
    P = bap_new_polynom_COEFF ();
    bap_set_polynom_COEFF (P, F);
    ba0_record (&B);
    size = ba0_range_mark (&A, &B);
    ba0_restore (&A);
    return size;
}

BAP_DLL void bap_switch_ring_polynom_COEFF 
				(bap_polynom_COEFF A, bav_differential_ring R)
{
    bav_switch_ring_term (&A->rang_total, R);
    bap_switch_ring_selection (&A->select, R);
    if (A->clot)
	bap_switch_ring_clot_COEFF (A->clot, R);
}

#endif
#undef BAD_FLAG_COEFF
