#include "bap_mul_polynom_COEFF.h"
#include "bap_parse_polynom_COEFF.h"
#include "bap_product_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/*
 * This structure permits us to represents products
 * $$p = c \, f_1^{a_1} \cdots f_t^{a_t}$$
 * where $c$ is a numerical factor, the
 * $f_i$ are polynomials not necessarily irreducible and not even pairwise
 * distinct. They are non constant. The exponents
 * $a_i$ may be zero.
 * 
 * The field {\tt num_factor} gives $c$.
 * The array {\tt tab} contains the pairs 
 * $(f_i, a_i)$. The field {\tt alloc} gives the number of pairs allocated
 * to {\tt tab}. The field {\tt size} gives the number of used pairs.
 * 
 * The product $1$ is coded by $c = 1$ and $t = 0$.
 * The product $0$ is coded by $c = 0$.
 * Observe that $0$ and $1$ are not necessarily recognized in the case of non
 * integral coefficient rings.
 */

/*
 * Powers are substructures of products.
 * They represent polynomials raised at some power.
 * In some functions, it is authorized that the polynomial is a constant,
 * possibly zero (for input purposes).
 */

/* Initializes $P$ to $1^1$.  */

BAP_DLL void bap_init_power_COEFF (bap_power_COEFF P)
{
    bap_init_polynom_one_COEFF (&P->factor);
    P->exponent = 1;
}

/* Allocates a power and initializes it to $1^1$.  */

BAP_DLL bap_power_COEFF bap_new_power_COEFF (void)
{   bap_power_COEFF P;

    P = (bap_power_COEFF)ba0_alloc (sizeof (struct bap_power_COEFF));
    bap_init_power_COEFF (P);
    return P;
}

/* Sets $P$ to $Q$.  */

BAP_DLL void bap_set_power_COEFF (bap_power_COEFF P, bap_power_COEFF Q)
{
    bap_set_polynom_COEFF (&P->factor, &Q->factor);
    P->exponent = Q->exponent;
}

/* Sets $P$ to $A^k$.  */

BAP_DLL void bap_set_power_polynom_COEFF 
			(bap_power_COEFF P, bap_polynom_COEFF A, bav_Idegree k)
{
    bap_set_polynom_COEFF (&P->factor, A);
    P->exponent = k;
}

/* Sets $P$ to $Q^k$.  */

BAP_DLL void bap_pow_power_COEFF 
			(bap_power_COEFF P, bap_power_COEFF Q, bav_Idegree k)
{
    if (P != Q)
	bap_set_polynom_COEFF (&P->factor, &Q->factor);
    P->exponent = Q->exponent * k;
}

/*
 * Lit une {\tt bap_power_COEFF} via l'analyseur lexical et l'affecte \`a~$A$.
 * Le param\`etre~$A$ peut \^etre nul.
 */

BAP_DLL void* bap_scanf_power_COEFF (void* A)
{   bap_power_COEFF P = (bap_power_COEFF)A;

    if (A == (void*)0)
	P = bap_new_power_COEFF ();
    bap_scanf_atomic_polynom_COEFF (&P->factor);
    ba0_get_token_analex ();
    if (ba0_sign_token_analex ("^"))
    {   ba0_get_token_analex ();
        if (ba0_type_token_analex () != ba0_integer_token)
            BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
        P->exponent = (bav_Idegree)atoi (ba0_value_token_analex ());
	if (P->exponent <= 0)
	    BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
    } else
    {   ba0_unget_token_analex ();
	P->exponent = 1;
    }
    return P;
}

/*
 * PRODUCTS
 */

static bool bap_has_non_trivial_factors_product_COEFF 
						(bap_product_COEFF, ba0_int_p);

/* Initializes {\tt prod} to $1$.  */

BAP_DLL void bap_init_product_COEFF (bap_product_COEFF prod)
{
    COEFF_init_set_ui (prod->num_factor, 1);
    prod->alloc = 0;
    prod->size = 0;
    prod->tab = (struct bap_power_COEFF*)0;
}

/* Initializes {\tt prod} to $0$.  */

BAP_DLL void bap_init_product_zero_COEFF (bap_product_COEFF prod)
{
    COEFF_init_set_ui (prod->num_factor, 0);
    prod->alloc = 0;
    prod->size = 0;
    prod->tab = (struct bap_power_COEFF*)0;
}

/*
 * Reallocs {\em prod} so that it can receive a product of at least $n$
 * powers. Existing factors of {\em prod} are kept.
 */

BAP_DLL void bap_realloc_product_COEFF (bap_product_COEFF prod, ba0_int_p n)
{   struct bap_power_COEFF* newtab;
    ba0_int_p i;

    if (n > prod->alloc)
    {	newtab = (struct bap_power_COEFF*)ba0_alloc 
					(sizeof (struct bap_power_COEFF) * n);
	memcpy (newtab, prod->tab, prod->size*sizeof(struct bap_power_COEFF));
	for (i = prod->size; i < n; i++)
	    bap_init_power_COEFF (&newtab [i]);
	prod->alloc = n;
	prod->tab = newtab;
    }
}

/* Allocates a new product and initializes it to $1$.  */

BAP_DLL bap_product_COEFF bap_new_product_COEFF (void)
{   bap_product_COEFF P;

    P = (bap_product_COEFF)ba0_alloc (sizeof (struct bap_product_COEFF));
    bap_init_product_COEFF (P);
    return P;
}

/* Allocates a new product and initializes it to $0$.  */

BAP_DLL bap_product_COEFF bap_new_product_zero_COEFF (void)
{   bap_product_COEFF P;

    P = (bap_product_COEFF)ba0_alloc (sizeof (struct bap_product_COEFF));
    bap_init_product_zero_COEFF (P);
    return P;
}

/* Sets $P$ to $1$.  */

BAP_DLL void bap_set_product_one_COEFF (bap_product_COEFF P)
{
    COEFF_set_si (P->num_factor, 1);
    P->size = 0;
}

/* Sets $P$ to $0$.  */

BAP_DLL void bap_set_product_zero_COEFF (bap_product_COEFF P)
{
    COEFF_set_si (P->num_factor, 0);
    P->size = 0;
}

/* Sets $P$ to $c$.  */

BAP_DLL void bap_set_product_numeric_COEFF (bap_product_COEFF P, COEFF_t c)
{
    P->size = 0;
    COEFF_set (P->num_factor, c);
}

/* Sets $P$ to $A^d$.  */

BAP_DLL void bap_set_product_polynom_COEFF 
		(bap_product_COEFF P, bap_polynom_COEFF A, bav_Idegree d)
{
    if (bap_is_zero_polynom_COEFF (A))
	bap_set_product_zero_COEFF (P);
    else if (d == 0)
	bap_set_product_one_COEFF (P);
    else if (bap_is_numeric_polynom_COEFF (A))
    {	struct ba0_mark M;
	COEFF_t c;

	ba0_push_another_stack ();
	ba0_record (&M);
	COEFF_init (c);
	COEFF_pow_ui (c, *bap_numeric_initial_polynom_COEFF (A), d);
	ba0_pull_stack ();
	bap_set_product_numeric_COEFF (P, c);
	ba0_restore (&M);
    } else
    {	bap_set_product_one_COEFF (P);
	bap_realloc_product_COEFF (P, 1);
	bap_set_polynom_COEFF (&P->tab [0].factor, A);
	P->tab [0].exponent = d;
	P->size = 1;
    }
}

/* Returns true if $P$ is the empty product.  */

static bool bap_has_non_trivial_factors_product_COEFF 
				(bap_product_COEFF, ba0_int_p);

BAP_DLL bool bap_is_one_product_COEFF (bap_product_COEFF P)
{
    return (! bap_has_non_trivial_factors_product_COEFF (P, 0)) && 
				COEFF_is_one (P->num_factor);
}

/* Returns true if $P$ is equal to zero.  */

BAP_DLL bool bap_is_zero_product_COEFF (bap_product_COEFF P)
{
    return COEFF_sgn (P->num_factor) == 0;
}

/* Returns true if $P$ is a constant.  */

BAP_DLL bool bap_is_numeric_product_COEFF (bap_product_COEFF P)
{
    return ! bap_has_non_trivial_factors_product_COEFF (P, 0);
}

/* Sets $P$ to $Q$. Does not copy the powers the exponents of which is zero. */

BAP_DLL void bap_set_product_COEFF (bap_product_COEFF P, bap_product_COEFF Q)
{   ba0_int_p p, q;

    if (P != Q)
    {	P->size = 0;
	bap_realloc_product_COEFF (P, Q->size);
	COEFF_set (P->num_factor, Q->num_factor);
	p = 0;
	for (q = 0; q < Q->size; q++)
	{   if (Q->tab [q].exponent > 0)
	    {	bap_set_polynom_COEFF 
				(&P->tab [p].factor, &Q->tab [q].factor);
		P->tab [p].exponent = Q->tab [q].exponent;
		p++;
	    }
	}
	P->size = p;
    }
}

/*
 * If $A$ is one of the factors of {\em prod}, returns its exponent.
 * Returns zero otherwise.
 */

BAP_DLL bav_Idegree bap_exponent_product_COEFF 
			(bap_product_COEFF prod, bap_polynom_COEFF A)
{   ba0_int_p i;

    for (i = 0; i < prod->size; i++)
	if (bap_equal_polynom_COEFF (&prod->tab [i].factor, A))
	    return prod->tab [i].exponent;
    return 0;
}

/*
 * Apply {\tt sort} over all the factors of $Q$. The result is in $P$. 
 * It is readonly.
 */

BAP_DLL void bap_sort_product_COEFF (bap_product_COEFF P, bap_product_COEFF Q)
{   ba0_int_p i;

    if (P == Q)
	for (i = 0; i < Q->size; i++)
	    bap_sort_polynom_COEFF (&P->tab [i].factor, &Q->tab [i].factor);
    else
    {	bap_set_product_numeric_COEFF (P, Q->num_factor);
	bap_realloc_product_COEFF (P, Q->size);
	for (i = 0; i < Q->size; i++)
	{   bap_sort_polynom_COEFF (&P->tab [i].factor, &Q->tab [i].factor);
	    P->tab [i].exponent = Q->tab [i].exponent;
	}
	P->size = Q->size;
    }
}

/* Apply {\tt physort} over all the factors of $P$.  */

BAP_DLL void bap_physort_product_COEFF 
			(bap_product_COEFF P, enum bap_typeof_transform type)
{   ba0_int_p i;

    for (i = 0; i < P->size; i++)
	bap_physort_polynom_COEFF (&P->tab [i].factor, type);
}

/* Sets $R$ to the product of the factors of {\em prod}.  */

BAP_DLL void bap_expand_product_COEFF 
				(bap_polynom_COEFF R, bap_product_COEFF prod)
{   bap_polynom_COEFF P, Q;
    struct bav_rank rg;
    ba0_int_p i, j;
    struct ba0_mark M;

    i = 0;
    while (i < prod->size && prod->tab [i].exponent == 0)
	i++;
    j = i + 1;
    while (j < prod->size && prod->tab [j].exponent == 0)
	j++;

    if (i == prod->size)
    {	rg = bav_constant_rank ();
	bap_set_polynom_crk_COEFF (R, prod->num_factor, &rg);
    } else if (j == prod->size)
    {	if (prod->tab [i].exponent > 1)
	{   if (COEFF_is_one (prod->num_factor))
		bap_pow_polynom_COEFF 
			(R, &prod->tab [i].factor, prod->tab [i].exponent);
	    else
	    {   ba0_push_another_stack ();
	        ba0_record (&M);
	        P = bap_new_polynom_COEFF ();
	        bap_pow_polynom_COEFF 
			(P, &prod->tab [i].factor, prod->tab [i].exponent);
	        ba0_pull_stack ();
	        bap_mul_polynom_numeric_COEFF (R, P, prod->num_factor);
	        ba0_restore (&M);
	    }
	} else if (COEFF_is_one (prod->num_factor))
	    bap_set_polynom_COEFF (R, &prod->tab [i].factor);
	else
	    bap_mul_polynom_numeric_COEFF 
			(R, &prod->tab [i].factor, prod->num_factor);
    } else
    {	ba0_push_another_stack ();
	ba0_record (&M);

	P = bap_new_polynom_COEFF ();
	Q = bap_new_polynom_COEFF ();
	bap_pow_polynom_COEFF 
			(P, &prod->tab [i].factor, prod->tab [i].exponent);
	do
	{
	    if (prod->tab [j].exponent > 1)
	    {   bap_pow_polynom_COEFF 
			(Q, &prod->tab [j].factor, prod->tab [j].exponent);
		bap_mul_polynom_COEFF (P, P, Q);
	    } else
		bap_mul_polynom_COEFF (P, P, &prod->tab [j].factor);
	    do 
		j++;
	    while (j < prod->size && prod->tab [j].exponent == 0);
	} while (j < prod->size);
	if (! COEFF_is_one (prod->num_factor))
	    bap_mul_polynom_numeric_COEFF (P, P, prod->num_factor);
	ba0_pull_stack ();
	bap_set_polynom_COEFF (R, P);
	ba0_restore (&M);
    }
}

/*
 * Sets {\em prod} to the product $\mbox{\em prod} \, A^k$.
 * Looks if $A$ is already a factor of {\em prod}.
 */

BAP_DLL void bap_mul_product_polynom_COEFF (bap_product_COEFF res, 
		bap_product_COEFF prod, bap_polynom_COEFF A, bav_Idegree k)
{   ba0_int_p i;

    if (bap_is_zero_polynom_COEFF (A))
	bap_set_product_zero_COEFF (res);
    else 
    {	if (res != prod)
        {   res->size = 0;
	    bap_realloc_product_COEFF (res, prod->size + 1);
	    bap_set_product_COEFF (res, prod);
	}

	if (bap_is_numeric_polynom_COEFF (A))
	{   COEFF_t p;
	    struct ba0_mark M;

	    ba0_push_another_stack ();
	    ba0_record (&M);
	    COEFF_init (p);
	    COEFF_pow_ui (p, *bap_numeric_initial_polynom_COEFF (A), k);
	    ba0_pull_stack ();
	    COEFF_mul (res->num_factor, res->num_factor, p);
	    ba0_restore (&M);
	} else if (k > 0)
	{   i = 0;
	    while (i < res->size && 
			! bap_equal_polynom_COEFF (&res->tab [i].factor, A))
		i++;
	    if (i < res->size)
		res->tab [i].exponent += k;
	    else
	    {   bap_realloc_product_COEFF (res, res->size + 1);
		bap_set_polynom_COEFF (&res->tab [i].factor, A);
		res->tab [i].exponent = k;
		res->size += 1;
	    }
	}
    }
}

BAP_DLL void bap_neg_product_COEFF (bap_product_COEFF R, bap_product_COEFF A)
{
    if (R == A)
	COEFF_neg (R->num_factor, R->num_factor);
    else
    {	bap_set_product_COEFF (R, A);
	COEFF_neg (R->num_factor, R->num_factor);
    }
}

/* Sets $R$ to the product $A \, c$.  */

BAP_DLL void bap_mul_product_numeric_COEFF 
			(bap_product_COEFF R, bap_product_COEFF A, COEFF_t c)
{
    if (COEFF_sgn (c) == 0 || bap_is_zero_product_COEFF (A))
	bap_set_product_zero_COEFF (R);
    if (R != A)
	bap_set_product_COEFF (R, A);
    COEFF_mul (R->num_factor, R->num_factor, c);
    if (COEFF_sgn (R->num_factor) == 0)
	bap_set_product_zero_COEFF (R);
}

/*
 * Sets {\em prod} to the product $\mbox{\em prod}\,T$.
 * The term $T$ is broken into elementary ranks.
 * Looks if each factor of $T$ is already a factor of {\em prod}.
 */

BAP_DLL void bap_mul_product_term_COEFF 
		(bap_product_COEFF res, bap_product_COEFF prod, bav_term T)
{   ba0_int_p i, j, nb_alloc, *tab;
    bool found;
    struct ba0_mark M;
    
    if (bav_is_one_term (T))
    {	bap_set_product_COEFF (res, prod);
	return;
    }

    ba0_push_another_stack ();
    ba0_record (&M);
    tab = (ba0_int_p*)ba0_alloc (sizeof (ba0_int_p) * T->size);
    ba0_pull_stack ();

    nb_alloc = 0;
    for (i = 0; i < T->size; i++)
    {	found = false;
	for (j = 0; ! found && j < prod->size; j++)
	{   struct bav_rank rg;

	    rg = bap_rank_polynom_COEFF (&prod->tab [j].factor);
	    if (bap_nbmon_polynom_COEFF (&prod->tab [j].factor) == 1 &&
				rg.var == T->rg [i].var && rg.deg == 1)
	    {	tab [i] = j;
		found = true;
	    }
	}
	if (! found)
	{   tab [i] = -1;
	    nb_alloc++;
	}
    }

    if (res != prod)
    {   res->size = 0;
	bap_realloc_product_COEFF (res, prod->size + nb_alloc);
	bap_set_product_COEFF (res, prod);
    } else
	bap_realloc_product_COEFF (res, res->size + nb_alloc);

    for (i = 0; i < T->size; i++)
    {	if (tab [i] >= 0)
	    res->tab [tab [i]].exponent += T->rg [i].deg;
	else
	{   bap_set_polynom_variable_COEFF 
			(&res->tab [res->size].factor, T->rg [i].var, 1);
	    res->tab [res->size].exponent = T->rg [i].deg;
	    res->size++;
	}
    }
    ba0_restore (&M);
}

/* Sets $R$ to the product $P\,Q$.  */

BAP_DLL void bap_mul_product_COEFF 
		(bap_product_COEFF R, bap_product_COEFF P, bap_product_COEFF Q)
{   ba0_int_p i;

    if (bap_is_zero_product_COEFF (P) || bap_is_zero_product_COEFF (Q))
    {	bap_set_product_zero_COEFF (R);
	return;
    }

    bap_realloc_product_COEFF (R, P->size + Q->size);

    if (R == Q)
    {	BA0_SWAP (bap_product_COEFF, P, Q);
    }

    if (R != P)
	bap_set_product_COEFF (R, P);

    COEFF_mul (R->num_factor, R->num_factor, Q->num_factor);
    for (i = 0; i < Q->size; i++)
	bap_mul_product_polynom_COEFF 
			(R, R, &Q->tab [i].factor, Q->tab [i].exponent);
}

/* Sets $P$ to $Q^k$.  */

BAP_DLL void bap_pow_product_COEFF 
		(bap_product_COEFF P, bap_product_COEFF Q, bav_Idegree k)
{   ba0_int_p j;

    if (bap_is_zero_product_COEFF (Q))
	bap_set_product_zero_COEFF (P);
    else if (k == 0)
	bap_set_product_one_COEFF (P);
    else
    {	bap_set_product_COEFF (P, Q);
	COEFF_pow_ui (P->num_factor, P->num_factor, k);
	for (j = 0; j < P->size; j++)
	     P->tab [j].exponent *= k;
    }
}

/*
 * Sets {\em prod} to the product $\mbox{\em prod} / A^k$.
 * The polynomial~$A$ must be a factor of {\em prod}.
 */

BAP_DLL void bap_exquo_product_polynom_COEFF (bap_product_COEFF res, 
		bap_product_COEFF prod, bap_polynom_COEFF A, bav_Idegree k)
{   COEFF_t p;
    ba0_int_p i;
    struct ba0_mark M;

    if (bap_is_zero_polynom_COEFF (A))
	BA0_RAISE_EXCEPTION (BAP_ERRNUL);

    if (res != prod)
	bap_set_product_COEFF (res, prod);

    if (bap_is_numeric_polynom_COEFF (A))
    {	
        ba0_push_another_stack ();
        ba0_record (&M);
        COEFF_init (p);
        COEFF_pow_ui (p, *bap_numeric_initial_polynom_COEFF (A), k);
        ba0_pull_stack ();

	COEFF_div (res->num_factor, res->num_factor, p);

	ba0_restore (&M);
    } else if (k > 0)
    {	i = 0;
	while (i < res->size && ! 
			bap_equal_polynom_COEFF (A, &res->tab [i].factor))
	    i++;

	if (i == res->size || res->tab [i].exponent < k)
	    BA0_RAISE_EXCEPTION (BAV_EXEXQO);

	if (res->tab [i].exponent > k)
	    res->tab [i].exponent -= k;
	else if (i == res->size - 1)
	    res->size -= 1;
	else
	{   BA0_SWAP (struct bap_power_COEFF, res->tab [i], 
						res->tab [res->size-1]);
	    res->size -= 1;
	}
    }
}

/*
 * Readonly static data
 */

static char _struct_product [] = "struct bap_product_COEFF";
#if defined (BAD_FLAG_mpz) || defined (BAD_FLAG_mpzm)
static char _struct_product_num [] = "bap_product_COEFF->num_factor._mp_d";
#else
#if defined (BAD_FLAG_mpq)
static char _struct_product_num_num [] = 
				"bap_product_COEFF->num_factor._mp_num._mp_d";
static char _struct_product_num_den [] = 
				"bap_product_COEFF->num_factor._mp_den._mp_d";
#endif
#endif
static char _struct_product_tab [] = "bap_product_COEFF->tab";

BAP_DLL ba0_int_p bap_garbage1_product_COEFF 
				(void* AA, enum ba0_garbage_code code)
{   bap_product_COEFF A = (bap_product_COEFF)AA;
    ba0_int_p i, n = 0;
    __mpz_struct* p;

    if (code == ba0_isolated)
	n += ba0_new_gc_info 
	    (A, sizeof(struct bap_product_COEFF), _struct_product);

#if defined (BAD_FLAG_mpz) || defined (BAD_FLAG_mpzm)
    p = A->num_factor;
    n += ba0_new_gc_info 
	(p->_mp_d, p->_mp_alloc * sizeof(mp_limb_t), 
	 _struct_product_num);
#else
#if defined (BAD_FLAG_mpq)
    p = &A->num_factor [0]._mp_num;
    n += ba0_new_gc_info 
	(p->_mp_d, p->_mp_alloc * sizeof(mp_limb_t),
	 _struct_product_num_num);
    p = &A->num_factor [0]._mp_den;
    n += ba0_new_gc_info 
	(p->_mp_d, p->_mp_alloc * sizeof(mp_limb_t),
	 _struct_product_num_den);
#else
    p = (__mpz_struct*)0;
#endif
#endif

    n += ba0_new_gc_info (A->tab, A->alloc * sizeof(struct bap_power_COEFF),
			  _struct_product_tab);
    for (i = 0; i < A->alloc; i++)
	n += bap_garbage1_polynom_COEFF (&A->tab [i].factor, ba0_embedded);

    return n;
}

BAP_DLL void* bap_garbage2_product_COEFF (void* AA, enum ba0_garbage_code code)
{   bap_product_COEFF A;
    ba0_int_p i;
    __mpz_struct* p;

    if (code == ba0_isolated)
	A = (bap_product_COEFF)ba0_new_addr_gc_info 
					(AA, _struct_product);
    else
	A = (bap_product_COEFF)AA;

#if defined (BAD_FLAG_mpz) || defined (BAD_FLAG_mpzm)
    p = A->num_factor;
    p->_mp_d = (mp_limb_t*)ba0_new_addr_gc_info 
			(p->_mp_d, _struct_product_num);
#elif defined (BAD_FLAG_mpq)
    p = &A->num_factor [0]._mp_num;
    p->_mp_d = (mp_limb_t*)ba0_new_addr_gc_info 
		(p->_mp_d, _struct_product_num_num);
    p = &A->num_factor [0]._mp_den;
    p->_mp_d = (mp_limb_t*)ba0_new_addr_gc_info 
		(p->_mp_d, _struct_product_num_den);
#else
    p = (__mpz_struct*)0;
#endif

    A->tab = (struct bap_power_COEFF*)ba0_new_addr_gc_info
					(A->tab, _struct_product_tab);
    for (i = 0; i < A->alloc; i++)
	bap_garbage2_polynom_COEFF (&A->tab [i].factor, ba0_embedded);

    return A;
}

BAP_DLL void* bap_copy_product_COEFF (void* A)
{   bap_product_COEFF B;

    B = bap_new_product_COEFF ();
    bap_set_product_COEFF (B, (bap_product_COEFF)A);
    return B;
}

/*
 * Lit un {\tt bap_product_COEFF} via l'analyseur lexical et l'affecte \`a~$A$.
 * Les factors num\'eriques sont regroup\'es dans {\tt num_factor}.
 * Le param\`etre~$A$ peut \^etre nul.
*/

BAP_DLL void* bap_scanf_product_COEFF (void* A)
{   bap_product_COEFF prod = (bap_product_COEFF)A;
    bap_power_COEFF P;
    ba0_list L;
    ba0_int_p i;
    bool negatif = false;
    bool nul = false;
    struct ba0_mark M;

    if (prod == (bap_product_COEFF)0)
	prod = bap_new_product_COEFF ();
    else
	bap_set_product_one_COEFF (prod);

    ba0_push_another_stack ();
    ba0_record (&M);

    if (ba0_sign_token_analex ("-"))
    {	negatif = true;
	ba0_get_token_analex ();
    }

    L = (ba0_list)0;
    P = bap_scanf_power_COEFF ((void*)0);
    if (bap_is_zero_polynom_COEFF (&P->factor))
	nul = true;
    L = ba0_cons_list (P, L);
    ba0_get_token_analex ();
    while (ba0_sign_token_analex ("*"))
    {	ba0_get_token_analex ();
	P = bap_scanf_power_COEFF ((void*)0);
	if (bap_is_zero_polynom_COEFF (&P->factor))
	    nul = true;
	L = ba0_cons_list (P, L);
	ba0_get_token_analex ();
    }
    ba0_unget_token_analex ();
    L = ba0_reverse_list (L);
    ba0_pull_stack ();

/*
   Affecte L a prod.
*/
    if (nul)
	bap_set_product_zero_COEFF (prod);
    else
    {	bap_realloc_product_COEFF (prod, ba0_length_list (L));
	i = 0;
	while (L != (ba0_list)0)
	{   P = (bap_power_COEFF)L->value;
	    if (bap_is_numeric_polynom_COEFF (&P->factor))
	    {   COEFF_t* lc = bap_numeric_initial_polynom_COEFF (&P->factor);
	        bav_Idegree e;
	        for (e = 0; e < P->exponent; e++)
		    COEFF_mul (prod->num_factor, prod->num_factor, *lc);
	    } else
	        bap_set_power_COEFF (&prod->tab [i++], P);
	    L = L->next;
        }
        prod->size = i;
        if (negatif)
	    COEFF_neg (prod->num_factor, prod->num_factor);
    }
  
    ba0_restore (&M);
    return prod;
}

/* Imprime le {\tt bap_product_COEFF} $A$.  */

BAP_DLL void bap_printf_product_COEFF (void* A)
{   bap_product_COEFF P = (bap_product_COEFF)A;
    COEFF_t bunk, *lc;
    ba0_int_p i;
    struct ba0_mark M;

    if (bap_is_zero_product_COEFF (P))
    {	ba0_put_char ('0');
	return;
    }

    ba0_push_another_stack ();
    ba0_record (&M);

    COEFF_init (bunk);

    if (COEFF_sgn (P->num_factor) < 0)
    {   ba0_put_string ("- ");
	COEFF_neg (bunk, P->num_factor);
    } else
	COEFF_set (bunk, P->num_factor);

    if (! COEFF_is_one (bunk))
    {	
#if defined (BAD_FLAG_mint_hp)
	ba0_printf ("%a_", &bunk);
#else
	ba0_printf ("%a_", bunk);
#endif
	if (bap_has_non_trivial_factors_product_COEFF (P, 0))
	{   if (ba0_LaTeX_flag)
		ba0_put_string ("\\,");
	    else
		ba0_put_char ('*');
	}
    } else if (! bap_has_non_trivial_factors_product_COEFF (P, 0))
    {
#if defined (BAD_FLAG_mint_hp)
	ba0_printf ("%a_", &bunk);
#else
	ba0_printf ("%a_", bunk);
#endif
    }

    for (i = 0; i < P->size; i++)
    {	if (P->tab [i].exponent == 0)
	    continue;
	lc = bap_numeric_initial_polynom_COEFF (&P->tab [i].factor);
	if (bap_nbmon_polynom_COEFF (&P->tab [i].factor) > 1 || 
							! COEFF_is_one (*lc))
	{   ba0_put_char ('(');
	    bap_printf_polynom_COEFF (&P->tab [i].factor);
	    ba0_put_char (')');
	} else
	    bap_printf_polynom_COEFF (&P->tab [i].factor);
	if (P->tab [i].exponent > 1)
	{   ba0_put_char ('^');
	    ba0_put_int_p (P->tab [i].exponent);
	}
	if (bap_has_non_trivial_factors_product_COEFF (P, i + 1))
	{   if (ba0_LaTeX_flag)
		ba0_put_string ("\\,");
	    else
		ba0_put_char ('*');
	}
    }

    ba0_pull_stack ();
    ba0_restore (&M);
}

/*
   Returns true if P has at least one non trivial factor in the
	range P->tab [i .. P->size - 1].
*/

static bool bap_has_non_trivial_factors_product_COEFF 
					(bap_product_COEFF P, ba0_int_p i)
{   ba0_int_p j;
    bool found;

    found = false;
    for (j = i; j < P->size && ! found; j++)
	found = P->tab [i].exponent > 0;
    return found;
}

#endif
#undef BAD_FLAG_COEFF
