#include "bap_selection.h"

/* 
 * A selection is a stack of pairs $(v, r)$ where~$v$ is a variable
 * and $r$ is a ordering number. 
 * Selections are used to suppress some factors in terms while accessing
 * to the terms of differential polynomials. They are part of the
 * {\tt bap_z_polynom_COEFF} data structure.
 * If $T$ is a term, one suppresses from~$T$ all the ranks $x^d$ such that
 * there exists a pair $(v,r)$ such that $x > v$ w.r.t. ordering~$r$.
 * The variable~$v$ may have value~$0$ or~$-1$.
 * In the first case, all the ranks of~$T$ are suppressed.
 * In the second case, no rank is suppressed.
 * These special value are not necessary at the top of the stack.
 * The ordering number~$r$ at the top of the stack is called the ordering
 * of the selection. It is used to sort the terms after stripping.
 */

BAP_DLL void bap_init_selection 
			(bap_selection s, bav_variable v, bav_Iordering r)
{
    s->i = 0;
    s->tab [0].varmax = v;
    s->tab [0].ordering = r;
}

/*
   Stores $\ast t$ in $\ast s$.
 */

BAP_DLL void bap_set_selection (bap_selection s, bap_selection t)
{
    *s = *t;
}

/*
   Returns the ordering of the selection.
*/

BAP_DLL bav_Iordering bap_ordering_selection (bap_selection s)
{
    return s->tab [s->i].ordering;
}

/*
   Returns the variable on the top of the stack.
*/

BAP_DLL bav_variable bap_variable_selection (bap_selection s)
{
    return s->tab [s->i].varmax;
}

/*
   Changes the ordering of the selection.
*/

BAP_DLL void bap_change_ordering_selection (bap_selection s, bav_Iordering r)
{
    s->tab [s->i].ordering = r;
}

/*
   Changes the variable at the top of the stack.
*/

BAP_DLL void bap_change_variable_selection (bap_selection s, bav_variable v)
{
    s->tab [s->i].varmax = v;
}

/*
   Stacks the pair $(v,r)$ in $\ast s$.
*/

BAP_DLL void bap_push_selection 
			(bap_selection s, bav_variable v, bav_Iordering r)
{
    if (bap_ordering_selection (s) == r)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    s->i++;
    if (s->i == BAP_SIZE_SELECTION)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
    s->tab [s->i].varmax = v;
    s->tab [s->i].ordering = r;
}

/*
   Unstacks.
*/

BAP_DLL void bap_pull_selection (bap_selection s)
{
    if (s->i == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    s->i--;
}

/*
 * The term $T$ is assumed to be ordered w.r.t. ordering~$r$.
 * Suppresses some ranks from $T$ (cf. the paragraph at the beginning
 * of the section) and then sorts the result if $r$ is different from
 * the ordering of the selection.
 */

BAP_DLL void bap_term_selection (bav_term T, bav_Iordering r, bap_selection s)
{   ba0_int_p i, j, n;
    bav_Iordering rr;
    bav_Inumber  nx;

    for (n = 0; n <= s->i; n++)
    {	if (s->tab [n].varmax == BAV_NOT_A_VARIABLE)
	{   bav_set_term_one (T);
	    return;
	} else if (s->tab [n].varmax != (bav_variable)-1)
	{   rr = s->tab [n].ordering;
	    nx = s->tab [n].varmax->number.tab [rr];
	    for (i = j = 0; j < T->size; j++)
		if (T->rg [j].var->number.tab [rr] <= nx)
		    T->rg [i++] = T->rg [j];
	    T->size = i;
	}
    }
    rr = bap_ordering_selection (s);
    if (rr != r)
    {	bav_R_push_ordering (rr);
	bav_sort_term (T);
	bav_R_pull_ordering ();
    }
}

/*
   Returns true if {\tt bap_term_selection (T, r, s)} leaves $T$ unchanged.
*/

BAP_DLL bool bap_identity_selection (bap_selection s, bav_Iordering r)
{   ba0_int_p i;

    if (r != bap_ordering_selection (s))
	return false;
    for (i = 0; i <= s->i; i++)
	if (s->tab [i].varmax != (bav_variable)-1)
	    return false;
    return true;
}

BAP_DLL void bap_switch_ring_selection 
				(bap_selection s, bav_differential_ring R)
{   ba0_int_p i;

    for (i = 0; i <= s->i; i++)
	if (s->tab [i].varmax != (bav_variable)-1)
	    s->tab [i].varmax = bav_switch_ring_variable (s->tab [i].varmax, R);
}


