#include "bap_termanager.h"

/*
 * They are subobjects of bap_clot_COEFF.
 * They handle terms identified by bap_termkeys in bap_mont_COEFF.
 * 
 * Term managers are able to handle many different representations.
 * The implemented ones are:
 * 
 * - terms are not compressed (bap_basic_term_manager)
 * - terms are compressed (bap_zipping_termanager)
 * - terms are stored in the term hash table (bap_hashing_term_manager)
 * 
 * The field rang_total contains a bound on the terms which are
 * recorded in the bap_clot_COEFF. It is an important field in the
 * case of the bap_zipping_termanager. It is currently not used by other
 * term managers.
 * 
 * The field u provides specific informations for the different term
 * managers about the meaning of the termkeys. 
 * 
 * - Basic term managers: a termkey is a pointer to a struct bav_term.
 *   The fields alloc and size are equal to the size of that structure
 *   counted in numbers ba0_int_p.
 * 
 * - Hashing term managers: no information is needed. Termkeys are hash 
 *   codes of the terms.
 * 
 * - Zipping term managers: 
 *   A termkey is a pointer to an array of ba0_int_p.
 *   The field alloc gives the number of ba0_int_p allocated
 *   to every termkey. If it is less than or equal to 1, no ba0_int_p
 *   is allocated: the information is stored in the pointer itself.
 *   The field size gives the number of ba0_int_p actually used.
 *   Degrees are stored in the ba0_int_p using the least number of bits
 *   possible. The number of bits attributed to a variable is about the log 
 *   of the degree of the variable in rang_total. The arrays nbbits and 
 *   mask have size rang_total.alloc. The integer nbbits [i] 
 *   gives the number of bits needed to represent the degree of 
 *   rang_total.rg [i].var. The integer mask [i]
 *   gives the mask needed to extract this information.
 */

/*
   The following function applies only to bap_zipping_termanagers.
   It computes  tgest->u.zipping.size.
	 	tgest->u.zipping.nbbits
		tgest->u.zipping.mask
   It *does not* compute tgest->u.zipping.alloc !
   For tgest->u.zipping.size, one must take care to the knapsack problem
   which may arise in the case of a change of ordering.
*/

static void compute_u_zipping (bap_termanager tgest)
{   bav_Idegree d;
    ba0_int_p s, m;
    ba0_int_p i, z;

    if (tgest->rang_total.size == 0)
    {	tgest->u.zipping.size = 0;
	return;
    }

    s = 0;
    m = -1;
/*
   s = sum of nbbits [i]
   m = max of nbbits [i]
*/
    for (i = 0; i < tgest->rang_total.size; i++)
    {   
        for (z = 0, d = tgest->rang_total.rg [i].deg; d > 0; d >>= 1)
            z++;
	tgest->u.zipping.nbbits [i] = (char)z;
        tgest->u.zipping.mask [i] = 
			(unsigned ba0_int_p)-1 >> (BA0_NBBITS_INT_P - z);
	s += z;
	if (z > m)
	    m = z;
    }
/*
   What is the value of tgest->u.zipping.size ?

   One knows that one must store s bits on ba0_int_p.
   At each change of ba0_int_p, one may "loose" at most m-1 bits

   We are thus looking for x such that
	BA0_NBBITS_INT_P * x >= s + (x - 1) * (m - 1)
   i.e. such that
	(BA0_NBBITS_INT_P - m + 1) * x >= s - m + 1
   Therefore
					          s - m + 1
	tgest->u.zipping.size = x = ceil ( ------------------------ )
					   BA0_NBBITS_INT_P - m + 1
*/
    {	ba0_int_p p, q;

	p = s - m + 1;
	q = BA0_NBBITS_INT_P - m + 1;
	tgest->u.zipping.size = p % q == 0 ? p / q : p / q + 1;
    }
}

/* 
 * Initializes {\em tgest}.
 * The total rank of the terms handled by this term manager must be a
 * divisor of~$T$.
 */

BAP_DLL void bap_init_termanager 
	(bap_termanager tgest, enum bap_typeof_termanager type, bav_term T)
{
    bav_init_term (&tgest->rang_total);
    bav_set_term (&tgest->rang_total, T);
    tgest->type = type;
    memset (&tgest->u, 0, sizeof (tgest->u)); /* for legibility */
    switch (type)
    {	case bap_basic_termanager:
	    tgest->u.basic.alloc = sizeof (struct bav_term)/sizeof (ba0_int_p);
	    tgest->u.basic.size = sizeof (struct bav_term)/sizeof (ba0_int_p);
	    break;
	case bap_zipping_termanager:
	    if (tgest->rang_total.alloc > 0)
	    {	tgest->u.zipping.nbbits = 
			(unsigned char*)ba0_alloc 
			   (sizeof (unsigned char) * tgest->rang_total.alloc);
	        tgest->u.zipping.mask = 
			(unsigned ba0_int_p*)ba0_alloc 
				(sizeof (unsigned ba0_int_p) * 
						tgest->rang_total.alloc);
	    } 
	    compute_u_zipping (tgest);
	    tgest->u.zipping.alloc = tgest->u.zipping.size;
	    break;
	case bap_hashing_termanager:
	    break;
    }
}

/* 
 * The term manager {\em tgest} is reinitialized with {\em type} and $T$.
 * Stores true in $\ast\mbox{\em reinit}$ if it is necessary to apply
 * {\tt bap_init_bap_termkey_array_termanager} to all the arrays formerly
 * handled by this term manager.
 */

BAP_DLL void bap_reset_termanager 
	(bap_termanager tgest, enum bap_typeof_termanager type, 
	 bav_term T, bool* reinit)
{   enum bap_typeof_termanager old_type = tgest->type;
    ba0_int_p old_rang_total_alloc = tgest->rang_total.alloc;

    bav_set_term (&tgest->rang_total, T);
    tgest->type = type;

    switch (old_type)
    {	case bap_basic_termanager:
	    switch (type)
	    {	case bap_basic_termanager:
		    *reinit = false;
		    break;
		case bap_zipping_termanager:
		    tgest->u.zipping.alloc = tgest->u.basic.alloc;
		    tgest->u.zipping.nbbits = (unsigned char*)ba0_alloc
			(sizeof (unsigned char) * tgest->rang_total.alloc);
		    tgest->u.zipping.mask = (unsigned ba0_int_p*)ba0_alloc
			(sizeof (unsigned ba0_int_p) * tgest->rang_total.alloc);
		    compute_u_zipping (tgest);
		    if (tgest->u.zipping.size > tgest->u.zipping.alloc)
		    {	*reinit = true;
			tgest->u.zipping.alloc = tgest->u.zipping.size;
		    } else
			*reinit = false;
		    break;
		case bap_hashing_termanager:
		    memset (&tgest->u, 0, sizeof (tgest->u));
		    *reinit = false;
		    break;
	    }
	    break;
	case bap_zipping_termanager:
	    switch (type)
	    {	case bap_basic_termanager:
/*
   *reinit = true even if the allocated area is large enough because
   the area is not initialized as a term data structure
*/
		    tgest->u.basic.alloc = 
				sizeof (struct bav_term)/sizeof(ba0_int_p);
		    tgest->u.basic.size =
				sizeof (struct bav_term)/sizeof(ba0_int_p);
		    *reinit = true;
		    break;
		case bap_zipping_termanager:
		    if (tgest->rang_total.alloc > old_rang_total_alloc)
		    {	tgest->u.zipping.nbbits = (unsigned char*)ba0_alloc 
			  (sizeof (unsigned char) * tgest->rang_total.alloc);
			tgest->u.zipping.mask = (unsigned ba0_int_p*)ba0_alloc 
				(sizeof (unsigned ba0_int_p) * 
						tgest->rang_total.alloc);
		    }
		    compute_u_zipping (tgest);
		    if (tgest->u.zipping.size > tgest->u.zipping.alloc)
		    {	*reinit = tgest->u.zipping.size > 1;
			tgest->u.zipping.alloc = tgest->u.zipping.size;
		    } else
		    	*reinit = false;
		    break;
		case bap_hashing_termanager:
		    memset (&tgest->u, 0, sizeof (tgest->u));
		    *reinit = false;
		    break;
	    }
	    break;
	case bap_hashing_termanager:
	    switch (type)
	    {	case bap_basic_termanager:
		    tgest->u.basic.alloc = 
				sizeof (struct bav_term)/sizeof (ba0_int_p);
		    tgest->u.basic.size =
				sizeof (struct bav_term)/sizeof (ba0_int_p);
		    *reinit = true;
		    break;
		case bap_zipping_termanager:
		    tgest->u.zipping.nbbits = (unsigned char*)ba0_alloc
			(sizeof (unsigned char) * tgest->rang_total.alloc);
		    tgest->u.zipping.mask = (unsigned ba0_int_p*)ba0_alloc
			(sizeof (unsigned ba0_int_p) * tgest->rang_total.alloc);
		    compute_u_zipping (tgest);
		    if (tgest->u.zipping.size > 1)
		    {	tgest->u.zipping.alloc = tgest->u.zipping.size;
			*reinit = true;
		    } else
		    {	*reinit = false;
			tgest->u.zipping.alloc = tgest->u.zipping.size;
		    }
		    break;
		case bap_hashing_termanager:
		    *reinit = false;
		    break;
	    }
    }
}

/*
   Returns true if termkeys are interpreted exactly the same way by A and B.
 */

BAP_DLL bool bap_equal_termanager (bap_termanager A, bap_termanager B)
{   bool b = false;

    if (A->type != B->type)
	return false;
    switch (A->type)
    {	case bap_basic_termanager:
	    b = true;
	    break;
	case bap_zipping_termanager:
	    b = bav_equal_term (&A->rang_total, &B->rang_total);
	    break;
	case bap_hashing_termanager:
	    b = true;
    }
    return b;
}

/*
   Initializes {\em dst} and stores a copy of {\em src} in it.
*/

BAP_DLL void bap_init_set_termanager (bap_termanager dst, bap_termanager src)
{   ba0_int_p alloc;

    bav_init_term (&dst->rang_total);
    bav_set_term (&dst->rang_total, &src->rang_total);
    dst->type = src->type;
    dst->u = src->u;
    switch (dst->type)
    {	case bap_basic_termanager:
	    break;
	case bap_zipping_termanager:
	    alloc = dst->rang_total.alloc;
    	    dst->u.zipping.nbbits = 
		(unsigned char*)ba0_alloc (sizeof (unsigned char) * alloc);
    	    dst->u.zipping.mask = (unsigned ba0_int_p*)ba0_alloc 
					(sizeof (unsigned ba0_int_p) * alloc);
	    memcpy (dst->u.zipping.nbbits, src->u.zipping.nbbits, 
					alloc * sizeof (unsigned char));
	    memcpy (dst->u.zipping.mask, src->u.zipping.mask, 
					alloc * sizeof (ba0_int_p));
	    break;
	case bap_hashing_termanager:
	    break;
    }
}

/*
   Initializes the elements of {\em tab}, which involves $n$ entries.
*/

BAP_DLL void bap_init_termkey_array_termanager 
			(bap_termanager tgest, bap_termkey* tab, ba0_int_p n)
{   ba0_int_p i;

    switch (tgest->type)
    {	case bap_basic_termanager:
	    for (i = 0; i < n; i++)
		tab [i] = (bap_termkey)bav_new_term ();
	    break;
	case bap_zipping_termanager:
	    if (tgest->u.zipping.alloc > 1)
		for (i = 0; i < n; i++)
		    tab [i] = (bap_termkey)ba0_alloc 
				(sizeof (ba0_int_p) * tgest->u.zipping.alloc);
	    break;
	case bap_hashing_termanager:
	    break;
    }
}

/*
   Stores s_tclef which is handled by s_tgest to *d_tclef which is handled by 
   d_tgest. This function is only called by functions of the kind of write_all. 
   It is thus only called when term managers are compatible, for efficiency.
*/

BAP_DLL void bap_set_termkey_from_termkey_termanager (bap_termanager d_tgest, 
	bap_termkey* d_tclef, bap_termanager s_tgest, bap_termkey s_tclef)
{
    if (d_tgest->type != s_tgest->type || 
	(d_tgest->type == bap_zipping_termanager && 
	 d_tgest->u.zipping.size != s_tgest->u.zipping.size))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    switch (d_tgest->type)
    {	case bap_basic_termanager:
	    bav_set_term (*(bav_term*)d_tclef, (bav_term)s_tclef);
	    break;
	case bap_zipping_termanager:
	    if (d_tgest->u.zipping.alloc <= 1)
	    {	if (s_tgest->u.zipping.alloc <= 1)
		    *d_tclef = s_tclef;
		else
		    *d_tclef = *(bap_termkey*)s_tclef;
	    } else
	    {	if (s_tgest->u.zipping.alloc <= 1)
		    **(bap_termkey**)d_tclef = s_tclef;
		else
		{   ba0_int_p *p, *q, i;
		    p = *(ba0_int_p**)d_tclef;
		    q = (ba0_int_p*)s_tclef;
		    for (i = 0; i < d_tgest->u.zipping.size; i++)
			*p++ = *q++;
		}
	    }
/*
	    {	struct bav_term T, U;
		struct ba0_mark M;
		ba0_record (&M);
		bav_init_term (&T);
		bav_init_term (&U);
		bap_set_term_from_termkey_termanager (d_tgest, &T, *d_tclef);
		bap_set_term_from_termkey_termanager (s_tgest, &U, s_tclef);
		if (! bav_equal_term (&T, &U))
		    BA0_RAISE_EXCEPTION (BA0_ERRALG);
		if (! bav_equal_term 
				(&d_tgest->rang_total, &s_tgest->rang_total))
		    BA0_RAISE_EXCEPTION (BA0_ERRALG);
		if (! bav_is_factor_term 
				(&d_tgest->rang_total, &T, (bav_term)0))
		    BA0_RAISE_EXCEPTION (BA0_ERRALG);
		ba0_restore (&M);
	    }
*/
	    break;
	case bap_hashing_termanager:
	    *d_tclef = s_tclef;
	    break;
    }
}

/*
   Stores the key of term~$T$ in *tclef
*/

BAP_DLL void bap_set_termkey_from_term_termanager 
			(bap_termanager tgest, bap_termkey* tclef, bav_term T)
{   ba0_int_p i, j, k, m, z, *tab;
/*
   Catches a hell a lot of bugs
*/
    if (! bav_is_factor_term (&tgest->rang_total, T, (bav_term)0))
	BA0_RAISE_EXCEPTION (BAP_ERRTGS);

    switch (tgest->type)
    {	case bap_basic_termanager:
	    bav_set_term ((bav_term)*tclef, T);
	    break;
	case bap_zipping_termanager:
/*
   tab points to the array of ba0_int_p which receives the compressed term
   i runs on tgest->rang_total
   j runs on T
   k runs on tab
   z is the ba0_int_p being built (i.e. tab [k])
   m = how many times the current degree must be shifted
*/
	    tab = tgest->u.zipping.alloc <= 1 ? 
				(ba0_int_p*)tclef : (ba0_int_p*)*tclef;
	    i = j = k = 0;
	    while (j < T->size)
	    {   m = 0;
		z = 0;
		while (j < T->size && 
			m + tgest->u.zipping.nbbits [i] <= 
						(ba0_int_p)BA0_NBBITS_INT_P)
		{   if (tgest->rang_total.rg [i].var == T->rg [j].var)
			z |= T->rg [j++].deg << m;
		    m += tgest->u.zipping.nbbits [i];
		    i++;
		}
		tab [k++] = z;
	    }
	    while (k < tgest->u.zipping.size)
		tab [k++] = 0;
/*
	    {	struct bav_term U;
		struct ba0_mark M;

		ba0_record (&M);
		bav_init_term (&U);
		bap_set_term_from_termkey_termanager (tgest, &U, *tclef);
		if (! bav_equal_term (&U, T))
		    BA0_RAISE_EXCEPTION (BA0_ERRALG);

		ba0_restore (&M);
	    }
*/
	    break;
	case bap_hashing_termanager:
	    *tclef = (bap_termkey)bav_hash_term (T);
	    break;
    }
}

/*
   Recovers the term from the key. Result in T.
*/

BAP_DLL void bap_set_term_from_termkey_termanager 
			(bap_termanager tgest, bav_term T, bap_termkey tclef)
{   ba0_int_p i, j, k, m, d, z, *tab;

    switch (tgest->type)
    {	case bap_basic_termanager:
	    bav_set_term (T, (bav_term)tclef);
	    break;
	case bap_zipping_termanager:
/*
   tab points to the array of ba0_int_p to be uncompressed
   i runs on tgest->rang_total
   j runs on T
   k runs on tab
   z = the ba0_int_p being uncompressed (i.e. tab [k])
   m = how many times z must be shifted to get the current degree

   the first loop determines the size of the term
   the second loop builds the term
*/
	    tab = tgest->u.zipping.alloc <= 1 ? 
				(ba0_int_p*)&tclef : (ba0_int_p*)tclef;
	    i = j = k = 0;
	    while (i < tgest->rang_total.size)
	    {	m = 0;
		z = tab [k];
		while (i < tgest->rang_total.size && 
			m + tgest->u.zipping.nbbits [i] <= 
						(ba0_int_p)BA0_NBBITS_INT_P)
		{   d = z & tgest->u.zipping.mask [i];
		    if (d != 0)
			j++;
		    m += tgest->u.zipping.nbbits [i];
		    z >>= tgest->u.zipping.nbbits [i];
		    i++;
		}
		k++;
	    }
	    bav_set_term_one (T);
	    bav_realloc_term (T, j);
	    i = j = k = 0;
	    while (i < tgest->rang_total.size)
	    {	m = 0;
		z = tab [k];
		while (i < tgest->rang_total.size && 
			m + tgest->u.zipping.nbbits [i] <= 
						(ba0_int_p)BA0_NBBITS_INT_P)
		{   d = z & tgest->u.zipping.mask [i];
		    if (d != 0)
		    {	T->rg [j].var = tgest->rang_total.rg [i].var;
			T->rg [j++].deg = d;
		    }
		    m += tgest->u.zipping.nbbits [i];
		    z >>= tgest->u.zipping.nbbits [i];
		    i++;
		}
		k++;
	    }
	    T->size = j;
	    break;
	case bap_hashing_termanager:
	    bav_unhash_term (T, (bav_hash_code)tclef);
	    break;
    }
}

/**************************************************************************
 GARBAGE COLLECTOR
 **************************************************************************/

/* 
   The termanager is inlined in a larger structure.

   --> embedded
*/

/*
* Readonly static data
*/

static char _termanager [] = "bap_termanager->rang_total.rg";
static char _termanager_nbbits [] = "bap_termanager->u.zipping.nbbits";
static char _termanager_mask [] = "bap_termanager->u.zipping.mask";

BAP_DLL ba0_int_p bap_garbage1_termanager (void* TT, enum ba0_garbage_code code)
{   bap_termanager T = (bap_termanager)TT;
    ba0_int_p n = 0;

    if (code != ba0_embedded)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (T->rang_total.alloc > 0)
	n += ba0_new_gc_info 
		(T->rang_total.rg, 
		 sizeof(struct bav_rank) * T->rang_total.alloc, 
		 _termanager);

    if (T->rang_total.alloc > 0 && T->type == bap_zipping_termanager)
    {	n += ba0_new_gc_info 
		(T->u.zipping.nbbits, 
		 ba0_ceil_align (sizeof(unsigned char) * T->rang_total.alloc),
		 _termanager_nbbits);
	n += ba0_new_gc_info
		(T->u.zipping.mask,
		 sizeof(unsigned ba0_int_p) * T->rang_total.alloc,
		 _termanager_mask);
    }

    return n;
}

/*
   The termanager is inlined in a larger structure.
   The returned value is thus pointless.
*/

BAP_DLL void* bap_garbage2_termanager (void* TT, enum ba0_garbage_code code)
{   bap_termanager T = (bap_termanager)TT;
 
    if (code != ba0_embedded)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (T->rang_total.alloc > 0)
	T->rang_total.rg = (struct bav_rank*)ba0_new_addr_gc_info
	    (T->rang_total.rg, _termanager);
    if (T->rang_total.alloc > 0 && T->type == bap_zipping_termanager)
    {	T->u.zipping.nbbits = (unsigned char*)ba0_new_addr_gc_info
		(T->u.zipping.nbbits, _termanager_nbbits);
	T->u.zipping.mask = (unsigned ba0_int_p*)ba0_new_addr_gc_info
		(T->u.zipping.mask, _termanager_mask);
    }
    return (void*)0;
}

/*
   Returns true if it is necessary to apply the garbage collector functions
   to the bap_termkeys handled by {\em tgest}.
*/

BAP_DLL bool bap_worth_garbage_termkey_termanager (bap_termanager tgest)
{   bool b = false;

    switch (tgest->type)
    {	case bap_basic_termanager:
	    b = true;
	    break;
	case bap_zipping_termanager:
	    b = tgest->u.zipping.alloc > 1;
	    break;
	case bap_hashing_termanager:
	    b = false;
	    break;
    }
    return b;
}

/*
 * Readonly static data
 */

static char _struct_term [] = "struct bav_term*";
static char _struct_term_rg [] = "bav_term->rg";
static char _termkey [] = "bap_termkey";

BAP_DLL ba0_int_p bap_garbage1_termkey_termanager 
	(bap_termanager tgest, bap_termkey key, enum ba0_garbage_code code)
{   bav_term T;
    ba0_int_p n = 0;

    if (code != ba0_embedded)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);

    switch (tgest->type)
    {	case bap_basic_termanager:
/*
   The allocated area may be larger than sizeof (struct bav_term).
   So we cannot call bav_garbage1_term.
*/
	    T = (bav_term)key;
	    n += ba0_new_gc_info (T, sizeof(ba0_int_p) * tgest->u.basic.alloc, 
						_struct_term);
	    if (T->rg)
		n += ba0_new_gc_info (T->rg, 
			sizeof (struct bav_rank) * T->alloc, _struct_term_rg);
	    break;
	case bap_zipping_termanager:
	    if (tgest->u.zipping.alloc <= 1)
		BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    n += ba0_new_gc_info 
			((void*)key, 
			 sizeof(ba0_int_p) * tgest->u.zipping.alloc, 
			 _termkey);
	    break;
	case bap_hashing_termanager:
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    break;
    }
    return n;
}

BAP_DLL bap_termkey bap_garbage2_termkey_termanager 
	(bap_termanager tgest, bap_termkey key, enum ba0_garbage_code code)
{   bav_term T;

    if (code != ba0_embedded)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    switch (tgest->type)
    {	case bap_basic_termanager:
	    T = (bav_term)key;
	    T = (bav_term)ba0_new_addr_gc_info (T, _struct_term);
	    if (T->rg)
		T->rg = (struct bav_rank*)ba0_new_addr_gc_info 
						(T->rg, _struct_term_rg);
	    key = (bap_termkey)T;
	    break;
	case bap_zipping_termanager:
	    key = (bap_termkey)ba0_new_addr_gc_info ((void*)key, _termkey);
	    break;
	case bap_hashing_termanager:
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    break;
    }
    return key;
}

BAP_DLL void bap_switch_ring_termanager 
			(bap_termanager tgest, bav_differential_ring R)
{
    switch (tgest->type)
    {	case bap_basic_termanager:
	case bap_hashing_termanager:
	    BA0_RAISE_EXCEPTION (BA0_ERRNYP);
	    break;
	case bap_zipping_termanager:
	    bav_switch_ring_term (&tgest->rang_total, R);
	    break;
    }
}

